/* translation of arith_valence.pas
{
  Copyright 1997-99 by Frank Heckenbach, heckenb@mi.uni-erlangen.de

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, version 2.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
}

unit Arith_Valence;

interface

uses GPC, Tools, Simplicial, GMP, Arith, Arith_GMP;

procedure NoStatus;
procedure SmithNormalForm_ValenceElim (	   Commentator		  : Pointer; const BigFaces,SmallFaces:TFaces;MaxRank,Dim:TCount;
                                       Mode			  : Integer; const Coeff:mpz_t;
                                       var PivotCount		  : TCount;var TorsionList:PTorsionList;
					   Numerator, Denominator : String );
procedure WriteMatrix_ValenceBB (const BigFaces,SmallFaces	  : TFaces;var FileName : String);
procedure WriteMatrix_ValenceBB_File (const BigFaces,SmallFaces	  : TFaces; const FileName : String);


procedure SmithNormalForm_ValenceBB (	 Commentator		: Pointer; const FileName : String;MaxRank,Dim:TCount;
                                     Mode			: Integer; const Coeff:mpz_t;
                                     var PivotCount		: TCount;var TorsionList:PTorsionList;
					 Numerator, Denominator	: String);

*/
void NoStatus(){}
void SmithNormalForm_ValenceElim ( 
   Commentator comm, const TFaces BigFaces, const TFaces SmallFaces,
   TCount MaxRank, TCount Dim,
   int Mode//  : Integer; const mpz_t Coeff, 
   TCount& PivotCount, PTorsionList& TorsionList,
   char* Numerator, char* Denominator );

procedure WriteMatrix_ValenceBB (const TFaces BigFaces, const TFaces SmallFaces, char* FileName);
procedure WriteMatrix_ValenceBB_File (const TFaces BigFaces, const TFaces SmallFaces, const char* FileName);

procedure SmithNormalForm_ValenceBB (	 
   Commentator comm, const char* FileName, 
   TCount MaxRank, TCount Dim,
   int Mode//  : Integer; const mpz_t Coeff, 
   TCount& PivotCount, PTorsionList& TorsionList,
   char* Numerator, char* Denominator );
/*
implementation

{$B-}

{ consistent with matrix.h }
type
  TSparseVectors = array [0 .. MaxArray] of TSV;
  PSparseVectors = ^TSparseVectors;

  TSparseMatrixArray = record
    Rows : PSparseVectors;
    NumRows, NumColumns : TCount
  end;
  PSparseMatrixArray = ^TSparseMatrixArray;

  TInvariantFactor = record
    Coefficient : mpz_t;
    Count : TCount
  end;

  TInvariantFactors = array [0 .. MaxArray] of TInvariantFactor;
  PInvariantFactors = ^TInvariantFactors;

  TSmithForm = record
    InvariantFactors : PInvariantFactors;
    Rows, Columns, NumberOfInvariantFactors : TCount
  end;

procedure MatrixSmithFormValence (    Commentator	     : Pointer;
                                  Matrix		     : PSparseMatrixArray;
                                  FileName		     : CString;
                                  MaxRank, Dim		     : TCount;
                                  Algorithm, Mode	     : Integer;
                                  protected var Coeff	     : mpz_t;
                                  var SmithForm		     : TSmithForm;
				      Numerator, Denominator : CString
); asmname 'MatrixSmithFormValence';

procedure NoStatus;
begin
end;

procedure WriteMatrix_ValenceBB (const BigFaces,SmallFaces:TFaces;var FileName : String);
begin
   FileName := GetTempFileName;
   WriteMatrix_ValenceBB_File(BigFaces, SmallFaces, FileName);
end;

procedure WriteMatrix_ValenceBB_File (const BigFaces,SmallFaces:TFaces; const FileName : String);
var
  i, j, Dummy : Integer;
  f : Text;
  v : TSV;
begin
  Rewrite (f, FileName);
  Writeln (f, FacesCount (BigFaces), ' ', FacesCount (SmallFaces), ' M');
  i := 0;
  while GetNextRow (BigFaces, SmallFaces, v, 0, i, Dummy, NoStatus) do
    for j := 0 to v.Count - 1 do
      with v.Components^[j] do Writeln (f, i, ' ', Index + 1, ' ', Val);
  Writeln (f, '0 0 0');
  Close (f)
end;

procedure SNF2TL (const SmithForm : TSmithForm; var PivotCount:TCount;var TorsionList:PTorsionList);
var
  i : Integer;
  pp : PPTorsionList;
begin
  with SmithForm do
    begin
      pp := @TorsionList;
      PivotCount := 0;
      for i := 0 to NumberOfInvariantFactors - 1 do
        with InvariantFactors^[i] do
          begin
            Inc (PivotCount, Count);
            if mpz_cmp_ui (Coefficient, 1) <> 0 then
              begin
                New (pp^);
                pp^^.TVal.GMPVal := Coefficient;
                pp^^.Count := Count;
                pp := @pp^^.Next
              end
          end;
      pp^ := nil
    end
end;

procedure SmithNormalForm_ValenceElim (	   Commentator		  : Pointer; const BigFaces,SmallFaces:TFaces;MaxRank,Dim:TCount;
                                       Mode			  : Integer; const Coeff:mpz_t;
                                       var PivotCount		  : TCount;var TorsionList:PTorsionList;
					   Numerator, Denominator : String );
var
  Matrix : TSparseMatrixArray;
  SmithForm : TSmithForm;
  i, j, Dummy : Integer;
begin
  Matrix.NumColumns := FacesCount (SmallFaces);
  Matrix.NumRows := FacesCount (BigFaces);
  GetMem (Matrix.Rows, Matrix.NumRows * SizeOf (Matrix.Rows^[1]));
  i := 0;
  repeat until not GetNextRow (BigFaces, SmallFaces, Matrix.Rows^[i], 0, i, Dummy, NoStatus);
  MatrixSmithFormValence (Commentator, @Matrix, nil, MaxRank, Dim, haValenceElim, Mode, Coeff, SmithForm, Numerator, Denominator);
  for i := 0 to Matrix.NumRows - 1 do Dispose (Matrix.Rows^[i].Components);
  Dispose (Matrix.Rows);
  SNF2TL (SmithForm, PivotCount, TorsionList)
end;

procedure SmithNormalForm_ValenceBB (Commentator : Pointer; const FileName : String;MaxRank,Dim:TCount;
                                     Mode:Integer; const Coeff:mpz_t;
                                     var PivotCount:TCount;var TorsionList:PTorsionList;
					 Numerator, Denominator : String);
var
  SmithForm : TSmithForm;
  f : Text;
begin
  MatrixSmithFormValence (Commentator, nil, FileName, MaxRank, Dim, haValenceBB, Mode, Coeff, SmithForm, Numerator, Denominator);
  SNF2TL (SmithForm, PivotCount, TorsionList);
  Assign (f, FileName);
  Erase (f)
end;

end.
