/* translation of  impl part of comp_hom.pas
{
  Copyright 1997-99 by Frank Heckenbach, heckenb@mi.uni-erlangen.de

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, version 2.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
}

unit Comp_Hom;

interface

uses GPC, Commentator_Pas, Tools, Simplicial, GMP, Arith, Arith_GMP, Arith_Valence;

procedure ComputeHomology (var Commentator : TCommentator; var Facets : TFacets; MinDim, MaxDim : TCount; Algorithm, Mode, Coeff_i : Integer; Numerator, Denominator : String);
procedure ComputeBoundaryMap (var Commentator : TCommentator; var Facets : TFacets; MinDim, MaxDim : TCount);
procedure ComputeBoundaryMapFile (var Commentator : TCommentator; var Facets : TFacets; MinDim : TCount; const FileName : String);

implementation

{$B-}

procedure ComputeHomology (var Commentator : TCommentator; var Facets : TFacets; MinDim, MaxDim : TCount; Algorithm, Mode, Coeff_i : Integer; Numerator, Denominator : String);
const HomReduce = 1;
var
  BigFaces,SmallFaces:TFaces;
  FaceNumbers,Ranks:array[Dim0] of TCount;
  FileNames:array[Dim0] of TString;
  MaxCard,i:Dim0;
  MDim,c,CurrentMode:Integer;
  MaxRank:TCount;
  Homology:THomTable;
  Homology_GMP:THomTable_GMP;
  Coeff : mpz_t;
  Msg, TorsionStr : TString;

  procedure StoreBetti (i : Dim0);
  var Betti:TCount;
  begin
    if (Mode <> hmNTorsionRank) and (i >= MinDim) then
      begin
        Betti:=FaceNumbers[i]-Ranks[i]-Ranks[i+1];
        if (Ranks[i]<0) or (Ranks[i+1]<0) or (Betti<0) then Abort ('Internal error in ComputeHomology');
        if Algorithm = haInteger
          then Homology[i].Betti:=Betti
          else Homology_GMP[i].Betti:=Betti;
        if IsPrinted (Commentator, LvlImp, PartialResult) <> 0 then
          case Mode of
            hmGroup : begin
                        if Algorithm = haInteger
                          then WriteStrTorsion(TorsionStr,Homology[i].Torsion,True)
                          else WriteStrTorsion_GMP(TorsionStr,Homology_GMP[i].Torsion,True);
                        WriteStr (Msg, 'H_', i, ' = ', Betti, TorsionStr, '.');
                        ActivityReport (Commentator, Msg, LvlImp, PartialResult)
                      end;
            hmBetti : begin
                        WriteStr (Msg, 'B_', i, ' = ', Betti, '.');
                        ActivityReport (Commentator, Msg, LvlImp, PartialResult)
                      end;
          end
      end
  end;

begin
  mpz_init_set_ui (Coeff, Coeff_i);
  for c:=0 to MaximumDim do
    begin
      FaceNumbers[c]:=0;
      Ranks[c]:=-1;
      Homology[c].Torsion:=nil;
      Homology_GMP[c].Torsion:=nil
    end;
  MaxCard:=GetDimension(Facets)+1;
  Ranks[0]:=HomReduce;
  Ranks[MaxCard]:=0;
  if MaxDim=-1 then MaxDim:=MaxCard-1;
  MaxDim:=Min(MaxCard-1,MaxDim);
  if MinDim > MaxDim then
    begin
      Write ('[]');
      Exit
    end;
  MinDim:=Min(MaxCard-1,Max(0,MinDim));
  MaxDim:=Max(0,MaxDim);
  MDim := Max (MinDim - Ord (Mode <> hmNTorsionRank), 0);
  StartActivity (Commentator, 'homology', 'Computing homology groups', LvlUnimp, InternalDescription);
  GetFaces(Facets,BigFaces,MDim);
  FaceNumbers[MDim]:=FacesCount(BigFaces);
  for i:=MDim to MaxDim do
    begin
      if i <> MaxCard - 1 then
        begin
          if i >= MinDim
            then CurrentMode:=Mode
            else CurrentMode:= hmBetti;
          SmallFaces:=BigFaces;
          GetFaces(Facets,BigFaces,i+1);
          FaceNumbers[i+1]:=FacesCount(BigFaces);
          MaxRank:=FaceNumbers[i]-Max(Ranks[i],0);
          case Algorithm of
            haInteger     : if QuickCheck(BigFaces,SmallFaces)=MaxRank
                               then Ranks[i+1]:=MaxRank
                               else SmithNormalForm(BigFaces,SmallFaces,MaxRank,i+1,CurrentMode,Coeff_i,Ranks[i+1],Homology[i].Torsion);
            haGMP         : if QuickCheck_GMP(BigFaces,SmallFaces)=MaxRank
                               then Ranks[i+1]:=MaxRank
                               else SmithNormalForm_GMP(BigFaces,SmallFaces,MaxRank,i+1,CurrentMode,TVV_GMP(Coeff),Ranks[i+1],Homology_GMP[i].Torsion);
            haValenceElim : SmithNormalForm_ValenceElim (Commentator,BigFaces,SmallFaces,MaxRank,i,CurrentMode,Coeff,Ranks[i+1],Homology_GMP[i].Torsion, Numerator, Denominator);
            haValenceBB   : WriteMatrix_ValenceBB (BigFaces,SmallFaces,FileNames[i]);
          end;
          DisposeFaces(SmallFaces)
        end;
      if Algorithm <> haValenceBB then StoreBetti (i)
    end;
  DisposeFaces(BigFaces);
  if Algorithm = haValenceBB then
    for i:=MDim to MaxDim do
      begin
        CurrentMode:=Mode;
        if i <> MaxCard - 1 then
          SmithNormalForm_ValenceBB (Commentator,FileNames [i],FaceNumbers[i]-Max(Ranks[i],0),i,CurrentMode,Coeff,Ranks[i+1],Homology_GMP[i].Torsion, Numerator, Denominator);
        StoreBetti (i)
      end;
  if Algorithm = haInteger
    then WriteHomology(Output, Homology, MinDim, MaxDim)
    else WriteHomology_GMP(Output, Homology_GMP, MinDim, MaxDim);
  mpz_clear (Coeff);
  StopActivity (Commentator, 'Homology groups computed', LvlUnimp, InternalDescription)
end;
*/
#include "comp_hom.h"
/*

procedure ComputeBoundaryMap (var Commentator : TCommentator; var Facets : TFacets; MinDim, MaxDim : TCount);
var
  BigFaces,SmallFaces:TFaces;
  MaxCard,i:Dim0;
  Columns, j, r, c, Dummy : Integer;
  Row : TSV;
begin
  MaxCard:=GetDimension(Facets)+1;
  if MaxDim=-1 then MaxDim:=MaxCard-1;
  MaxDim:=Min(MaxCard-1,MaxDim);
  if MinDim > MaxDim then
    begin
      Write ('[]');
      Exit
    end;
  MinDim:=Max(1,Min(MaxCard-1,MinDim));
  MaxDim:=Max(0,MaxDim);
  GetFaces(Facets,BigFaces,MinDim - 1);
  Write ('[');
  for i:=MinDim to MaxDim do
    begin
      if i <> MinDim then Write (',');
      SmallFaces:=BigFaces;
      GetFaces(Facets,BigFaces,i);
      Columns := FacesCount (SmallFaces);
      r := 0;
      Write ('[');
      while GetNextRow (BigFaces, SmallFaces, Row, 0, r, Dummy, NoStatus) do
        begin
          if r <> 1 then Write (',');
          Write ('[');
          j := 0;
          for c := 0 to Columns - 1 do
            begin
              if c <> 0 then Write (',');
              if (j < Row.Count) and (Row.Components^[j].Index = c) then
                begin
                  Write (Row.Components^[j].Val);
                  Inc (j)
                end
              else
                Write ('0')
            end;
          Write (']');
          Dispose (Row.Components)
        end;
      Write (']');
      DisposeFaces(SmallFaces)
    end;
  Write (']');
  DisposeFaces(BigFaces)
end; { ComputeBoundaryMap }
*/


/*
procedure ComputeBoundaryMapFile (var Commentator : TCommentator; var Facets : TFacets; MinDim : TCount; const FileName : String);
var
  BigFaces,SmallFaces:TFaces;
  MaxCard,i:Dim0;
  Columns, j, r, c, Dummy : Integer;
  Row : TSV;
begin
   MaxCard:=GetDimension(Facets)+1;
   MinDim:=Max(1,Min(MaxCard-1,MinDim));
   GetFaces(Facets,SmallFaces,MinDim - 1);
   GetFaces(Facets,BigFaces,MinDim);
   WriteMatrix_ValenceBB_File(BigFaces, SmallFaces, FileName);
   DisposeFaces(SmallFaces);
   DisposeFaces(BigFaces);
end; { ComputeBoundaryMap }

end.
*/

void ComputeBoundaryMapFile (Commentator comm, TFacets& Facets, TCount MinDim,
					const char* FileName)
{
   TFaces BigFaces,SmallFaces;
   Dim0 MaxCard;

   MaxCard=GetDimension(Facets)+1;
   MinDim=Max(1,Min(MaxCard-1,MinDim));
   GetFaces(Facets,SmallFaces,MinDim - 1);
   GetFaces(Facets,BigFaces,MinDim);
   WriteMatrix_ValenceBB_File(BigFaces, SmallFaces, FileName);
}

