/*  manual translation to C++ -bds
{
  Program to read in one simplicial complex in GAP output format from standard
  input and compute its homology.
  Copyright 1997-99 by Frank Heckenbach, heckenb@mi.uni-erlangen.de

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, version 2.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
}

program Homology_GAP;

uses GPC, Commentator_Pas, Tools, Simplicial, Comp_Hom;

const
  numint  = 7;
  numargs = 10;

var
  r, c : Integer;
  Args : array [1 .. numint] of Integer;
  Facets : TFaces;
  ErrMsg : TString;

begin
   ErrMsg := '';
   if ParamCount <> numargs then ErrMsg := 'Wrong number of command-line arguments';
   for c := 1 to numint do
      if ErrMsg = '' then
      begin
	 Val (ParamStr (c), Args [c], r);
	 if r <> 0 then ErrMsg := 'Command-line argument is not an integer'
      end;
   Commentator := InitializeCommentator (Args [4], Args [3]);
   if ErrMsg <> '' then Abort (ErrMsg);
   ReadComplex (Facets);
   if Args [6] = hmMatrixFile
      then ComputeBoundaryMapFile (Commentator, Facets, Args [1], ParamStr (10))
   else if Args [6] = hmMatrix
      then ComputeBoundaryMap (Commentator, Facets, Args [1], Args [2] )
   else ComputeHomology (Commentator, Facets, Args [1], Args [2], Args [5], Args [6], Args [7], ParamStr (8), ParamStr (9));
   DisposeFaces (Facets)
end.
   
*/

//External:
////Type TFaces
////ReadComplex()
////hmMatrixFile
////ComputeBoundaryMapFile()
////hmMatrix
////ComputeBoundaryMap()
////ComputeHomology()
////DisposeFaces()
#include "commentator.h"
#include "comp_hom.h"

main(int argc, char* ParamStr[])
{ 
   const int numint = 7, numargs = 10;
   int r, c;
   long Args[numint+1];

   if (argc != numargs + 1)
   {  cerr << "Wrong number of command-line arguments (" << (argc-1) << ").\n"; 
      return 1;
   }
   for (c = 1; c <= numint; ++c)
      Args [c] = atoi(ParamStr[c]);
   Commentator comm(Args [4], Args [3]);

   TFaces Facets;
   ReadComplex (Facets);
   if (Args [6] == hmMatrixFile)
      ComputeBoundaryMapFile (comm, Facets, Args [1], ParamStr [10]);
   else if (Args [6] == hmMatrix)
      ComputeBoundaryMap (comm, Facets, Args [1], Args [2] );
   else ComputeHomology (comm, Facets, Args [1], Args [2], Args [5], Args [6], Args [7], ParamStr [8], ParamStr [9]);
   DisposeFaces (Facets);
}
