/* translation of interface portion of simplicial.pas
{
  Copyright 1997-99 by Frank Heckenbach, heckenb@mi.uni-erlangen.de

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, version 2.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
}

unit Simplicial;

interface

uses GPC, Commentator_Pas, Tools;

const
  MaxN=256;
  MaxLength=$FF;

type
  n0=0..MaxN;
  n1=1..MaxN;

  {Typ fuer einen Simplex}
  PPFace=^PFace;
  PFace=^TFace;
  TFace=array[0..MaxN] of TElementCount;

  PPFaceArray=^TPFaceArray;
  TPFaceArray=array[0..MaxArray] of PFace;

  {Typ fuer eine Menge von Simplizes, gespeichert als dynamisches
   Array mit Angabe ihrer Anzahl}
  TFaces=record
    Count:TCount;
    FaceP:PPFaceArray
  end;

  TFacets=TFaces;

{Legt eine neue Menge von Simplizes in der angegebenen Anzahl an}
procedure NewFaces(var Faces:TFaces;Count:TCount);

{Entfernt eine Menge von Simplizes}
procedure DisposeFaces(Faces:TFaces);

{Liefert einen Zeiger auf einen bestimmten Simplex in einer Menge}
function FaceGet(const Faces:TFaces;Index:TCount):PFace;

{Liefert einen Zeiger auf den Zeiger auf einen bestimmten Simplex in einer Menge}
function FacePGet(const Faces:TFaces;Index:TCount):PPFace;

{Sucht einen Simplex in einer Menge von Simplizes}
function FindFace(const Face:TFace;const Faces:TFaces):TCount;

{Liefert die Anzahl der Elemente einer Seite}
function FaceCount(const Face:TFace):n0;

{Liefert die Anzahl der Seiten einer Menge von Seiten}
function FacesCount(const Faces:TFaces):TCount;

{Liefert die Dimension (+1) eines Komplexes}
function GetDimension(const Facets:TFacets):Dim0;

{Erzeugt alle Simplizes eines Komplexes mit gegebenen Facetten}
procedure GetFaces(var Facets:TFacets;var Faces:TFaces;Dim:Dim0);

{Liefert eine Zeile der Randoperator-Matrix}
function GetOneRow(const BigFaces,SmallFaces:TFaces;RowCount:TCount):PSparseIntVector;

{ Liest einen Komplex ein }
procedure ReadComplex (var Facets:TFaces);
*/
#include "tools.h"

const int MaxN=256;
const int MaxLength=256;// MaxLength=$FF;

typedef int n0; // n0=0..MaxN;
typedef int n1; // n1=1..MaxN;

typedef TElementCount* TFace;
typedef TFace* PFace;

typedef PFace* TPFaceArray;

struct TFaces
{
   TFaces(int count = 0): Count(count){ FaceP = new PFace[count]; }
   ~TFaces(){ delete [] FaceP; }

   TCount Count;
   TPFaceArray FaceP;
};

typedef TFaces TFacets;

void NewFaces(TFaces* Faces, TCount Count)
{ Faces = new TFaces[Count]; }

void DisposeFaces(TFaces Faces)
{ Faces.~TFaces(); }

PFace FaceGet(const TFaces Faces, TCount Index)
{ return Faces.FaceP[Index]; }

TCount FindFace(const TFace Face, const TFaces Faces);

n0 FaceCount(const TFace Face);

TCount FacesCount(const TFaces Faces)
{ return Faces.Count; }

Dim0 GetDimension(const TFacets Facets);

void GetFaces(TFacets& Facets, TFaces&  Faces, Dim0 Dim);

PSparseIntVector GetOneRow(const TFaces BigFaces, const TFaces SmallFaces,
			      TCount RowCount);

void ReadComplex (TFaces& Facets);
/*********
==================
??
procedure FaceCopy(const Src:TFace;var Dest:TFace);
var i:n0;
begin
  for i:=0 to FaceCount(Src) do Dest[i]:=Src[i]
end;


{Vergleicht zwei Simplizes lexikographisch}
function FaceLess(const A,B:TFace):Boolean;
var i:n0;
begin
  i:=0;
  while (A[i]<>0) and (A[i]=B[i]) do Inc(i);
  FaceLess:=A[i]<B[i]
end;

{Stellt fest, ob $A=B$ gilt}
function FaceEqual(const A,B:TFace):Boolean;
var i:n0;
begin
  i:=0;
  while (A[i]<>0) and (A[i]=B[i]) do Inc(i);
  FaceEqual:=(A[i]=0) and (B[i]=0)
end;

{Sucht einen Simplex in einer Menge von Simplizes}
{Binaere Suche, da diese Funktion zeitkritisch ist}
function FindFace(const Face:TFace;const Faces:TFaces):TCount;
var
  a,b,c:TCount;
  p:PFace;
  Found:Boolean;
begin
  a:=0;
  b:=Faces.Count-1;
  repeat
    c:=(a+b) div 2;
    p:=FaceGet(Faces,c);
    Found:=FaceEqual(Face,p^);
    if not Found then
      if FaceLess(Face,p^) then b:=c-1 else a:=c+1
  until Found or (a>b);
  if Found
    then FindFace:=c
    else Abort ('Internal error in FindFace.')
end;
********/

/*
{Liefert die Anzahl der Elemente einer Seite}
function FaceCount(const Face:TFace):n0;
var Result:n0;
begin
  Result:=0;
  while (Result<MaxN) and (Face[Result]<>0) do Inc(Result);
  FaceCount:=Result
end;

{Liefert die Dimension (+1) eines Komplexes}
function GetDimension(const Facets:TFacets):Dim0;
var
  d:Dim0;
  c:TCount;
  Msg:TString;
begin
  d:=0;
  for c:=0 to Facets.Count-1 do
    d:=Max(d,FaceCount(FaceGet(Facets,c)^));
  WriteStr (Msg,'Simplicial complex of dimension ', d-1,' with ',Facets.Count,' facets');
  ActivityReport (Commentator, Msg, LvlNormal, InternalDescription);
  GetDimension:=d-1
end;

{Erzeugt alle Simplizes eines Komplexes mit gegebenen Facetten}
procedure GetFaces(var Facets:TFacets;var Faces:TFaces;Dim:Dim0);
const MaxI=2*MaxN;
type
  {Typ fuer eine Liste von Simplizes}
  PPFaceList=^PFaceList;
  PFaceList=^TFaceList;
  TFaceList=record
    Next:PFaceList;
    FaceP:PFace
  end;

  {Typ fuer eine Liste von Simplizes und die Anzahl der Elemente}
  TFaceMerge=record
    Count:TCount;
    List:PFaceList
  end;

var
  c:TCount;
  FacetCard:n0;
  j,k:0..MaxN+1;
  i:1..MaxI;
  si:n1;
  Facet,SubSet:TFace;
  SubSetIndex:array[n0] of n0;
  Msg, FacesStr : TString;

  {Ein Array aus Listen von Simplizes.
   Dieses Array wird so benutzt, dass fuer jedes i die i-te Liste
   zu jeder Zeit hoechstens 2^(i-1) Elemente besitzt.
   Neue Elemente werden immer in die erste Liste eingefuegt.
   Sobald eine Liste zu gross wird, wird sie mit der jeweils
   naechsten Liste durch Merge-Sort verknuepft. Dabei fallen evtl.
   schon einige mehrfach enthaltene Elemente heraus. Dies spart viel
   Speicherplatz, da beim Erzeugen der Simplizes jeder Simplex i.a.
   sehr oft auftritt. (Daher waere es unpraktisch, erst alle
   Simplizes zu erzeugen und dann alle mehrfach aufgetretenen
   auszusortieren.)}
  FM:array[1..MaxI] of TFaceMerge;

  {Simplizes werden mit Merge-Sort sortiert.
   Der Rueckgabewert ist die Anzahl der dabei entfernten
   doppelten Seiten.}
  function Merge(Dest:PPFaceList;Source:PFaceList):TCount;
  var
    t1,t2:PFaceList;
    c:TCount;
  begin
    c:=0;
    while (Source<>nil) and (Dest^<>nil) do
      if FaceEqual(Source^.FaceP^,Dest^^.FaceP^)
        then
          begin
            t1:=Source;
            Source:=Source^.Next;
            Dispose(t1^.FaceP);
            Dispose(t1);
            Inc(c)
          end
        else
          begin
            if FaceLess(Source^.FaceP^,Dest^^.FaceP^) then
              begin
                t1:=Dest^;
                t2:=Source^.Next;
                Dest^:=Source;
                Source^.Next:=t1;
                Source:=t2
              end;
            Dest:=@Dest^^.Next
          end;
    if Source<>nil then Dest^:=Source;
    Merge:=c
  end;

  {Fuegt einen neuen Simplex ein, und verknuepft verschiedene
   Listen, soweit noetig}
  procedure NewFace(const NFace:TFace);
  var
    pn:PFaceList;
    i:1..MaxI;
    c,j:n0;
  begin
    New(pn);
    c:=FaceCount(NFace);
    GetMem(pn^.FaceP,(c+1)*SizeOf(pn^.FaceP^[0]));
    for j:=0 to c do pn^.FaceP^[j]:=NFace[j];
    pn^.Next:=nil;
    Inc(FM[1].Count);
    if FM[1].Count=1
      then FM[1].List:=pn
      else
        if FaceEqual(NFace,FM[1].List^.FaceP^)
          then
            begin
              Dispose(pn^.FaceP);
              Dispose(pn);
              Dec(FM[1].Count)
            end
          else
            begin
              if FaceLess(NFace,FM[1].List^.FaceP^)
                then
                  begin
                    pn^.Next:=FM[1].List;
                    FM[1].List:=pn
                  end
                else FM[1].List^.Next:=pn;
              i:=1;
              while (i<MaxI) and (FM[i].Count>TCount(1) shl(i-1)) do
                begin
                  if FM[i+1].Count=0
                    then FM[i+1]:=FM[i]
                    else Inc(FM[i+1].Count,FM[i].Count-
                           Merge(@FM[i+1].List,FM[i].List));
                  FM[i].Count:=0;
                  FM[i].List:=nil;
                  Inc(i)
                end
            end
  end;

  {Verknuepft (am Ende) alle Listen}
  procedure CollectFaces;
  var
    i,j:1..MaxI+1;
    p,pt:PFaceList;
    c:TCount;
  begin
    i:=1;
    while (i<MaxI) and (FM[i].Count=0) do Inc(i);
    c:=FM[i].Count;
    j:=i;
    while i<MaxI do
      begin
        Inc(i);
        while (i<=MaxI) and (FM[i].Count=0) do Inc(i);
        if i<=MaxI then
          begin
            Inc(c,FM[i].Count-Merge(@FM[i].List,FM[j].List));
            j:=i
          end
      end;
    p:=FM[j].List;
    NewFaces(Faces,c);
    c:=0;
    while p<>nil do
      begin
        FacePGet(Faces,c)^:=p^.FaceP;
        Inc(c);
        pt:=p;
        p:=p^.Next;
        Dispose(pt)
      end
  end;

  procedure DoReport;
  begin
    ProgressReport (Commentator, 'facets', LvlUnimp, c + 1, Facets.Count)
  end;

{Findet zu jeder Facette alle Dim-dimensionalen Teilmengen}
begin
  WriteStr (FacesStr, 'faces of dimension ', Dim);
  StartActivity (Commentator, 'faces', 'Finding ' + FacesStr, LvlNormal, InternalDescription);
  for i:=1 to MaxI do
    with FM[i] do
      begin
        Count:=0;
        List:=nil
      end;
  for j:=0 to MaxN do SubSet[j]:=0;
  for c:=0 to Facets.Count-1 do
    begin
      FaceCopy(FaceGet(Facets,c)^,Facet);
      FacetCard:=FaceCount(Facet);
      if FacetCard>Dim then
        begin
          for j:=0 to Dim do
            begin
              SubSetIndex[j]:=j;
              SubSet[j]:=Facet[j]
            end;
          repeat
            NewFace(SubSet);
            j:=Dim + 1;
            while (j<>0) and (SubSetIndex[j-1]=FacetCard-(Dim + 1)+j-1)
              do Dec(j);
            if j<>0 then
              begin
                Inc(SubSetIndex[j-1]);
                si:=SubSetIndex[j-1];
                SubSet[j-1]:=Facet[si];
                for k:=j to Dim do
                  begin
                    Inc(si);
                    SubSetIndex[k]:=si;
                    SubSet[k]:=Facet[si]
                  end
              end
          until j=0
        end;
      if (MsgStep<>0) and ((c+1) mod MsgStep=0) then DoReport
    end;
  if (MsgStep<>0) and ((c+1) mod MsgStep<>0) then DoReport;
  CollectFaces;
  WriteStr (Msg, 'Found ', Faces.Count, ' ', FacesStr);
  StopActivity (Commentator, Msg, LvlNormal, InternalDescription)
end;

function GetOneRow(const BigFaces,SmallFaces:TFaces;RowCount:TCount):PSparseIntVector;
var
  i,k,s,Temp:TCount;
  Sign:-1..1;
  SmallFace:TFace;
begin
  FaceCopy(FaceGet(BigFaces,BigFaces.Count-1-RowCount)^,SmallFace);
  New(Result,FaceCount(SmallFace));
  Result^.Count:=Result^.Size;
  k:=0;
  Sign:=1;
  for i:=Result^.Count downto 1 do
    begin
      Temp:=SmallFace[i-1];
      SmallFace[i-1]:=k;
      k:=Temp;
      s:=FindFace(SmallFace,SmallFaces);
      if (i<>Result^.Count) and
         (Result^.Elements[Result^.Count-i-1].Index>=s) then
        Abort ('Internal error with vertex ordering.');
      with Result^.Elements[Result^.Count-i] do
        begin
          Index:=s;
          Val:=Sign
        end;
      Sign:=-Sign
    end
end;

procedure ReadComplex (var Facets:TFaces);
type
  {Liste von Facetten, die beim Einlesen angelegt wird.}
  PFacetList=^TFacetList;
  TFacetList=record
    Next:PFacetList;
    Facet:PFace
  end;

  {Liste von Strings.}
  PHashList=^THashList;
  THashList=record
    Next:PHashList;
    Index:TElementCount;
    s:PString
  end;

var
  HashList:array [0..Hash_Size-1] of PHashList;
  MaxIndex:TElementCount;

  procedure InitHash;
  var c:TCount;
  begin
    for c:=0 to Hash_Size-1 do HashList[c]:=nil;
    MaxIndex:=0
  end;

  procedure DoneHash;
  var
    c:TCount;
    ph:PHashList;
  begin
    for c:=0 to Hash_Size-1 do
      while HashList[c]<>nil do
        begin
          ph:=HashList[c];
          HashList[c]:=HashList[c]^.Next;
          Dispose(ph)
        end
  end;

  function GetIndex(const s:TString):TElementCount;
  var
    p:^PHashList;
    i,n,Hash:Integer;
  begin
    Hash:=Length(s);
    n:=10;
    for i:=1 to Length(s) do
      begin
        Hash:=Hash+n*Ord(s[i]);
        n:=(n*27) mod Hash_Size
      end;
    Hash:=Hash mod Hash_Size;
    p:=@HashList[Hash];
    while (p^<>nil) and (s<>p^^.s^) do p:=@p^^.Next;
    if p^=nil then
      begin
        New(p^);
        p^^.Next:=nil;
        New(p^^.s,Length(s));
        p^^.s^:=s;
        Inc(MaxIndex);
        p^^.Index:=MaxIndex
      end;
    GetIndex:=p^^.Index
  end;

  function GetCh = Result : Char;
  begin
    while not EOF and (Input^ in [#0, ' ', #9{TAB},#13{CR},#10{LF}]) do Get (Input);
    if EOF then
      Result := #0
    else
      Read (Result)
  end;

var
  FacetCount,SCount,c:TCount;
  FacetList,pf:PFacetList;
  Facet:TFace;
  i,j,z1,z2:n0;
  t:TElementCount;
  s:TString;
  Ch:Char;
  f:Boolean;

begin
  InitHash;
  FacetList:=nil;
  FacetCount:=0;
  Ch := ' ';
  if GetCh <> '[' then Abort ('Beginning of complex not found.');
  repeat
    Ch:=GetCh;
    if Ch = ']' then Break;
    if Ch <> '[' then Abort ('Beginning of facet not found.');
    i:=0;
    repeat
      s:='';
      SCount:=0;
      repeat
        Ch:=GetCh;
        if Ch = #0 then Abort ('Unexpected end of file!');
        f:=(SCount=0) and ((Ch=',') or (Ch=']'));
        if Ch='[' then Inc(SCount);
        if Ch=']' then Dec(SCount);
        if not f and (Length(s)<MaxLength) then
          begin
            {$X+}SetLength(s,Length(s)+1);{$X-}
            s[Length(s)]:=Ch
          end
      until f;
      if Length(s)=MaxLength then Abort ('String too long ('+s+')!');
      if i>=MaxN then Abort ('Facet too big (increase MaxN)!');
      Facet[i]:=GetIndex(s);
      Inc(i)
    until Ch=']';

    {Sortieren der Punkte mit Bubble sort - auf Effizienz
     kommt es an dieser Stelle nicht an.}
    if i>=2 then
      for z1:=0 to i-2 do
        for z2:=i-1 downto z1+1 do
          if Facet[z1]>Facet[z2] then
            begin
              t:=Facet[z1];
              Facet[z1]:=Facet[z2];
              Facet[z2]:=t
            end;

    {Aussortieren doppelter Punkte}
    z2:=1;
    for z1:=1 to i-1 do
      if Facet[z1]<>Facet[z1-1] then
        begin
          Facet[z2]:=Facet[z1];
          Inc(z2)
        end;
    Facet[z2]:=0;

    pf:=FacetList;
    New(FacetList);
    FacetList^.Next:=pf;
    GetMem(FacetList^.Facet,(z2+1)*Sizeof(FacetList^.Facet^[0]));
    for j:=0 to z2 do FacetList^.Facet^[j]:=Facet[j];
    Inc(FacetCount);
    Ch := GetCh
  until Ch <> ',';
  if Ch <> ']' then Abort ('End of complex not found.');
  DoneHash;

  {Facetten-Liste in ein Array umwandeln}
  if FacetCount=0 then Abort ('No facets given!');
  NewFaces(Facets,FacetCount);
  c:=0;
  while FacetList<>nil do
    begin
      FacePGet(Facets,c)^:=FacetList^.Facet;
      Inc(c);
      pf:=FacetList;
      FacetList:=FacetList^.Next;
      Dispose(pf)
    end
end;

end.
*/
