#ifndef _ARRAY0_H_
#define _ARRAY0_H_
// ========================================================================== //
// Array0.h : base class for one dimensional array of elements of type T.
// (c) copyright GIVARO 1994 
// author: Th. Gautier     
// version : V2.7
// date: 1995
// bugs: unknown
// ========================================================================== //

#include "givtype.h"
#include "giverror.h"
// #include "buffer.h"
#include "givcounter.h"

// Short description: 
// 
// If the dimensional of the two array differs in assigment operator
// we resize the assigned array.
//
// The functions on array are :
//      - constructor
//      - destructor
//      - logcpy        : logical copy
//      - memcpy        : physical copy
//      - operator=     : creation of an array and do assignement of each operator   

template <class T>
class Array0 {
public :
inline Array0 ( ) ;

  // Default cstor : ctsor of s size array
inline Array0 ( unsigned long s)  ;

  // Recopy cstor : logical copy
inline Array0 (const Array0<T>& p)   ;

  // Destructor
inline ~Array0 () ;

  // Allocation of an array of s elements
inline void allocate( const unsigned long s ) ;

  // Reallocation of an array of s elements
inline void reallocate( const unsigned long s ) ;

  // Freeing of the array
inline void freeing( ) ;

  //  Assignment operator : copy the array and do assignment on each element
inline  Array0<T>& operator= (const Array0<T>& p) ;

  // Logical recopy: copy p into this and returns it
inline Array0<T>& logcpy(const Array0<T>& p) ;

  // Physical recopy: same as operator= 
inline Array0<T>& memcpy(const Array0<T>& p) ;

  // Return the occuped size of the array
inline long size() const ;

  // Return the physical size of the array
inline long phsize() const ;

  // Return the base ptr to the array 
inline T* baseptr() ;
inline const T* baseptr() const ;

  // Access to the ith element:
inline const T& operator[] (const int i)  const  ;
inline T& operator[] (const int i)  ;
inline void write( const int i, const T& val)  ;
inline void read ( const int i, T& val ) const ;

  // Map and apply functions :
inline Array0<T>& map( void (*f)(T&) ) ;
inline Array0<T>& map( T (*f)(const T&) ) ;
inline Array0<T>  apply( T (*f)(const T&) ) const ;
inline Array0<T>  apply( void (*f)(T&) ) const ;

protected :  // --------------------- Public Internal representation
  long _size ;      // actual size of the array. If ==0 then _psz =0
  long _psz ;       // physical size of the array
  Counter* _count ; // reference counter
  T*  _d ;          // ptr to the memory
} ;


#include "array0.inl"

#endif
