// ========================================================================== //
// Array0.C : one dimensional array over elements of type T.
// (c) copyright GIVARO
// version: 2.7 
// author: Th. Gautier     
// date: 1995
// ======================================================================= //
// Class Array0<T>

  // Default cstor : ctsor of s size array
template<class T>
inline Array0<T>::Array0 () 
{
  _psz = _size = 0;
  _d = 0; // JGD 050599
}

template<class T>
inline Array0<T>::Array0 ( unsigned long s) 
{
  _psz = _size = s ;
  if (s != 0) 
  {
     _d = new T[s] ;
     _count = new Counter(1) ;
  } else
     _d = 0; // JGD 050599
}

  // Recopy cstor : logical copy
template<class T>
inline Array0<T>::Array0 (const Array0<T>& p) 
{
  _psz = _size = p._size ;
  if (_size !=0)
  {
    _d = p._d ;
    _count = p._count ;
    _count->incr() ;
  } else
    _d = 0; // JGD 050599
}

  // Freeing of the array
template<class T>
inline void Array0<T>::freeing( ) 
{
  if (_size !=0) 
  {
    if (_count->decr() ==0) 
    {
       delete [] _d ;
       delete _count ;
    }
  } else
  	_d = 0; // JGD 050599
}

  // Logical destructor: identical to free
template<class T>
inline Array0<T>::~Array0 ()
{ 
  if (_size !=0) 
  {
    if (_count->decr() ==0) 
    {
       delete [] _d ;
       delete _count ;
    }
  } else
  	_d = 0; // JGD 050599
}

  // Allocation of an array of s elements
template<class T>
inline void Array0<T>::allocate( const unsigned long s ) 
{
  if (_psz >= s) { _size = s ; return ; }
  // Must alloc a new structure
  this->freeing() ;
  _psz = _size = s ;
  if (s != 0) 
  {
     _d = new T[s] ;
     _count = new Counter(1) ;
  } else
  	_d = 0; // JGD 050599
}

  // Reallocation of an array of s elements
  // and recopy the min(_size,s) first elements
template<class T>
inline void Array0<T>::reallocate( const unsigned long s ) 
{
  if (_psz >= s) { _size = s ; return ; }
  if (s !=0)
  {
    T* tmp ;
    // Must alloc a new structure
    tmp = new T[s] ;
    const T* baseThis = _d ;
    long min = (_size < s ? _size : s ) ;
    for (long i=0 ; i<min; i++) tmp[i] = baseThis[i] ;
    _psz = s ;
    if (_size !=0) this->freeing() ; 
    _size = s ;
    _d = tmp ; 
    _count = new Counter(1) ;
  }
  else {
  	 _size = 0 ; // here, may be _psz >0
	 _d = 0; // JGD 050599
	 }
}

// Logical recopy of p in this
template <class T>
inline Array0<T>& Array0<T>::logcpy (const Array0<T>& p)
{
  if (this == &p) return *this ;
  if (_d == p._d ) return *this ;
  this->freeing() ;
  _size = p._size ;
  _psz = p._psz ;
  if (_size !=0)
  {
    _d = p._d ; 
    _count = p._count ;
    _count->incr() ;
    _psz = p._psz ;
  }
  else
  {
    _d = 0 ;
    _count = 0 ;
    _psz = 0 ;
  }
  return *this ;
}

// Physical copy : recopy and assignement on each element
template <class T>
inline Array0<T>& Array0<T>::memcpy (const Array0<T>& p)
{ 
//cerr << "a0 memcpy : " << p._size << " --> " << (*this)._size << endl;
//cerr << "a0 memcpy &  : " << &p << " --> " << this << endl;
  if (this == &p) return *this ;
//cerr << "a0 memcpy &_d : " << &_d << " --> " << &(p._d) << endl;
//cerr << "a0 memcpy _d : " << _d << " --> " << p._d << endl;
  if (_d == p._d ) return *this ;
//cerr << "a0 memcpy real copy " << endl;
  this->allocate( p._size ) ;
  if (_size >0)
  {
    const T* baseP = p._d ;
    T* baseThis = _d ;
    for (long i=0 ; i<_size; i++) baseThis[i] = baseP[i] ;
  }
//cerr << "a0 memcpy : " << (*this)._size << endl;
  return *this ;
}

// Assignement op: like memcopy 
template<class T>
inline Array0<T>& Array0<T>::operator=(const Array0<T>& p) 
{
   return memcpy( p ) ;
}

template<class T>
inline long Array0<T>::size() const 
{
  return _size ;
}

template<class T>
inline long Array0<T>::phsize() const 
{
  return _psz ;
}

template<class T>
inline T* Array0<T>::baseptr() 
{
  return _d ;
}

template<class T>
inline const T* Array0<T>::baseptr() const 
{
  return _d ;
}

  // This foloowing functions directly access to protected
  // member of BaseArray class
template <class T>
inline const T& Array0<T>::operator[] (const int i) const
{
//#ifdef GIVARO_DEBUG
  if (_size ==0)
    throw GivError("*** try to access to an element of null size Array0") ;
  if ((i <0) || (i>=_size)) 
    throw GivError("*** index out of bounds in Array0") ;
//#endif
  return _d[i] ;
}

// Access operator : Write access
template <class T>
inline T& Array0<T>::operator[] (const int i)
{
//#ifdef GIVARO_DEBUG
  if (_size ==0)
    throw GivError("*** try to access to an element of null size Array0") ;
  if ((i <0) || (i>=_size)) 
    throw GivError("*** index out of bounds in Array0") ;
//#endif
  return _d[i] ;
}

template <class T>
inline void Array0<T>::write( const int i, const T& val)
{
   _d[i] = val ;
}

template <class T>
inline void Array0<T>::read ( const int i, T& val ) const
{
   val = _d[i] ;
}


// Map functions:
template<class T>
Array0<T>& Array0<T>::map( void (*f)(T&) )
{
   if (_size ==0) return *this ;
   if (_count->val() >0)
   {
      Array0<T> Tmp ;
      Tmp.memcpy(*this) ;
      this->logcpy(Tmp) ; 
   }
   for (int i=0; i < _size ; i++) (*f) ( _d[i]) ;
   return *this ;
}


template<class T>
Array0<T>& Array0<T>::map( T (*f)(const T&) )
{
   if (_size ==0) return *this ;
   if (_count->val() >0)
   {
      Array0<T> Tmp ;
      Tmp.memcpy(*this) ;
      this->logcpy(Tmp) ; 
   }
   for (int i=0; i < _size ; i++) _d[i] = (*f) ( _d[i]) ;
   return *this ;
}


// Apply
template<class T>
Array0<T> Array0<T>::apply( T (*f)(const T&) ) const
{
   Array0<T> Tmp ;
   Tmp.memcpy(*this) ; // dump de this
   return Tmp.map(f) ;
}

template<class T>
Array0<T> Array0<T>::apply( void (*f)(T&) ) const
{
   Array0<T> Tmp ;
   Tmp.memcpy(*this) ; // dump de this
   return Tmp.map(f) ;
}
