#ifndef _ARRAY1_H_
#define _ARRAY1_H_
// ========================================================================== //
// Array1.h : User class to one dimensional array of elements of type T.
// (c) copyright GIVARO 1994 
// author: Th. Gautier     
// date: 1994                
// bugs: unknown
// ========================================================================== //
#include "array0.h"

// 
// The functions are :
//      - constructor
//      - destructor
//      - size()	: returns the numbers of elements of an array
//      - length()	: returns the size in bytes of an array
//      - operator=     : logical copy
//      - operator==    : equality test
//      - operator!=    : inequality test 
//      - map           : application d'une fonction  sur les elements
//      - apply         : construction  d'un nouveau tableau apres
//                        application d'une fonction  sur les elements

#ifdef __GNUC__
template <class T>
class Array1;

template <class T>
inline ulong length(const Array1<T>& A)  ;
#endif


template <class T>
class Array1 : public Array0<T> {

protected:
  // Cstor from Array0: logical copy
inline  Array1 (const Array0<T>& p)   ;

public :

  // Default cstor : ctsor of s size array
inline  Array1 (const unsigned long  s = 0)  ;

  // Recopy cstor : logical copy
inline  Array1 (const Array1<T>& p)   ;

  // Destructor
//inline  ~Array1 () ;

  //  Assignment operator : assignment of each element of p to a new array this
inline  Array1<T>& operator= (const Array1<T>& p) ;

  // Logical recopy: logical copy of p into this and returns it.
inline  Array1<T>& logcpy(const Array1<T>& p) ;

  // Physical recopy: same implementation as operator=
inline  Array1<T>& memcpy(const Array1<T>& p) ;

  // Returns the nb of elements and the size int bytes of an array  
#ifdef __GNUC__
friend ulong length<>(const Array1<T>& A)  ;
#else
friend ulong length(const Array1<T>& A)  ;
#endif

  // Comparaizon operator
inline int operator==(const Array1<T>&p) const ;
inline int operator!=(const Array1<T>&p) const ;

  // Map and Apply functions
inline Array1<T>& map( void (*f)(T&) ) ;
inline Array1<T>& map( T (*f)(const T&) ) ;
inline Array1<T>  apply( T (*f)(const T&) ) const ;
inline Array1<T>  apply( void (*f)(T&) ) const ;

} ;


#include "array1.inl"

#endif
