// ========================================================================== //
// Array1.C : one dimensional array over elements of type T.
// (c) copyright GIVARO 1994 
// author: Th. Gautier     
// date: 1995                
// 
// ======================================================================= //

  // Cstor from Array0: logical copy
template<class T>
inline  Array1<T>::Array1(const Array0<T>& p) : Array0<T>(p) 
{ } 


// Default cstor : ctsor of s size array
template<class T>
inline Array1<T>::Array1 (const unsigned long s) : Array0<T>(s)
{ }


// Recopy cstor : logical copy
template<class T>
inline Array1<T>::Array1 (const Array1<T>& p) : Array0<T>(p) 
{ }

// Logical destructor:
//template<class T>
//inline Array1<T>::~Array1 ()
//{}

// Logical recopy of p in this
template <class T>
inline Array1<T>& Array1<T>::logcpy (const Array1<T>& p)
{
   Array0<T>::logcpy (p) ;
   return *this ;
}

// Assignement operator: physical
template <class T>
inline Array1<T>& Array1<T>::operator= (const Array1<T>& p)
{
  Array0<T>::memcpy (p) ;
  return *this ;
}


// Physical copy : by recursive assignement of each element
template<class T>
inline Array1<T>& Array1<T>::memcpy(const Array1<T>& p) 
{
  Array0<T>::memcpy (p) ;
  return *this ;
}


// Comparaisons
template<class T>
inline int Array1<T>::operator==(const Array1<T>& p) const
{
  if (this == &p) return 1 ;
  if (size() != p.size()) return 0 ;

  int i, egal = 1 ;
  const T *z1 = baseptr(), *z2 = p.baseptr() ; 
  int s = size() ;
  for (i=0; (i < s) && egal ; i++)
     egal = egal && (z1[i] == z2[i]) ;
  return egal ;
}

template<class T>
inline int Array1<T>::operator!=(const Array1<T>& p) const
{ 
  return ! operator==(p) ;
}


// Returns the size in bytes of the array A
template<class T>
inline unsigned long length( const Array1<T>& A) 
{
  int sz = A.size() ;
  if (sz ==0) return 0 ;
  unsigned long  sum = 0 ;
  const T* z1 = A.baseptr() ;
  for (int i=0 ; i < sz ; i++) sum += length(z1[i]) ;
  return sum ;
}

template<class T>
Array1<T>& Array1<T>::map( void (*f)(T&) ) 
{ Array0<T>::map(f) ; return *this ; }

template<class T>
Array1<T>& Array1<T>::map( T (*f)(const T&) ) 
{ Array0<T>::map(f) ; return *this ; }

template<class T>
Array1<T>  Array1<T>::apply( T (*f)(const T&) ) const 
{ return Array0<T>::apply(f) ; }

template<class T>
Array1<T>  Array1<T>::apply( void (*f)(T&) ) const 
{ return Array0<T>::apply(f) ; }

