#include "integer.h"
#include "compgmp.h"
extern "C" 
{ 
#include "gmp.h"
}


#ifndef GIVABS
#define GIVABS(l) ((l) <0 ? -l : l)
#endif

// returns 1 if a > b, 0 if a == b and -1 otherwise.  
int compare(const Integer &a, const Integer& b) 
{
  if (&a == &b) return 0 ;

/* Old version of Gmp
  if (a.size > b.size) return 1;
  if (a.size < b.size) return -1;

  if ((a.size == 0) && (b.size == 0)) return 0 ;
  if (a.size == 0)   return (b.size <0) ? 1 : -1 ;
  if (b.size == 0) return (a.size <0) ? -1 : 1 ;
*/

  int cmp = mpz_cmp ( (mpz_ptr)&a.gmp_rep, (mpz_ptr)&b.gmp_rep );
  return (cmp <0 ? -1 : (cmp >0 ? 1 : 0));
/*
  return (a.size <0 ? -cmp : cmp ) ;
*/
}

// May be we can do that with mpz_ function ?
int absCompare(const Integer &a, const Integer &b) 
{
  if (&a == &b) return 0 ;
  int u_as = GIVABS(a.gmp_rep.size) ;
  int u_bs = GIVABS(b.gmp_rep.size) ;
  if (u_as > u_bs) return 1;
  else if (u_as < u_bs) return -1;
  else if ((u_as == 0) && (u_bs == 0)) return 0 ;
  else if (u_as == 0) return -1 ;
  else if (u_bs == 0) return 1 ;
  int cmp = mpn_cmp (a.gmp_rep.d, b.gmp_rep.d, u_as);
  return (cmp <0 ? -1 : (cmp >0 ? 1 : 0));
}

int Integer::operator != (const int l) const 
{ return *this != Integer(l) ; }

int Integer::operator != (const long l) const 
{ return *this != Integer(l) ; }

int Integer::operator > (const int l) const 
{ return *this > Integer(l) ; }

int Integer::operator > (const long l) const 
{ return *this > Integer(l) ; }

int Integer::operator < (const int l) const 
{ return *this < Integer(l) ; }

int Integer::operator < (const long l) const 
{ return *this < Integer(l) ; }


