#include "integer.h"
#include "compgmp.h"
extern "C" {
#include "gmp.h"
}

#include <math.h> //for ceil
#include <strstream.h>

//------------------------------------- predefined null and one
const Integer Integer::Zero (0) ;
const Integer Integer::One (1) ;


// Cstor and Dstor

//-----------------------------~Integer()
Integer::~Integer() 
{
  if (cpt->decr() == 0) { mpz_clear((mpz_ptr)&gmp_rep) ; delete cpt ;}
}


//-----------------------------Integer()
Integer::Integer() 
{
  mpz_init_set_ui((mpz_ptr)&gmp_rep, 0L) ;
  cpt = new Counter(1) ;
}


//-----------------------------Integer(int n)
Integer::Integer(int n) 
{
  mpz_init_set_si((mpz_ptr)&gmp_rep, n) ;
  cpt = new Counter(1) ;
}

//-----------------------------Integer(ulong n)
Integer::Integer(ulong n) 
{
  mpz_init_set_ui((mpz_ptr)&gmp_rep, n) ;
  cpt = new Counter(1) ;
}

//-----------------------------Integer(long n)
Integer::Integer(long n) 
{
  mpz_init_set_si((mpz_ptr)&gmp_rep, n) ;
  cpt = new Counter(1) ;
}

//-----------------------------Integer(const type neutral n)
Integer::Integer(const Neutral n) 
{
  if (n == zero) mpz_init_set_ui((mpz_ptr)&gmp_rep, 0L) ;
  else  mpz_init_set_ui((mpz_ptr)&gmp_rep, 1L) ;
  cpt = new Counter(1) ;
}


//------------------------------------Integer(const char *s)
Integer::Integer(const char *s) 
{
  mpz_init_set_str((mpz_ptr)&gmp_rep, s, 10) ;
  cpt = new Counter(1) ;
}

// ------------------------------------Int2double ()
const double Integer::Int2double() const {
	return  mpz_get_d ( (mpz_ptr)&gmp_rep ) ;
}

// ------------------------------------Int2long ()
const long Integer::Int2long() const {
	return  mpz_get_si ( (mpz_ptr)&gmp_rep ) ;
}

// ------------------------------------Int2str ()
char * Integer::Int2str() const {
  //!!!
  int base = 10;
  size_t strSize = mpz_sizeinbase((mpz_ptr)&(gmp_rep), base)+2;
  char *str = ::new char[strSize];
  mpz_get_str(str, base, (mpz_ptr)&(gmp_rep));
  return str;
} 

//-------------------------------Integer(const Integer &n)
Integer::Integer(const Integer &n) 
{ 
   gmp_rep = n.gmp_rep ;
   cpt = n.cpt ;
   cpt->incr() ;
}


//------------------------------------------operator = (const Integer &n)
Integer& Integer::logcpy(const Integer &n)
{
  //don't try to copy n to itself !
  if (this == &n) return *this ;
  if (cpt->decr() ==0) { mpz_clear((mpz_ptr)&gmp_rep) ; delete cpt ;}
  gmp_rep = n.gmp_rep ;
  cpt = n.cpt ;
  cpt->incr() ;
  return *this ;
}

Integer& Integer::operator = (const Integer &n)
{ return logcpy(n) ; }


Integer& Integer::memcpy(const Integer &n)
{
  //don't try to copy n to itself !
  if (this == &n) return *this ;
  if (cpt->decr() ==0) { mpz_clear((mpz_ptr)&gmp_rep) ; delete cpt ;}
  mpz_init_set((mpz_ptr)&gmp_rep, (mpz_ptr)&(n.gmp_rep)) ;
  cpt = new Counter(1) ;
  return *this ;
}
