#include "integer.h"
#include "giverror.h"
#include "compgmp.h"
extern "C" 
{ 
#include "gmp.h" 
}



// Returns the quotient of *this / n 
const Integer Integer::operator / (const Integer &n) const
{
  if (iszero(n)) {
     throw GivMathDivZero("*** division by zero, in operator / (const Integer&)") ;
     return n ;
  }
  if (iszero(*this)) return *this ;

  if (absCompare(*this,n) == -1) return Integer(0L) ;

  Integer Res ; 
  mpz_div( (mpz_ptr)&(Res.gmp_rep), (mpz_ptr)&gmp_rep, (mpz_ptr)&(n.gmp_rep)) ;
  return Res ;
}
  

//
const Integer Integer::operator / (const ulong l) const 
{
  Integer Res ; 
  mpz_div_ui( (mpz_ptr)&(Res.gmp_rep), (mpz_ptr)&gmp_rep, l ) ;
  return Res ;
}

const Integer Integer::operator / (const int l) const
{
  ulong ul = ABS(l) ;
  Integer Res ; 
  mpz_div_ui( (mpz_ptr)&(Res.gmp_rep), (mpz_ptr)&gmp_rep, ul ) ;
  if (sign(l) <0) return -Res ;
  return Res ;
}

const Integer Integer::operator / (const long l) const
{ 
  ulong ul = ABS(l) ;
  Integer Res ; 
  mpz_div_ui( (mpz_ptr)&(Res.gmp_rep), (mpz_ptr)&gmp_rep, ul ) ;
  if (sign(l) <0) return -Res ;
  return Res ;
}


// Returns q and r such that a=b*q + r, with |r| < |b|
void divide(const Integer &a, const Integer &b, Integer &q, Integer &r)
{
  if (iszero(b)) {
    throw GivMathDivZero("*** division by zero, in 'divide' (Integer)") ;
    return ;
  }
  if ( absCompare(a,b) == -1 ) {
// JGD 03.06.98 : et non pas dans l'ordre inverse !!!
    r = a ;
    q = Integer(zero) ;
    return ;
  }

  mpz_divmod( (mpz_ptr)&(q.gmp_rep), (mpz_ptr)&(r.gmp_rep), 
              (mpz_ptr)&(a.gmp_rep), (mpz_ptr)&(b.gmp_rep)) ;
  return ;
}


// Returns q such that a=b*q, so the division must be exact
const Integer divexact(const Integer &a, const Integer &b)
{
  if (iszero(b)) {
    throw GivMathDivZero("*** division by zero, in 'divide' (Integer)") ;
    return b;
  }
  if ( absCompare(a,b) == -1 ) {
    return Integer(zero) ;
  }

  Integer q;
  mpz_divexact( (mpz_ptr)&(q.gmp_rep),
              (mpz_ptr)&(a.gmp_rep), (mpz_ptr)&(b.gmp_rep)) ;
  return q;
}

