#ifndef _INTEGER_H_
#define _INTEGER_H_
// ========================================================================== //
// Defintion of class Integer
// (c) copyright GIVARO 1994
// author: M.Samama, T.Gautier
// date: 1994/1996
// Gmp version : 2.0
// ========================================================================== //

#include "givtype.h"
#include "givcounter.h"

// forward decl
//class ostream ; class istream ; 
#include <iostream>

//========================   CLASS "Integer"  ================================//

class Integer {

public:

//--------------------------------------cstors & dstors
  Integer();
  Integer(int n);
  Integer(long n);
  // Creates a new Integer from an unsigned long 
  Integer(ulong n);
  Integer(const char *s);
  Integer(const Integer& n);
  Integer(const Neutral n) ;
  ~Integer();

//------------------------------------- predefined null and one
static const Integer Zero ;
static const Integer One ;

  Integer& operator = (const Integer& n);
  Integer& memcpy(const Integer& n);
  Integer& logcpy(const Integer& n);
  
//------------------Equalities and inequalities between integers and longs
  int operator != (const int l) const ;
  int operator != (const long l) const ;
  
  friend int compare(const Integer& a, const Integer& b) ;
  friend int absCompare(const Integer& a, const Integer& b) ;

  int operator > (const int l) const ;
  int operator > (const long l) const ;
  int operator < (const int l) const ;
  int operator < (const long l) const ;

//----------------Elementary arithmetic between Integers & longs
  const Integer operator + (const Integer& n) const;  
  const Integer operator + (const int l) const;
  const Integer operator + (const long l) const;

  const Integer operator - (const Integer& n) const;
  const Integer operator - (const int l) const;
  const Integer operator - (const long l) const;
  const Integer operator -() const;

  const Integer operator * (const Integer& n) const;
  const Integer operator * (const int l) const;
  const Integer operator * (const long l) const;


  const Integer operator / (const Integer& n) const;
  const Integer operator / (const int l) const ;
  const Integer operator / (const long l) const ;
  const Integer operator / (const ulong l) const ;

  const Integer operator % (const Integer& n) const;
  const int     operator % (const int l) const;
  const long    operator % (const long l) const;
  const ulong   operator % (const ulong l) const;
  const double  Int2double () const;
  const long  Int2long () const;
  char * Int2str () const;
  
  //-----------------------------------------Arithmetic functions
  friend const Integer gcd (const Integer& a, const Integer& b) ;
  friend const Integer gcd (const Integer& a, const Integer& b, 
                            Integer& u, Integer& v) ;

//   friend const Integer powmod(const Integer&, const Integer&, const Integer&);
  friend const Integer pow(const Integer& n, const long l);
  friend const Integer pow(const Integer& n, const int l);
  friend void divide(const Integer& num, const Integer &den, 
                     Integer &q, Integer &r);
  friend const Integer divexact(const Integer& num, const Integer &den); 
  friend Integer sqrt(const Integer& p);
  friend Integer sqrt(const Integer& p, Integer& r);


  friend const Integer fact ( ulong l) ;
  
  //-----------------------------------------Miscellaneous
  friend int odd(const Integer& n);
  friend int even(const Integer& n);
  friend int probab_prime(const Integer& p);
  friend int probab_prime(const Integer& p, int r);
  friend int jacobi(const Integer& u, const Integer& v) ;
  friend int legendre(const Integer& u, const Integer& v) ;

  friend const Integer abs(const Integer& n);

  const Integer operator << (unsigned int l) const; // lshift
  const Integer operator >> (unsigned int l) const; // rshift
  const Integer operator << (unsigned long l) const; // lshift
  const Integer operator >> (unsigned long l) const; // rshift

private:
  int priv_sign() const ;

public:
  friend inline ulong length (const Integer& a) ; 
  friend inline int sign   (const Integer& a) ;
  friend inline int iszero (const Integer& a) ;
  friend inline int isone  (const Integer& a) ;


  //----------------------------------------------I/O

  friend std::istream &operator >> (std::istream &i, Integer& n);
  friend std::ostream &operator << (std::ostream &o, const Integer& n);
  friend std::ostream& absOutput(std::ostream &o, const Integer&n) ;

public:
  //---------------------------------------------- Internal representation
  // this struct must match the MP_INT struct
  struct {
    long int alloc ;
    long int size ;
    ulong* d ;
  }  gmp_rep ;
protected:
  Counter* cpt ; // associated to d

  // Creates a new Integer from a size sz and a array of ulong d 
  Integer(ulong* d, long size);

protected:
  //---------------------------------------------- Annex functions
//static const Integer LehmerGcd(const Integer& a, const Integer& b) ;
static const Integer GMPGcd(const Integer& a, const Integer& b) ;
private:
};//----------------------------------------------- End of Class Integer

// -- return the prime part of Q in P:
extern const Integer pp( const Integer& P, const Integer& Q ) ;

#include "integer.inl"

#endif

