// =================================================================== //
// Givaro : Prime numbers
//              Factors,
// Needs list structures : stl ones for instance
// Time-stamp: <01 Jun 99 14:54:53 Jean-Guillaume.Dumas@imag.fr> 
// =================================================================== //

#ifndef _GIVARO_FACTORISATION_
#define _GIVARO_FACTORISATION_

#include <list>
using std::list;
#include "integer.h"
#include "intprime.h"
#include "intfactor.h"

// =================================================================== //
// Set or list of divisors, factors.
// =================================================================== //
unsigned long factorset(const Integer& n, list<Integer>& Lf, list<unsigned long>& Lo) {
        // n = * Lf[i] ^ Lo[i]
        // But Lf[i] might not be prime (cf. Pollard probability)
    
    Integer nn,g,r,u;
    nn = ABS(n);
    unsigned long c,nb=0;
    while(nn > 1) {        
        g=Integer_factor(nn);
        Lf.push_back(g);
        c=0;r=0;
	u = divexact(nn,g);
        while(r == 0) {
        //	nn = nn / g;
        //	r = nn % g;
		nn.memcpy(u);
		divide(nn,g,u,r);
            	c++;
        }
        Lo.push_back( c );
        nb++;
    }
    return nb;
}

void factorset(const Integer& n, list<Integer>& Lf, list<long>& Lo) {
        // n = * Lf[i] ^ Lo[i]
        // But Lf[i] might not be prime (cf. Pollard probability)
    Integer nn,g,r,u;
    nn = n;
    long c;
    while(nn > 1) {        
        g=Integer_factor(nn);
        Lf.push_back(g);
        c=0;r=0;
	u = divexact(nn,g);
        while(r == 0) {
        //	nn = nn / g;
        //	r = nn % g;
		nn.memcpy(u);
		divide(nn,g,u,r);
            c++;
        }
        Lo.push_back( c );
    }
}           

void factorset_Erathostene(const Integer& p, list<Integer>& Lf) {
        // Deterministic algorithm
        // Valid for p < BOUNDARY_factor
        // Lf is the list of factors
    int n = p%int(BOUNDARY_factor);
    if (! (n & 0x1)) {
        Lf.push_back(Integer(2));
        do
            n >>= 1;
        while (!(n & 0x1));
    }
    int * IP = new int[n+1];
    int i;
    for(int i=n+1;i--;)
	IP[i] = 0L;
    i=3;
    int j, ii, cofact;
    while (i<=sqrt(n)) {
        ii= i << 1;
        j = i+ii;
        while (j<=n) {
            IP[j] = 1L;
            j+=ii;
        }
        if ((j-ii) == n) {
            Lf.push_back(Integer(i));
            do
                n /= i;
            while (!(n%i));
        }
        j = i+1;
        while (IP[++j]) { j++;}
        i = j;
    }
    if (!(IP[n]) && (n>1)) Lf.push_back(Integer(n));
}
 

void factorset(const Integer& n, list<Integer>& Lf) {
        // big_factor is executed until
        // a (sometimes probably) prime factor is found.
        // Lf is a list of divisors
        // Lf is a list of factors with probability -- TO EXPLICIT
        // something like : 1 - (big_factor to be composite)*(big_isprime)
    Integer nn,g,r,u;
    nn = n;
    while(nn > 1) {
        g=Integer_factor(nn);
        r=0;
	u = divexact(nn,g);
        while(r == 0) {
        //	nn = nn / g;
        //	r = nn % g;
		nn.memcpy(u);
		divide(nn,g,u,r);
        }
        if (isprime(g))
            Lf.push_back(g);
        else
            Integer_factorset(g, Lf);
    }
}

void divisors(const Integer& n, list<Integer>& L) {
    list<Integer> Lf;
    list<ulong> Le;
    factorset(n,Lf,Le);
    list<Integer>::iterator li = Lf.begin();
    list<ulong>::iterator lj = Le.begin();
    list<Integer> Res(1,Integer(1));
    list<Integer> Res2;
    list<Integer>::iterator lr;
    Integer Itmp;
    for(;li!=Lf.end();++li,++lj) {
        for(lr = Res.begin();lr!=Res.end();++lr) {
		Itmp = *lr;
      		for(ulong i=*lj;i--;) {
			Itmp = Itmp * *li;
            		Res2.push_back( Itmp );
		}
	}
        Res.splice(Res.end(),Res2);
    }
    L = Res;
}



//cerr << "a nn: " << nn << "  u: " << u << "  g: " << g << "  r: " << r << endl;
#endif
