#include "integer.h"
#include "compgmp.h"
extern "C" {
#include "gmp.h"
}




// ==========================================================================
// Computes and returns the gcd of the two integers a and b.
// The algorithm used is this of Gmp.
const Integer  Integer::GMPGcd(const Integer& a, const Integer& b)
                                  //Integer& u, Integer& v)
{
  Integer Res ;
  mpz_gcd( (mpz_ptr)&(Res.gmp_rep), (mpz_ptr)&(a.gmp_rep), (mpz_ptr)&(b.gmp_rep) ) ;
  if (sign(Res) <0) return -Res ;
  return Res ;
}




// ==========================================================================
// Computes and returns the gcd of the two integers a and b.
// Switch to different methods along the size of the input arguments
const Integer  gcd(const Integer& a, const Integer& b)
{
   if (&a == &b) return a ;
   return Integer::GMPGcd(a,b) ;
}


// ==========================================================================
// Computes and returns the gcd g of the two integers a and b such that
// g = a*u + b*v .
// The algorithm used is this of Gmp.
const Integer  gcd (const Integer& a, const Integer& b, Integer& u, Integer& v)
{
  Integer Res ;
  mpz_gcdext( (mpz_ptr)&(Res.gmp_rep), (mpz_ptr)&(u.gmp_rep), (mpz_ptr)&(v.gmp_rep), 
              (mpz_ptr)&(a.gmp_rep), (mpz_ptr)&(b.gmp_rep) ) ;
  if (sign(Res) <0) { Res = -Res ; u = -u ; v = -v ; }
  return Res ;
}

const Integer pp( const Integer& P, const Integer& Q ) 
{
   Integer U = P ; 
   Integer V = gcd(P,Q) ; 
   // -- computes the prime part U of g relatively to U
   while ( V != Integer::One ) 
   {
      U = U / V ;
      V = gcd( U,V) ;
   } 
   return U ;
} 

