#ifndef _GIVARO_INTEGERS_MISC_
#define _GIVARO_INTEGERS_MISC_




#include "integer.h"
#include "compgmp.h"
extern "C" {
#include "gmp.h"
}



//------------------------------------------GMP isprime
//     If this function returns 0, OP is definitely not prime.  If it
//     returns 1, then OP is `probably' prime.  The probability of a
//     false positive is (1/4)^r.  A reasonable value of r is 25.

int probab_prime(const Integer &p)
{
  return mpz_probab_prime_p ((mpz_ptr)&(p.gmp_rep),1) ;
}

int probab_prime(const Integer &p, int r)
{
  return mpz_probab_prime_p ((mpz_ptr)&(p.gmp_rep),r) ;
}

// ==========================================================================
// Computes and returns the Jacobi and Legendre symbols (u/v) of the integers u and v.
// The algorithm used is Gmp's.
int jacobi(const Integer& u, const Integer& v)
{
  return mpz_jacobi ((mpz_ptr)&(u.gmp_rep),(mpz_ptr)&(v.gmp_rep)) ;
}

int legendre(const Integer& u, const Integer& v)
{
  return mpz_legendre ((mpz_ptr)&(u.gmp_rep),(mpz_ptr)&(v.gmp_rep)) ;
}


//------------------------------------------odd (const Integer &n)
int odd(const Integer &n)
{
  if (iszero(n)) return 1;
  return ( (n.gmp_rep.d[0] & 0x1) == 0 );
}

//------------------------------------------even (const Integer &n)
int even(const Integer &n)
{
  if (iszero(n)) return 1;
  return ( (n.gmp_rep.d[0] & 0x1) != 0 );
}

/*
//floor( log_2(n) ).
//may be 1 too small if n = 2^k
//-------------------------------------------lg
int log2(const Integer &n)
{
  if (n.size == 0) return 0;
  int nbit = 31;
  long sz = (size < 0 ? -n.size : n.size) ;
  mp_limb Hi = n.d[ sz - 1 ];
  
  while (! (Hi & (1 << 31)) ) Hi <<= 1, --nbit;
  return 32 * (n.size - 1) + nbit;
}
*/


//-------------------------------------------abs (const Integer &n)
const Integer abs(const Integer &n)
{
  if (sign(n) >= 0) return n;
  return -n ;
}

//-------------------------------------------fact (ulong l)
const Integer fact ( ulong l) 
{
  Integer Res ;
  mpz_fac_ui( (mpz_ptr)&(Res.gmp_rep), l ) ;
  return Res ;
}



//-------------------------------------------square root
Integer sqrt(const Integer &a)
{
  Integer q;
  mpz_sqrt( (mpz_ptr)&(q.gmp_rep),
              (mpz_ptr)&(a.gmp_rep)) ;
  return q;
}

Integer sqrt(const Integer &a, Integer& r)
{
  Integer q;
  mpz_sqrtrem( (mpz_ptr)&(q.gmp_rep),
              (mpz_ptr)&(r.gmp_rep), (mpz_ptr)&(a.gmp_rep)) ;
  return q;
}


//--------------------------------------------Integer::operator <<   // shift left
// N O T   I M P L E M E N T E D 
const Integer Integer::operator << (unsigned int l) const 
{ return *this; }

const Integer Integer::operator << (unsigned long l) const 
{ return *this; }


//--------------------------------------------Integer::operator >>   // shift right
// N O T   I M P L E M E N T E D 
const Integer Integer::operator >> (unsigned int l) const
{ return *this; }

const Integer Integer::operator >> (unsigned long l) const
{ return *this; }

#endif
