#include "integer.h"
#include "giverror.h"
#include "compgmp.h"
extern "C" {
#include "gmp.h"
}

 

  // Returns *this % l. The sign of the result is the same as *this.
const Integer Integer::operator % (const Integer &n) const
{
  if (iszero(n)) {
     throw GivMathDivZero("*** argument is null in operator% on Integer") ;
     return n;
  }

  Integer Res ;
  mpz_mod( (mpz_ptr)&(Res.gmp_rep), (mpz_ptr)&gmp_rep, (mpz_ptr)&(n.gmp_rep) );
  // Correction in order to set the sign of Res to those of this (Gmp 1.906)
  int signThis = sign(*this) ;
  if ((signThis == -1) && (!iszero(Res))) return Res - n ;
  return Res ;
}
  

  // Returns *this % l. The sign of the result is the same as *this.
const int Integer::operator % (const int l) const
{
  if (l == 0)
    {
      throw GivMathDivZero("*** argument is null in operator% on Integer") ;
      return l ;
    }
  ulong ll = ABS(l) ;
  Integer Res ;
  mpz_mod_ui( (mpz_ptr)&(Res.gmp_rep), (mpz_ptr)&gmp_rep, ll);
  // Correction in order to set the sign of Res to those of this (Gmp 1.906)
  int signThis = sign(*this) ;
  if ((signThis ==-1) && (!iszero(Res))) return Res.gmp_rep.d[0] - ll ;
  return Res.gmp_rep.d[0] ; 
}

  // Returns *this % l. The sign of the result is the same as *this.
const long Integer::operator % (const long l) const
{
  if (l == 0)
    {
      throw GivMathDivZero("*** argument is null in operator% on Integer") ;
      return l ;
    }
  ulong ll = ABS(l) ;
  Integer Res ;
  mpz_mod_ui( (mpz_ptr)&(Res.gmp_rep), (mpz_ptr)&gmp_rep, ll);
  // Correction in order to set the sign of Res to those of this (Gmp 1.906)
  int signThis = sign(*this) ;
  if ((signThis ==-1) && (!iszero(Res))) return Res.gmp_rep.d[0] - ll ;
  return Res.gmp_rep.d[0] ; 
}
  
  // Returns *this % l. The sign of the result is the same as *this.
const ulong Integer::operator % (const ulong l) const
{
  if (l == 0)
    {
      throw GivMathDivZero("*** argument is null in operator% on Integer") ;
      return l ;
    }
  Integer Res ;
  mpz_mod_ui( (mpz_ptr)&(Res.gmp_rep), (mpz_ptr)&gmp_rep, l);
  // Correction in order to set the sign of Res to those of this (Gmp 1.906)
  int signThis = sign(*this) ;
  if ((signThis ==-1) && (!iszero(Res))) return Res.gmp_rep.d[0] - l ;
  return Res.gmp_rep.d[0] ; 
}
  

