// =================================================================== //
// Givaro : Euler's phi function
//          Primitive roots.
// Needs list structures : stl ones for instance
// Time-stamp: <28 May 99 18:44:28 Jean-Guillaume.Dumas@imag.fr> 
// =================================================================== //

#ifndef _GIVARO_NUMTHEORY_
#define _GIVARO_NUMTHEORY_

#include <list>
using std::list;
#include "integer.h"
#include "intprime.h"
#include "intfactor.h"

// =================================================================== //
// Givaro : Theoreme Chinois des Restes
// =================================================================== //
const Integer tcr(list<Integer>& Lr, list<Integer>& Lm);
const Integer& tcr(Integer& res, Integer& pm, list<Integer>& Lr, list<Integer>& Lm);

// =================================================================== //
// Euler's phi function
// =================================================================== //
const Integer phi(const Integer& n);

// =================================================================== //
// Primitive Root
// =================================================================== //
const Integer prim_root(const Integer& n) ;
const Integer lowest_prim_root(const Integer& n) ;
int is_prim_root(const Integer& p, const Integer& n) ;
const Integer order(const Integer& p, const Integer& n) ;


// =================================================================== //
// RSA public-key cipher codes
// =================================================================== //

// =================================================================== //
// log[10]
// =================================================================== //
long log(const Integer& n);

// =================================================================== //
// Text conversions
// =================================================================== // 
void ecriture_str(const Integer& n);
void ecriture_Int(const Integer& p, const Integer& size);
void encipher(const Integer& m, const Integer& k);
void decipher(const Integer& m, const Integer& u);

// =================================================================== //
// Keys generation
// public keys are m and k, the secret key is u.
// ciphering is computing       : x^k mod m
// deciphering is computing     : b^u mod m
// since for any x, x^(k.u) = x mod m
// =================================================================== //

// =================================================================== //
// Here m = p*q
// p and q are prime numbers of respective sizes psize, qsize
// Moreover p-1 and q-1 have one prime factor of respective size 2/3
// since k.u = 1 mod (p-1)(q-1)
// =================================================================== //
void keys_gen(long psize, long qsize, Integer& m, Integer& k, Integer& u);

// =================================================================== //
// Breaking codes : finding u knowing only m an k ...
// =================================================================== //
Integer& point_break(Integer& u, const Integer& m, const Integer& k);

#endif
