// =================================================================== //
// Givaro : Prime numbers
//              Modular powering,
//              Fermat numbers,
//              Primality tests, Factorization :
//                      (There are parameters to fix)
// Time-stamp: <21 Aug 99 14:38:55 Jean-Guillaume.Dumas@imag.fr> 
// =================================================================== //
#ifndef _GIVARO_INTEGERS_PRIME_H_
#define _GIVARO_INTEGERS_PRIME_H_

#include <math.h>
#include "timer.h"
#include "integer.h"
#include "compgmp.h"


// =================================================================== //
// Modular powering
// =================================================================== //
const Integer powmod( const Integer& ti, const Integer l, const Integer& fn) ;
// =================================================================== //
// Fermat numbers
// =================================================================== //
const Integer fermat ( const long  n ) ;
int pepin (const long n);
// =================================================================== //
// Random generator
// =================================================================== //
const Integer Random_Integer (Integer& r, const Integer& size) ;
const Integer Random_Integer (Integer& r, long size) ;
// =================================================================== //
// Primality tests and factorization algorithms
// =================================================================== //

// Those macros are parameters to fix

// primes known
// first array
#define LOGMAX 3512
#define TABMAX 32768
// second array
#define LOGMAX2 3031
#define TABMAX2 65536

// Bounds between big and small
#define BOUNDARY_factor TABMAX2
#define BOUNDARY_isprime TABMAX    
#define BOUNDARY_2_isprime TABMAX2    

// Algorithms chosen
#define big_isprime isprime_Miller

// First primes
#define PROD_first_primes 223092870
#define factor_first_primes(n) ((n==1)?1:((!(n % 29))?29:((!(n % 23))?23:((!(n % 19))?19:((!(n % 17))?17:((!(n % 2))?2:((!(n % 3))?3:((!(n % 5))?5:((!(n % 7))?7:((n % 11)?13:11))))))))))

// Switch between big and small
//#define isprime(n) probab_prime(n)
#define isprime(n) ( ((n)<BOUNDARY_isprime)?isprime_Erathostene(n):probab_prime(n))

// =================================================================== //
// Primality tests 
// =================================================================== //

int isprime_Miller(const Integer& n) ;
const Integer test_Lehmann(const Integer& n) ;
int isprime_Lehmann(const Integer& n) ;
int isprime_Erathostene(const Integer& p) ;
const Integer nextprime(const Integer& p) ;
const Integer prevprime(const Integer& p) ;




// =================================================================== //
// Factorizations
// =================================================================== //
const Integer Erathostene(const Integer& p) ;

// Those macros are parameters for Pollard's algorithms
// Pfct : must be somewhat a "random" function in Z/nZ
// Pcst is a slight alternative for for the Pfct x^2+1
#define Pollard_cst 1   
#define Pollard_fct(x,n) (x*x+Pollard_cst)%n

const Integer Pollard(const Integer&) ;
// This is the maximum number of iterations
// Result of Pollard_easy might not be prime then
const Integer Pollard_easy(const Integer&, const unsigned long );



#ifndef __GIVARO_EASY_FACTORIZATION__
#define big_factor(n) Pollard(n)
#else
unsigned long GIV_EASY_FACT() { return 100000; }
unsigned long GIV_EASY_FACT(const unsigned long a) { return a; }  

#define big_factor(n) Pollard_easy((n), GIV_EASY_FACT( __GIVARO_EASY_FACTORIZATION__) )
#endif


#define small_factor Erathostene

#define Integer_factor(n) (gcd(PROD_first_primes,n)==1?((n)<BOUNDARY_factor ? small_factor(n) : big_factor(n)):factor_first_primes(n))




#endif
