// =================================================================== //
// Givaro : Euler's phi function
//          Primitive roots.
// Needs list structures : stl ones for instance
// Time-stamp: <28 May 99 18:46:19 Jean-Guillaume.Dumas@imag.fr> 
// =================================================================== //

#ifndef _GIVARO_PRIMITIVE_ROOT_
#define _GIVARO_PRIMITIVE_ROOT_

#include <list>
using std::list;
#include "integer.h"
#include "intprime.h"
#include "intfactor.h"

// =================================================================== //
// Euler's phi function
// =================================================================== //

const Integer phi(const Integer& n) {
    if (n <= 1) return n;
    if (n <= 3) return n-1;
    list<Integer> Lf;
    Integer_factorset(n,Lf);
    Integer res = n;
    for(list<Integer>::iterator f=Lf.begin(); f!=Lf.end(); f++)
        res = (res / (*f)) * (*f - 1);
    return res;
}


// =================================================================== //
// Primitive Root
// =================================================================== //

const Integer prim_root(const Integer& n) {
        // n must be in {2,4,p^m,2p^m} where p is an odd prime
        // else infinite loop
    if (n <= 4) return n-1;
    if (! (n%int(4))) return Integer(0);
    Integer phin = phi(n);
    list<Integer> Lf;
    Integer_factorset(phin,Lf);
    list<Integer>::iterator f;
    for(f=Lf.begin();f!=Lf.end();++f)
            *f = phin / (*f);
    Integer A;
    int found;
    //srand48(BaseTimer::seed());
    do {
        do {
            Random_Integer(A,n-3);
            A += 3;
        } while ( (! (A%int(2))) || (gcd(A,n) != 1) );
        found = 1;
        for(f=Lf.begin();(f!=Lf.end() && found);f++)
            found = ( powmod(A,*f,n) != 1);
    } while (! found);
    return A;
}

    
        
const Integer lowest_prim_root(const Integer& n) {
        // n must be in {2,4,p^m,2p^m} where p is an odd prime
        // else returns zero
    if (n <= 4) return n-1;
    if (! (n%int(4))) return Integer(0);
    Integer phin = phi(n);
    list<Integer> Lf;
    Integer_factorset(phin,Lf);
    list<Integer>::iterator f;
    for(f=Lf.begin();f!=Lf.end();++f)
            *f = phin / (*f);
    int found=0;
    Integer A(1);
    for(;((A<n) && (! found));A+=2) {
        if (gcd(A,n) == 1) {
            found = 1;
            for(f=Lf.begin();(f!=Lf.end() && found);f++)
                found = ( powmod(A,*f,n) != 1);
        }
    }
    if (A<n)
        return A-2;
    else
        return Integer(0); 
}

int is_prim_root(const Integer& p, const Integer& n) {
        // returns 0 if failed
    int found=0;
    Integer phin = phi(n);
    list<Integer> Lf;
    Integer_factorset(phin,Lf);
    list<Integer>::iterator f=Lf.begin();
    Integer A = p % n;
    if (gcd(A,n) == 1) {
        found = 1;
        for(;(f!=Lf.end() && found);f++) {
            found = ( powmod(A,phin / (*f),n) != 1);
        }
    }
    return found;
}


const Integer order(const Integer& p, const Integer& n) {
        // returns 0 if failed
    Integer A = p % n;
    if (iszero(A))
	return Integer(zero);
    if (isone(A))
	return Integer(one);
    int primroot=0;
    Integer phin = phi(n),g,gg,un(one);
    list<Integer> Lf;
    Integer_factorset(phin,Lf);
    Lf.sort();
    list<Integer>::iterator f=Lf.begin();
    if (gcd(A,n) == 1) {
        primroot = 1;
        for(;f!=Lf.end();++f)
            if (! (primroot = ( powmod(A, g = (phin / (*f)),n) != un) ) )
                break;
        if (primroot)
            return phin;
        else {
            for(;f!=Lf.end();++f)
                while (iszero(g % *f) && powmod(A, gg = (g / (*f)),n) == un)
                    g = gg;
            return g;
        }
    }
    return Integer(zero);
}



#endif
