// =================================================================== //
// Givaro : Prime numbers
//              RSA public-key cipher codes
// Time-stamp: <29 Jan 99 19:51:26 Jean-Guillaume.Dumas@imag.fr> 
// =================================================================== //

#ifndef _GIVARO_RSA_Public_KEY_
#define _GIVARO_RSA_Public_KEY_

#include <iostream.h>
#include "integer.h"
#include "intprime.C"

// =================================================================== //
// log[10]
// =================================================================== //

long log(const Integer& n) {
    Integer p = n;
    long res = 0;
    do {
        p = p / 10;
        res++;
    } while (p > 0);
    return res;
}   
    

// =================================================================== //
// Text conversions
// =================================================================== //

                
void ecriture_str(const Integer& n) {
    Integer p = n;
    do {
        cout << char(p % 1000);
        p = p / 1000;
    } while (p>0);
}


void ecriture_Int(const Integer& p, const Integer& size) {
    if (p>0) {
        for(int i=0;i<(size - log(p)); i++)
            cout << "0";
        cout << p;
    }
}


void encipher(const Integer& m, const Integer& k) {
    long lm = log(m);
    int x;
    Integer pas, res;
    do {
        res = 0;
        pas = 1;
        for(int i = 0; i<lm/3; i++) {
            x = cin.get();
            if (cin.eof())
                break;
            res += x * pas;
            pas *= 1000;
        }
        ecriture_Int(powmod(res,k,m),lm);
    } while (! cin.eof());
}           

void decipher(const Integer& m, const Integer& u) {
    long lm = log(m);
    Integer pas, res;
    int * TC = new int[9];
    for(int j=0;j<10;j++)
        TC[j] = j;
    int * TR = &TC[-48];
    int * envers = new int[lm];
    do {
        res = 0;
        pas = 1;
        for(int i=0; i<lm; i++)
            envers[i] = TR[cin.get()];
        for(i=1;i<=lm;i++) {
            res += envers[lm-i]*pas;
            pas *= 10;
        }
        ecriture_str(powmod(res,u,m));
    } while (! cin.eof());
}           

// =================================================================== //
// Keys generation
// public keys are m and k, the secret key is u.
// ciphering is computing	: x^k mod m
// deciphering is computing	: b^u mod m
// since for any x, x^(k.u) = x mod m
// =================================================================== //
void keys_gen(long psize, long qsize, Integer& m, Integer& k, Integer& u) {
    Integer p1,q1;
    //srand48(BaseTimer::seed());
    Random_Integer(p1,long((2*psize)/3));
    Random_Integer(q1,long((2*qsize)/3));
    
    p1 = nextprime(p1);
    q1 = nextprime(q1);

    Integer d,l;
    
    Random_Integer(d,long(psize/3));
    Random_Integer(l,long(qsize/3));

    if (d%2)
        d = d + 1;
    if (l%2)
        l = l + 1;
    
    Integer p,q;

    do {
        p = p1 * d + 1;
        d = d + 2;
    } while (! (isprime(p) && isprime(p)) );
    
    do {
        q = q1 * l + 1;
        l = l + 2;
    } while (! (isprime(q) && isprime(q)) );
    
    Integer phim = (p-1)*(q-1);
    m = p*q;

    Integer v;
    
    do {
        Random_Integer(k,phim);
    } while (gcd(k,phim,u,v) != 1);
    u = u % phim;
}

// =================================================================== //
// Breaking codes
// =================================================================== //
Integer& point_break(Integer& u, const Integer& m, const Integer& k) {
    Integer p,v;
    p = Integer_factor(m);
    gcd(k,(p-1)*((m/p)-1),u,v);
    return u;
}

#endif
