#ifndef __B_B_Gershgorin_H_
#define __B_B_Gershgorin_H_

// ============================================================
// Copyright (C) LinBox Project 1999
//
// Black Blox Gershgorin
// An example of generic algorithm
// file : lin_b_b_gersh.h
// Preconditions : Field must provide a comparison function
// ============================================================

#ifndef GIVMAX
#define GIVMAX(a,b) ((a)<(b) ? b:a)
#endif

#ifndef GIVMIN
#define GIVMIN(a,b) ((a)<(b)?(a):(b))
#endif

#ifndef GIVABS
#define GIVABS(l) ((l)<0 ? -(l): (l))
#endif


template<class BBA, class TT>
class BBGershgorin {
private:
    BBA * _amat;

public:
    typedef BBA self_BlackBox;
    typedef typename BBA::PreferredInMatrix Right;
    typedef typename BBA::PreferredOutMatrix Left;
    typedef double K;
    
    BBGershgorin(){};
    BBGershgorin(BBA* a) : _amat(a){};

    const K& getradius() { return _radius; }
    const K& getataradius() { return _ata_radius; }
    const K& getaatradius() { return _aat_radius; }
    const K& getdiag() { return _diag; }
    const K& getatadiag() { return _ata_diag; }
    const K& getaatdiag() { return _aat_diag; }
    const K& getbound() { return _eigen_bound; }
    
    void makedata() {
        
        Indice i,j;
        K Zero(zero);
        _radius = Zero;
        _diag = Zero;
        _ata_diag = Zero;
        _ata_radius = Zero;
        _aat_diag = Zero;
        _aat_radius = Zero;
        K One(one);

        Right u( _amat->n_col() );
        Left v( _amat->n_row() );
        Right w( _amat->n_col() );
        Left y( _amat->n_row() );

        for(Indice i=_amat->n_row()-1; i>=0 ; i--)
            v[i] = One;

        for(Indice i=_amat->n_row()-1; i>=0 ; i--) {
            v[i] = One;
            _amat->ApplyTrans(u,v);
            K local_rad=Zero;
            for(Indice j=_amat->n_col()-1; j>=0 ; j--)
                if (j != i) local_rad += GIVABS(u[j]);
                else _diag = GIVMAX( _diag, GIVABS(u[j]) );
            _radius = GIVMAX( _radius, local_rad );

            _amat->Apply(y,u);
            local_rad=Zero;
            for(Indice j=_amat->n_row()-1; j>=0 ; j--)
                if (j != i) local_rad += GIVABS(y[j]);
                else _ata_diag = GIVMAX( _diag, GIVABS(y[j]) );
            _ata_radius = GIVMAX( _ata_radius, local_rad);

            v[i] = Zero;
        }

        K laat = _aat_diag+_aat_radius, lata = _ata_diag+_ata_radius;       
        _eigen_bound = GIVMIN( laat, lata);
    };      
    
private:
    K _radius;
    K _ata_radius;
    K _aat_radius;
    K _diag;
    K _ata_diag;
    K _aat_diag;
    K _eigen_bound;
};



#endif __B_B_Gershgorin_H_
