// ============================================================
// Copyright (C) LinBox Project 1999
//
// Black Blox SPARSE Gershgorin
// An example ofspecialization
// file : lin_sparse_gersh.h
// Preconditions : Field must provide a comparison function
// 
// ============================================================
#ifndef __SPARSE_B_B_Gershgorin_C_
#define __SPARSE_B_B_Gershgorin_C_

#include "sparse_vector.h"
#include "lin_sparse_b_b.h"
#include "lin_gershgorin.C"

template<class K>
class BBGershgorin < BlackBox < _IBB_VECTOR_ < Sparse_Vector < K > > > , double >  {
public:
    BBGershgorin(){};
    BBGershgorin(BlackBox < _IBB_VECTOR_ < Sparse_Vector < K > > > * a) : _amat(a){};

    const K& getradius() { return _radius; }
    const K& getataradius() { return _ata_radius; }
    const K& getaatradius() { return _aat_radius; }
    const K& getdiag() { return _diag; }
    const K& getatadiag() { return _ata_diag; }
    const K& getaatdiag() { return _aat_diag; }
    const K& getbound() { return _eigen_bound; }
    
    void makedata() {    
        Indice i,j;
        K Zero(zero);
        _radius = Zero;
        _diag = Zero;
        _ata_diag = Zero;
        _ata_radius = Zero;
        _aat_diag = Zero;
        _aat_radius = Zero;
        K One(one);

        for(i=_amat->n_row()-1; i>=0 ; --i) {
            K local_rad=Zero;
            K local_aat_diag=Zero;
            K local;
            for(j=(*_amat)[i].size()-1; j>=0 ; --j)
                if (j != i) {
                    local = GIVABS( (*_amat)[i][j].getvalue() );
                    local_rad += local;
                    local_aat_diag += local*local;
                } else _diag = GIVMAX( _diag, GIVABS( (*_amat)[i][j].getvalue() ) );
            _radius = GIVMAX( _radius, local_rad );
            _aat_diag = GIVMAX( _aat_diag, local_aat_diag );
        }

        Left u( _amat->n_col() );
        Right v( _amat->n_row() );
        Left w( _amat->n_col() );

        for(j=_amat->n_col()-1; j>0 ; --j)
                u[j] = Zero;
        u[0] = One;
        _amat->ApplyTrans(w,_amat->Apply(v,u));
        _ata_diag = w[0];

        for(j=_amat->n_col()-1; j>=0 ; --j)
                u[j] = One;
        _amat->ApplyTrans(w,_amat->Apply(v,u));
        for(j=_amat->n_col()-1; j>=0 ; --j)
            _ata_radius = GIVMAX (_ata_radius, GIVABS(w[j]-_ata_diag));

        Right y( _amat->n_row() );
        for(i=_amat->n_row()-1; i>=0 ; --i)
                y[i] = One;
        _amat->Apply(v,_amat->ApplyTrans(w,y));
        for(i=_amat->n_row()-1; i>=0 ; --i)
            _aat_radius = GIVMAX (_aat_radius, GIVABS(v[i]-_aat_diag));

        K laat = _aat_diag+_aat_radius, lata = _ata_diag+_ata_radius;       
        _eigen_bound = GIVMIN( laat, lata);
    };   
    
private:
    typedef typename BlackBox < _IBB_VECTOR_ < Sparse_Vector < K > > >::PreferredInMatrix Left;
    typedef typename BlackBox < _IBB_VECTOR_ < Sparse_Vector < K > > >::PreferredOutMatrix Right;
    BlackBox < _IBB_VECTOR_ < Sparse_Vector < K > > > * _amat;

    K _radius;
    K _ata_radius;
    K _aat_radius;
    K _diag;
    K _ata_diag;
    K _aat_diag;
    K _eigen_bound;
};



#endif __SPARSE_B_B_Gershgorin_C_
