#ifndef __LINBOX_DOMAIN_INTEGER_SMITH_NORMAL_FORM_C__
#define __LINBOX_DOMAIN_INTEGER_SMITH_NORMAL_FORM_C__
// ======================================================================= //
// Copyright (C) LinBox project 1999
// Givaro / Athapascan-1
// Smith form
// Time-stamp: <31 Jul 00 21:17:16 Jean-Guillaume.Dumas@imag.fr> 
// ======================================================================= //

#include <stdlib.h>
#include <stdio.h>

#include <givdomgfq.C>
#include "lin_dom_wiedemann.C"     // Symetric Wiedemann
#include "lin_dom_sparse_b_b.h"    // givaro sparse matrices
#include "lin_erase_bb.h"    // givaro sparse matrices

// #define MAXGFQSIZE 131072
#define MAXGFQSIZE 65536
#define NORMGFQSIZE 65536
#define MINGFQSIZE 32678


void wiedemann_rank(ulong MOD, char * File_Name, ulong & rank, double &wrong) {
    startActivity(GlobalCommentator, "Wiedemann", "Computing rank via Wiedemann algorithm", LVL_NORMAL, INTERNAL_DESCRIPTION); 
    typedef GFqDom<long>               GFqDomain;
    typedef GFqDomain::Residu_t        Residu;
//     SparseBlackBoxDom< GFqDomain >::Storage_t TheMat;
    SparseBlackBoxEraseDom< GFqDomain >::Storage_t TheMat;
//     long ni = TheMat.get_n_row(File_Name);
//     long nj = TheMat.n_col();
    unsigned long ni, nj; 
    {
        SparseBlackBoxEraseDom< GFqDomain > MF;
        MF.read (ni, nj, File_Name);
//cerr << "ni: " << ni << ", nj: " << nj << endl;
        if (ni < nj) {
            SparseBlackBoxStorage< long > TempMat;
            TempMat.init_transpose(File_Name);
            TempMat.write(File_Name);
        }
    }
    long sq = ni*nj;

    Residu expo = 0, qMOD = MOD, pMOD = 1;
    
    
    while ((qMOD <= MAXGFQSIZE) && ( (pMOD <= 50*sq) || (qMOD <= NORMGFQSIZE) ) ) {
        ++expo;
        pMOD = qMOD;
        qMOD *= MOD;
//cerr << "e: " << expo << ", p: " << pMOD << ", q: " << qMOD << endl;
    }    
    GFqDomain F(MOD,expo);
    char * UT = new char[200]; 
    sprintf(UT,"Computing rank mod %ld in GF(%ld).",MOD,pMOD); 
    activityReport(GlobalCommentator, UT, LVL_UNIMP, INTERNAL_DESCRIPTION ); 

//     SparseBlackBoxDom< GFqDomain > MF( F );
//     WiedemannDom< SparseBlackBoxDom< GFqDomain > >  WD( MF );
    SparseBlackBoxEraseDom< GFqDomain > MF( F );
    WiedemannDom< SparseBlackBoxEraseDom< GFqDomain > >  WD( MF );
    

//     MF.init(TheMat, File_Name);
    unsigned long addrank = MF.init_erase(TheMat, ni, nj, File_Name);
    if (addrank > 0) {
        sprintf(UT,"Reordering, ... matrix reduced, preliminary rank found : %ld.", addrank);
        activityReport(GlobalCommentator, UT, LVL_UNIMP, INTERNAL_DESCRIPTION ); 
        activityReport(GlobalCommentator, "Now computing the rank of the remaining matrix.", LVL_UNIMP, INTERNAL_DESCRIPTION );
    }
    int level;
    if ( (ni > 0) && (nj > 0)) {
//     WD.precondition(TheMat);
        WD.precondition(TheMat, ni, nj);
//     WD.seq_rank_sw(rank, TheMat);
        WD.seq_rank_sw(rank, TheMat, nj);
//     char * UT = new char[100]; 
//     if ( (long sq = TheMat.n_row()*TheMat.n_col()) > qMOD )
//     cerr << "sq: " << sq << ", whereas field size is : " << F.size() << endl;
        rank += addrank;
    
        if ( (6*sq) > pMOD ) {
            level = LVL_ALWAYS;
            sprintf(UT,"Rank %ld, computed via a heuristic, probability of success unknown.", rank); 
            wrong = 1;
        } else {
            level = LVL_IMP;
            sprintf(UT,"Rank %ld, correct with probability at least 1 - %e", rank, wrong = double(5.5*sq+1)/double(pMOD));
        }
    } else {
        level = LVL_NORMAL;
        sprintf(UT,"Rank %ld, computed deterministically.", rank = addrank); 
    }    
    stopActivity(GlobalCommentator, UT, level, INTERNAL_DESCRIPTION ); 
    delete [] UT;
}

    
void wiedemann_rank_N_torsion(Invariants<ulong>& Inv, char * File_Name, Integer IN) {
    startActivity(GlobalCommentator, "N-Torsion", "Rank N-Torsion Computation via wiedemann algorithm", LVL_NORMAL, INTERNAL_DESCRIPTION); 

    UIndice rank;
    Integer valence;
    ulong nbfactor;
    list<Integer> Lm;
    list<ulong> Le;
    
    nbfactor = factorset(IN,Lm,Le);

    list<Integer>::iterator im=Lm.begin();
    list<ulong>::iterator ie=Le.begin();

    Indice ni, nj;
    FILE* FileDes = fopen(File_Name, "r");
    if (FileDes != NULL) {
        fscanf(FileDes,"%ld %ld M\n",&ni, &nj) ;
    }
    fclose(FileDes);


    UIndice * r = new UIndice[nbfactor];
    UIndice ir;
    ulong ulim;
    Indice k, minr= GIVMIN(ni,nj);
    Integer pr=1;
    for( k=0,im=Lm.begin(),ie=Le.begin();im != Lm.end(); ++im, ++ie,++k)
        if (*ie > 0) {
            pr *= *im;
            double rank_wrong;
            wiedemann_rank(*im%(ulong)MAXPRIMESIZE, File_Name, r[k],rank_wrong);
            minr = GIVMIN(minr,r[k]);
            char * UT = new char[70]; sprintf(UT,"rank mod %ld : %ld", (*im).Int2long(), r[k]);
            activityReport(GlobalCommentator, UT, LVL_NORMAL, PARTIAL_RESULT);
        }
    
    Integer p( prevprime(MAXGFQSIZE - (ulong)lrand48()%(ulong)MAXGFQSIZE/2)) ;
    while (iszero(pr % p))
      p = prevprime(p);
    double rank_wrong;
    wiedemann_rank(p.Int2long(),File_Name,ir,rank_wrong);
    char * UT = new char[130]; sprintf(UT,"Integer rank (mod %ld) : %ld", (p).Int2long(), ir);
    activityReport(GlobalCommentator, UT, LVL_NORMAL, PARTIAL_RESULT);
    if (ir > minr) {
        Indice ssize = ir-minr;
        ulong * Smith = new ulong[ssize];
        Indice ll=ssize;
        for (;ll--;)
            Smith[ll] = 1;
    

        UIndice l=0,l1=0, count_ie;
        for( k=0,im=Lm.begin(),ie=Le.begin();im != Lm.end(); ++im, ++ie,++k)
            if (*ie > 0) {
                count_ie = 0;
                l = r[k];
                ulong p = (*im).Int2long();
                ulong pp = p;
//                 while (l < ir) {
//                     for(Indice lk=ir-l;lk;--lk)
//                         Smith[ssize-lk] *= p;
//                     pp *= p;
//                     if (++count_ie < *ie) {
//                         activityReport(GlobalCommentator, "Rank mod powers is NOT YET IMPLEMENTED, so result is valid only for the primes", LVL_NORMAL, PARTIAL_RESULT);
//                     } else
//                         break;
//                 }
                if (l < ir) {
                    Smith[ssize-ir+l] *= p;
                    sprintf(UT,"Rank mod powers is NOT YET IMPLEMENTED, so rank mod %ld should be checked.", (Smith[ssize-ir+l]*p));
                    activityReport(GlobalCommentator, UT, LVL_NORMAL, INTERNAL_DESCRIPTION);
                }
            }

        Inv.PushInvariant(1,minr);
        Indice lj=0;
        for(ll=0;ll<ssize;) {
            while ((++lj<ssize) && (Smith[lj] == Smith[ll])) {};
            Inv.PushInvariant(Smith[ll],lj-ll);
            ll = lj;
        }
        delete [] Smith;
    
    } else {
        Inv.PushInvariant(1,ir);
    }
    
    delete [] UT;
    delete [] r;
    stopActivity(GlobalCommentator, "Torsions computed via a heuristic, probability of success unknown.", LVL_ALWAYS, INTERNAL_DESCRIPTION ); 
};




#endif
