#ifndef __LINBOX_GAUSS_C__
#define __LINBOX_GAUSS_C__
// ================================================================================== //
// Copyright (C) Linbox project 1999.
// Givaro / Athapascan-1
// Calcul de rang par la mthode de Gauss pivot par ligne, sur matrice creuse
// Calculation of the rank by Gauss' method, (pivot by row ?? ), 
//   on sparse matrix. :-)
// Time-stamp: <02 Aug 00 11:36:54 Jean-Guillaume.Dumas@imag.fr> 
// ================================================================================== //

#include <time.h>
#include <stdio.h>
#include <fstream.h>
#include <timer.h>

#include <vector>
using std::vector;
#include <sparse_vector.h>
#include <givtype.h>

// --------------------------------------------
// Modulo operators
// int iszero(const Modulo& a ) { return !a ;};

template<class Ring>
Ring MY_Zpz_bezout(Ring a, Ring b, Ring& u, Ring& v)
{
    Ring u1,u2,u3 ;
    Ring v1,v2,v3 ;
    u1 = 1 ; u2 = 0 ; u3 = a ;
    v1 = 0 ; v2 = 1 ; v3 = b ;
    while (v3 != 0)
    {
        Ring q , t1, t2 ,t3 ;
        q = u3 / v3 ;
        t1 = u1 - q * v1 ; t2 = u2 - q * v2 ; t3 = u3 - q * v3 ;
        u1 = v1 ; u2 = v2 ; u3 = v3 ; v1 = t1 ; v2 = t2 ; v3 = t3 ;
    }
    u = u1 ; v = u2 ;
    return u3 ;
} ;

template<class Modulo>
Modulo MY_Zpz_inv (const Modulo a, const Modulo pp)
{
    typedef typename Signed_Trait<Modulo>::signed_type Ring;
    if (a == 1) return a;
    if (a == -1) return a;
    Ring u,v, d ;
    d = MY_Zpz_bezout(Ring(a),Ring(pp),u,v) ;

    if (d == -1) { d = -d ;  u = -u ; v = -v ; }
//  if (d != 1) {
//      cerr << "*** division by zero in: Modulo div " << endl;
//      cerr << a << '*'<< u << '+' << v << '*' << pp << '=' << u*a+v*pp << endl ;
//      cerr << "and gcd = " << d << endl ;
//      return 0;
//    }
    if (u <0) u += pp ;
    return u ;
}


template<class Ring>
Ring MY_gcd(Ring a, Ring b) {
  Ring q,r,d;
  Ring ma=a, mb=b;
  while (mb != 0) {
        q = ma / mb;
        r = ma - q * mb;
        ma = mb;
        mb = r;
  }
  return ma;
};

template<class Ring>
short MY_divides(Ring a, Ring b) {
    return iszero(b%a);
};

// ------------------------------------------------
// Pivot Searchers and column strategy
// ------------------------------------------------

template<class Vecteur>
void CherchePivot( Vecteur& lignepivot, Indice& indcol , Indice& indpermut ) {
    Indice nj =  lignepivot.size() ;
    if (nj) {
       indpermut= lignepivot[0].j();
       if (indpermut != indcol)
           lignepivot[0].change_j(indcol);
       indcol++ ;
    } else
	indpermut = -1;
}

template<class Vecteur, class D>
void CherchePivot( Vecteur& lignepivot, Indice& indcol , Indice& indpermut, D& columns ) {
    typedef typename Vecteur::elements E;
    typedef typename Vecteur::coefficientSpace F;
    Indice nj =  lignepivot.size() ;
    if (nj) {
       indpermut= lignepivot[0].j();


       Indice ds = columns.decr(indpermut),dl,p=0;
       for(Indice j=1;j<nj;++j) {
           if ( (dl=columns.decr(lignepivot[j].j()) ) < ds ) {
               ds = dl;
               p = j;
           }
       }
       if (p != 0) {
           if (indpermut == indcol) {
               F ttm = lignepivot[p].getvalue();
               indpermut = lignepivot[p].j();
               lignepivot[p].affect(lignepivot[0].getvalue());
               lignepivot[0].affect(ttm);
           } else {
               E ttm = lignepivot[p];
               indpermut = ttm.j();
               for(Indice m=p;m;--m)
                   lignepivot[m] = lignepivot[m-1];
               lignepivot[0] = ttm;
           }
       }        


       if (indpermut != indcol)
// no need to decrement/increment, already done during the search
           lignepivot[0].change_j(indcol);
       indcol++ ;
    } else
	indpermut = -1;
}

template<class Modulo, class Vecteur, class D>
void CherchePivot(Modulo PRIME, Vecteur& lignepivot, Indice& indcol , Indice& indpermut, D& columns ) {
    typedef typename Vecteur::elements E;
    typedef typename Vecteur::coefficientSpace F;
    Indice nj =  lignepivot.size() ;
    if (nj) {
       indpermut = lignepivot[0].j();
       Indice pp=0;
       for(;pp<nj;++pp)
           if (! MY_divides(PRIME,lignepivot[pp].getvalue()) ) break;

       if (pp < nj) {
           
           Indice ds = columns.getvalue( lignepivot[pp].j() ),dl,p=pp,j=pp;
           for(++j;j<nj;++j)
               if ( ( (dl=columns.getvalue(lignepivot[j].j()) ) < ds ) && (! MY_divides(PRIME,lignepivot[j].getvalue()) ) ) {
                   ds = dl;
                   p = j;
               }
           if (p != 0) {
               if (indpermut == indcol) {
                   F ttm = lignepivot[p].getvalue();
                   indpermut = lignepivot[p].j();
                   lignepivot[p].affect(lignepivot[0].getvalue());
                   lignepivot[0].affect(ttm);
               } else {
                   E ttm = lignepivot[p];
                   indpermut = ttm.j();
                   for(Indice m=p;m;--m)
                       lignepivot[m] = lignepivot[m-1];
                   lignepivot[0] = ttm;
               }
           } 
           if (indpermut != indcol)
               lignepivot[0].change_j(indcol);
           indcol++ ;
           for(j=nj;j--;)
               columns.decr(lignepivot[j].j()); 
       } else
           indpermut = -2;
    } else
	indpermut = -1;
}

template<class Vecteur>
void modtwo_CherchePivot( Vecteur& lignepivot, Indice& indcol , Indice& indpermut ) {
    Indice np =  lignepivot.size() ;
    if (np) {
       indpermut =  lignepivot[0];
       if (indpermut != indcol)
            lignepivot[0] = indcol;
       indcol++ ;
    } else
	indpermut = -1;
}

template<class Vecteur, class De>
void modtwo_CherchePivot( Vecteur& lignepivot, Indice& indcol , Indice& indpermut, De& columns) {
    Indice nj =  lignepivot.size() ;
    if (nj) {
       indpermut =  lignepivot[0];
       Indice ds = columns.decr(indpermut),dl,p=0;
       for(Indice j=1;j<nj;++j) {
           if ( (dl=columns.decr(lignepivot[j]) ) < ds ) {
               ds = dl;
               p = j;
           }
       }
       if (p != 0) {
           if (indpermut == indcol) {
               indpermut = lignepivot[p];
           } else {
               indpermut = lignepivot[p];
               for(Indice m=p;m;--m)
                   lignepivot[m] = lignepivot[m-1];
               lignepivot[0] = indpermut;
           }
       }        

       if (indpermut != indcol)
            lignepivot[0] = indcol;
       indcol++ ;
    } else
	indpermut = -1;
}

template<class T>
void CherchePivot(vector<T>& lignepivot, Indice& k, Indice& indpermut ) {
     Indice n =  lignepivot.size();
     Indice j = k;
     for (; j< n ; ++j )
        if (! iszero(lignepivot[j])) break ;
     if (j == n )
       indpermut = -1 ;
     else {
       indpermut = j ;
       if (indpermut != k) {
         T tmp = lignepivot[k] ;
         lignepivot[k] = lignepivot[j] ;
         lignepivot[j] = tmp ;
       }
       ++k;
     }
}


template<class Modulo, class Vecteur>
void PivotTotal(Modulo FMOD, Modulo& MOD, Modulo PRIME, Vecteur * LigneA, Indice Ni, Indice r, Indice& indcol , Indice& indpermut ) {
    typedef typename Vecteur::coefficientSpace F;
    typedef typename Vecteur::elements E;

    int divided = 0;
    Indice tt_size = 0;
    Indice k,j;
    Indice pk,pj;
    for(k=r; k<Ni; k++)
        tt_size += LigneA[k].size();
    if (tt_size) {
        int pivot_found = 0;
        int unit_found = 0;
        do {
            for(k = r; k<Ni; k++) {
                Indice nj = LigneA[k].size();
                if (nj) {
                    for (j = 0; j < nj ; j++ ) {
//                         Modulo tmp = MY_gcd(LigneA[k][j].getvalue(), PRIME);
//                         if ( GIVABS(tmp) == 1) break;
                        if (! MY_divides(PRIME,LigneA[k][j].getvalue()) ) break;
                    }
                    if (j == nj)
                        continue;
                    else {
                        if (pivot_found) {

                            if (LigneA[pk][pj].getvalue() > LigneA[k][j].getvalue()) {
                                pk = k;
                                pj = j;
                            }
                        } else {
                            pivot_found = 1;
                            pk = k;
                            pj = j;
                        }
                        if (LigneA[k][j].getvalue() == 1) {
                            unit_found = 1;
                            break;
                        }
                    }
                }
            }
            if (pivot_found) {
                if (! unit_found) {
                    k = pk;
                    j = pj;
                }
                Indice j_piv = LigneA[k][j].j();
                indpermut = j_piv  ;
                if (j_piv != indcol) {
                    Indice nj = LigneA[k].size();
                    Indice wj(0);
                    for(; wj<nj; ++wj)
                        if (LigneA[k][wj].j() >= indcol) break;
                    if ((wj<nj) && (LigneA[k][wj].j() == indcol)) {
                        if (wj == 0) {
                            F tmp = LigneA[k][j].getvalue();
                            LigneA[k][j].affect( LigneA[k][0].getvalue() );
                            LigneA[k][0].affect( tmp );
                        }
                    } else
                        if (j) {
                            E tmp = LigneA[k][j];
                            for(Indice l=j;l>0;l--)
                                LigneA[k][l] = LigneA[k][l-1];
                            LigneA[k][0] = tmp;
                        }
                        
                    
                }

                LigneA[k][0].change_j(indcol);
                ++indcol ;
                if (k != r) {
                        // Must copy physically ??
                    Vecteur tmpk (LigneA[k].size());
                    for(Indice i=0; i< LigneA[k].size(); i++) {
                        tmpk[i] = LigneA[k][i];
		    }	

		    Vecteur tmpr (LigneA[r].size());
		    for(Indice ii=0;ii< LigneA[r].size(); ii++) {
			tmpr[ii] = LigneA[r][ii];
		    }

                    LigneA[k] = tmpr;
                    LigneA[r] = tmpk;
                }
            } else {
                    // PRIME is in gcd(A)
                for(k = r; k<Ni; k++)
        	    if (LigneA[k].size())
                	for(Indice i=0;i<LigneA[k].size();i++)
                            LigneA[k][i].affect( LigneA[k][i].getvalue() / PRIME );
                MOD = MOD / PRIME;
                divided = 1;
		}
        } while (! pivot_found);
    } else
        indpermut = -1;
//     if (divided)
//         cerr << "In mod " << FMOD << ", subrank mod " << FMOD / MOD << " : " <<  indcol-1 << endl;
}





    
// ---------------------------------------------------------
// Line Elimination techniques with local permutations
// In the dense case L can be computed in place via LU
// ---------------------------------------------------------

template<class T>
void LU( vector<T>& lignecur,
         const vector<T>& lignepivot,
         Indice indcol ,
         Indice indpermut ) {

    Indice n =  lignecur.size() ;
    Indice k = indcol - 1 ;

        // permutation if one has been performed to compute the pivot
    if (indpermut != k) {
        T tmp = lignecur[k] ;
        lignecur[k] = lignecur[indpermut] ;
        lignecur[indpermut] = tmp ;
    }

    lignecur[k] = -lignecur[k] / lignepivot[k] ;
    T headcoeff = lignecur[k] ;
    Indice j = 0 ;
    for (; j < k ; ++j ) 
        lignecur[j] += headcoeff * lignepivot[j] ;
    for (++j ; j < n ; ++j )
        lignecur[j] += headcoeff * lignepivot[j] ;
}

template<class T>
void FaireElimination( vector<T>& lignecur,
                       const vector<T>& lignepivot,
                       Indice indcol ,
                       Indice indpermut ) {

    Indice n =  lignecur.size() ;
    Indice k = indcol - 1 ;

        // permutation if one has been performed to compute the pivot
    if (indpermut != k) {
        T tmp = lignecur[k] ;
        lignecur[k] = lignecur[indpermut] ;
        lignecur[indpermut] = tmp ;
    }

    lignecur[k] = -lignecur[k] / lignepivot[k] ;
    T headcoeff = lignecur[k] ;
    for ( Indice j = k+1 ; j < n ; ++j )
            lignecur[j] += headcoeff * lignepivot[j] ;
}


template<class Modulo, class Vecteur>
void FaireElimination( Modulo MOD,
                       Vecteur& lignecourante,
                       const Vecteur& lignepivot,
                       const Indice& indcol,
                       const Indice& indpermut ) {
    
    typedef typename Vecteur::coefficientSpace F;
    typedef typename Vecteur::elements E;
    

    Indice k = indcol - 1;
    Indice nj =  lignecourante.size() ;
    if (nj) {
        Indice j_head(0);
        for(; j_head<nj; ++j_head)
            if (lignecourante[j_head].j() >= indpermut) break;
        Indice bjh(j_head-1);
        if ((j_head<nj) && (lignecourante[j_head].j() == indpermut)) {
                // -------------------------------------------
                // Permutation
            if (indpermut != k) {
                if (lignecourante[0].j() == k) {     
                        // non zero  <--> non zero
                    F tmp = lignecourante[0].getvalue() ;
                    lignecourante[0].affect(lignecourante[j_head].getvalue() );
                    lignecourante[j_head].affect(tmp);
                } else {
                        // zero <--> non zero
                    E tmp = lignecourante[j_head];
                    tmp.change_j(k);
                    for(Indice l=j_head; l>0; l--)
                        lignecourante[l] = lignecourante[l-1];
                    lignecourante[0] = tmp;
                } 
                j_head = 0;
            }
                // -------------------------------------------
                // Elimination
            Indice npiv = lignepivot.size();
            Vecteur construit(nj + npiv);
                // construit : <-- ci
                // courante  : <-- m
                // pivot     : <-- l
            typedef typename Vecteur::iterator Viter;
            Viter ci = construit.begin();
            Indice m=1;
            Indice l(0);
                // A[i,k] <-- A[i,k] / A[k,k]
            lignecourante[0].affect(  ((ulong)( ( MOD-(lignecourante[0].getvalue()) ) * ( MY_Zpz_inv( lignepivot[0].getvalue(), MOD) ) ) ) % (ulong)MOD ) ;
            F headcoeff = lignecourante[0].getvalue() ;
        
            Indice j_piv;
            F tmp;
            for(;l<npiv;++l)
                if (lignepivot[l].j() > k) break;
                // for all j such that (j>k) and A[k,j]!=0
            for(;l<npiv;++l) {
                j_piv = lignepivot[l].j();
                    // if A[k,j]=0, then A[i,j] <-- A[i,j]
                for (;(m<nj) && (lignecourante[m].j() < j_piv);) 
                    *ci++ = lignecourante[m++];
                    // if A[i,j]!=0, then A[i,j] <-- A[i,j] - A[i,k]*A[k,j]
                if ((m<nj) && (lignecourante[m].j() == j_piv)) {
                    lignecourante[m].affect( ((ulong)( headcoeff  *  lignepivot[l].getvalue()  + lignecourante[m].getvalue() ) ) % (ulong)MOD );
                    if (! iszero(lignecourante[m].getvalue()))
                        *ci++ = lignecourante[m++];
                    else 
                        m++;
                } else if (! iszero(tmp = ((ulong)(headcoeff * lignepivot[l].getvalue())) %(ulong)MOD)) 
                    *ci++ =  E(j_piv, tmp );
            }
                // if A[k,j]=0, then A[i,j] <-- A[i,j] 
            for (;m<nj;)  
                *ci++ = lignecourante[m++];        
        
            construit.erase(ci,construit.end());
            lignecourante = construit;
        } else
                // -------------------------------------------
                // Permutation
            if (indpermut != k) {
                Indice l(0);
                for(; l<nj; ++l)
                    if (lignecourante[l].j() >= k) break;
                if ((l<nj) && (lignecourante[l].j() == k))  {
                        // non zero <--> zero
                    E tmp = lignecourante[l];
                    tmp.change_j(indpermut);
                    for(;l<bjh;l++)
                        lignecourante[l] = lignecourante[l+1];
                    lignecourante[bjh] = tmp;
                } // else
                    // zero <--> zero
            }
    }
}

template<class Modulo, class Vecteur, class De>
void FaireElimination( Modulo MOD,
                       Vecteur& lignecourante,
                       const Vecteur& lignepivot,
                       const Indice& indcol,
                       const Indice& indpermut,
                       De& columns) {
    
    typedef typename Vecteur::coefficientSpace F;
    typedef typename Vecteur::elements E;
    

    Indice k = indcol - 1;
    Indice nj =  lignecourante.size() ;
    if (nj) {
        Indice j_head(0);
        for(; j_head<nj; ++j_head)
            if (lignecourante[j_head].j() >= indpermut) break;
        Indice bjh(j_head-1);
        if ((j_head<nj) && (lignecourante[j_head].j() == indpermut)) {
                // -------------------------------------------
                // Permutation
            if (indpermut != k) {
                if (lignecourante[0].j() == k) {     
                        // non zero  <--> non zero
                    F tmp = lignecourante[0].getvalue() ;
                    lignecourante[0].affect(lignecourante[j_head].getvalue() );
                    lignecourante[j_head].affect(tmp);
                } else {
                        // zero <--> non zero
                    E tmp = lignecourante[j_head];
                    columns.decr(tmp.j());
                    columns.incr(k);
                    tmp.change_j(k);
                    for(Indice l=j_head; l>0; l--)
                        lignecourante[l] = lignecourante[l-1];
                    lignecourante[0] = tmp;
                } 
                j_head = 0;
            }
                // -------------------------------------------
                // Elimination
            Indice npiv = lignepivot.size();
            Vecteur construit(nj + npiv);
                // construit : <-- ci
                // courante  : <-- m
                // pivot     : <-- l
            typedef typename Vecteur::iterator Viter;
            Viter ci = construit.begin();
            Indice m=1;
            Indice l(0);
                // A[i,k] <-- A[i,k] / A[k,k]
            lignecourante[0].affect(  ((ulong)( ( MOD-(lignecourante[0].getvalue()) ) * ( MY_Zpz_inv( lignepivot[0].getvalue(), MOD) ) ) ) % (ulong)MOD ) ;
            F headcoeff = lignecourante[0].getvalue() ;
            columns.decr(lignecourante[0].j());
            
            Indice j_piv;
            F tmp;
            for(;l<npiv;++l)
                if (lignepivot[l].j() > k) break;
                // for all j such that (j>k) and A[k,j]!=0
            for(;l<npiv;++l) {
                j_piv = lignepivot[l].j();
                    // if A[k,j]=0, then A[i,j] <-- A[i,j]
                for (;(m<nj) && (lignecourante[m].j() < j_piv);) 
                    *ci++ = lignecourante[m++];
                    // if A[i,j]!=0, then A[i,j] <-- A[i,j] - A[i,k]*A[k,j]
                if ((m<nj) && (lignecourante[m].j() == j_piv)) {
                    lignecourante[m].affect( ((ulong)( headcoeff  *  lignepivot[l].getvalue()  + lignecourante[m].getvalue() ) ) % (ulong)MOD );
                    if (! iszero(lignecourante[m].getvalue()))
                        *ci++ = lignecourante[m++];
                    else 
                        columns.decr(lignecourante[m++].j());
//                         m++;
                } else if (! iszero(tmp = ((ulong)(headcoeff * lignepivot[l].getvalue())) %(ulong)MOD)) {
                    columns.incr(j_piv);
                    *ci++ =  E(j_piv, tmp );
                }
            }
                // if A[k,j]=0, then A[i,j] <-- A[i,j] 
            for (;m<nj;)  
                *ci++ = lignecourante[m++];        
        
            construit.erase(ci,construit.end());
            lignecourante = construit;
        } else
                // -------------------------------------------
                // Permutation
            if (indpermut != k) {
                Indice l(0);
                for(; l<nj; ++l)
                    if (lignecourante[l].j() >= k) break;
                if ((l<nj) && (lignecourante[l].j() == k))  {
                        // non zero <--> zero
                    E tmp = lignecourante[l];
                    columns.decr(tmp.j());
                    columns.incr(indpermut);
                    tmp.change_j(indpermut);
                    for(;l<bjh;l++)
                        lignecourante[l] = lignecourante[l+1];
                    lignecourante[bjh] = tmp;
                } // else
                    // zero <--> zero
            }
    }
}


template<class Vecteur>
void modtwo_FaireElimination( Vecteur& lignecourante,
                              const Vecteur& lignepivot,
                              const Indice& indcol,
                              const Indice& indpermut ) {
    
    Indice k = indcol - 1;
    Indice nj =  lignecourante.size() ;
    if (nj) {
        Indice j_head(0);
        for(; j_head<nj; j_head++)
            if (lignecourante[j_head] >= indpermut) break;
        Indice bjh(j_head-1);
        if ((j_head<nj) && (lignecourante[j_head] == indpermut)) {
                // -------------------------------------------
                // Permutation
            if (indpermut != k) {
                if (lignecourante[0] != k) {
                        // zero <--> non zero
                    for(Indice l=j_head; l>0; --l)
                        lignecourante[l] = lignecourante[l-1];
                    lignecourante[0] = k;
                }
                j_head = 0;
            }
                // -------------------------------------------
                // Elimination
                // Precondition npiv != 0 : because of existing pivot !
            Indice npiv = lignepivot.size(); 
            Vecteur construit(nj + npiv);
                // construit : <-- ci
                // courante  : <-- m
                // pivot     : <-- l
            typedef typename Vecteur::iterator Viter;
            Viter ci = construit.begin();
            Indice m(1);
            Indice l(0);

            Indice j_piv;
            for(; l<npiv; l++)
                if (lignepivot[l] > k) break;
                    // for all j such that (j>k) and A[k,j]!=0
            while(l < npiv) {
                j_piv = lignepivot[l++];
                    // if A[k,j]=0, then A[i,j] <-- A[i,j]
                for (; (m<nj) && (lignecourante[m] < j_piv); )
                    *ci++ = lignecourante[m++];
                
                    // if A[i,j]!=0, then A[i,j] <-- A[i,j] - A[i,k]*A[k,j] == 0 
                if ((m<nj) && (lignecourante[m] == j_piv)) {
                    m++;
                } else {
                        // A[i,j] <-- - A[i,k]*A[k,j] == 1 
                    *ci++ = j_piv;
                }
            }
                // if A[k,j]=0, then A[i,j] <-- A[i,j] 
            for (;m<nj;)
                *ci++ = lignecourante[m++];
        
            construit.erase(ci,construit.end());
            lignecourante = construit;
        } else
                // -------------------------------------------
                // Permutation
            if (indpermut != k) {
                Indice l(0);
                for(; l<nj; ++l)
                    if (lignecourante[l] >= k) break;
                if ((l<nj) && (lignecourante[l] == k))  {
                        // non zero <--> zero
                    for(;l<bjh; ++l)
                        lignecourante[l] = lignecourante[l+1];
                    lignecourante[bjh] = indpermut;
                } // else
                    // zero <--> zero
            }
    }
}


template<class Vecteur, class De>
void modtwo_FaireElimination( Vecteur& lignecourante,
                              const Vecteur& lignepivot,
                              const Indice& indcol,
                              const Indice& indpermut,
                              De& columns) {
    
    Indice k = indcol - 1;
    Indice nj =  lignecourante.size() ;
    if (nj) {
        Indice j_head(0);
        for(; j_head<nj; j_head++)
            if (lignecourante[j_head] >= indpermut) break;
        Indice bjh(j_head-1);
        if ((j_head<nj) && (lignecourante[j_head] == indpermut)) {
                // -------------------------------------------
                // Permutation
            if (indpermut != k) {
                if (lignecourante[0] != k) {
                        // zero <--> non zero
                    columns.decr(lignecourante[0]);
                    columns.incr(k);
                    for(Indice l=j_head; l>0; --l)
                        lignecourante[l] = lignecourante[l-1];
                    lignecourante[0] = k;
                }
                j_head = 0;
            }
                // -------------------------------------------
                // Elimination
                // Precondition npiv != 0 : because of existing pivot !
            Indice npiv = lignepivot.size(); 
            Vecteur construit(nj + npiv);
                // construit : <-- ci
                // courante  : <-- m
                // pivot     : <-- l
            typedef typename Vecteur::iterator Viter;
            Viter ci = construit.begin();
            Indice m(1);
            Indice l(0);
            columns.decr(lignecourante[0]);

            Indice j_piv;
            for(; l<npiv; l++)
                if (lignepivot[l] > k) break;
                    // for all j such that (j>k) and A[k,j]!=0
            while(l < npiv) {
                j_piv = lignepivot[l++];
                    // if A[k,j]=0, then A[i,j] <-- A[i,j]
                for (; (m<nj) && (lignecourante[m] < j_piv); )
                    *ci++ = lignecourante[m++];
                
                    // if A[i,j]!=0, then A[i,j] <-- A[i,j] - A[i,k]*A[k,j] == 0 
                if ((m<nj) && (lignecourante[m] == j_piv)) {
                    columns.decr(lignecourante[m++]);
                } else {
                        // A[i,j] <-- - A[i,k]*A[k,j] == 1 
                    columns.incr(j_piv);
                    *ci++ = j_piv;
                }
            }
                // if A[k,j]=0, then A[i,j] <-- A[i,j] 
            for (;m<nj;)
                *ci++ = lignecourante[m++];
        
            construit.erase(ci,construit.end());
            lignecourante = construit;
        } else
                // -------------------------------------------
                // Permutation
            if (indpermut != k) {
                Indice l(0);
                for(; l<nj; ++l)
                    if (lignecourante[l] >= k) break;
                if ((l<nj) && (lignecourante[l] == k))  {
                        // non zero <--> zero
                    columns.decr(lignecourante[l]);
                    columns.incr(indpermut);
                    for(;l<bjh; ++l)
                        lignecourante[l] = lignecourante[l+1];
                    lignecourante[bjh] = indpermut;
                } // else
                    // zero <--> zero
            }
    }
}



template<class Vecteur>
void FaireElimination( Vecteur& lignecourante,
                       const Vecteur& lignepivot,
                       const Indice& indcol,
                       const Indice& indpermut ) {
    
    typedef typename Vecteur::coefficientSpace F;
    typedef typename Vecteur::elements E;

    Indice k = indcol - 1;
    Indice nj =  lignecourante.size() ;
    if (nj) {
       Indice j_head(0);
       for(; j_head<nj; ++j_head)
           if (lignecourante[j_head].j() >= indpermut) break;
       Indice bjh(j_head-1);
       if ((j_head<nj) && (lignecourante[j_head].j() == indpermut)) {
                // -------------------------------------------
                // Permutation
            if (indpermut != k) {
                if (lignecourante[0].j() == k) {     
                        // non zero  <--> non zero
                    F tmp = lignecourante[0].getvalue() ;
                    lignecourante[0].affect(lignecourante[j_head].getvalue() );
                    lignecourante[j_head].affect(tmp);
                } else {
                        // zero <--> non zero
                    E tmp = lignecourante[j_head];
                    tmp.change_j(k);
                    for(Indice l=j_head; l>0; l--)
                        lignecourante[l] = lignecourante[l-1];
                    lignecourante[0] = tmp;
                } 
                 j_head = 0;
            }
                // -------------------------------------------
                // Elimination
            Indice npiv = lignepivot.size();
            Vecteur construit(nj + npiv);
                // construit : <-- j
                // courante  : <-- m
                // pivot     : <-- l
            Indice j=0;
            Indice m=j_head+1;
                // A[i,k] <-- - A[i,k] / A[k,k]

            lignecourante[j_head].affect(  -lignecourante[j_head].getvalue() / lignepivot[0].getvalue() ) ;
            F headcoeff = lignecourante[j_head].getvalue() ;
        
                // if A[k,j]=0, then A[i,j] <-- A[i,j]
            while(j < j_head) {
                construit[j] = lignecourante[j];
                j++;
            }
            Indice bef_piv = 0, j_piv;

            Indice l(0);
            for(; l<npiv; l++)
                if (lignepivot[l].j() > k) break;
                    // for all j such that (j>k) and A[k,j]!=0
                while(l < npiv) {
                    j_piv = lignepivot[l].j();
                        // if A[k,j]=0, then A[i,j] <-- A[i,j]
                     while ((m<nj) && (lignecourante[m].j() < j_piv))
                        construit[j++] = lignecourante[m++];
                        // if A[i,j]!=0, then A[i,j] <-- A[i,j] - A[i,k]*A[k,j]
                    if ((m<nj) && (lignecourante[m].j() == j_piv)) {
//                         lignecourante[m].affect( headcoeff  *  lignepivot[l].getvalue()  + lignecourante[m].getvalue() );
//                         if (! iszero(lignecourante[m].getvalue()))
//                             construit[j++] = lignecourante[m++];
                        F tmp; tmp.muladd(headcoeff,lignepivot[l].getvalue(),lignecourante[m].getvalue());
                        if (! iszero(tmp)) {
                            lignecourante[m].affect( tmp );
                            construit[j++] = lignecourante[m++];
                        }
                        else 
                            m++;
                    
                    } else {
                        F tmp = headcoeff * lignepivot[l].getvalue();
                        if (! iszero(tmp))
                            construit[j++] = E(j_piv, tmp );
                    }
                    l++;
                }
        
                // if A[k,j]=0, then A[i,j] <-- A[i,j] 
            while (m<nj)
                construit[j++] = lignecourante[m++];
        
//             Vecteur final(j);
//             if (j)
//                 for(Indice p=0;p<j;p++)
//                     final[p] = construit[p];
//             lignecourante = final;
            construit.resize(j);
            lignecourante = construit;
        } else
                // -------------------------------------------
                // Permutation
            if (indpermut != k) {
                Indice l(0);
                for(; l<nj; ++l)
                    if (lignecourante[l].j() >= k) break;
                if ((l<nj) && (lignecourante[l].j() == k))  {
                        // non zero <--> zero
                    E tmp = lignecourante[l];
                    tmp.change_j(indpermut);
                    for(;l<bjh;l++)
                        lignecourante[l] = lignecourante[l+1];
                    lignecourante[bjh] = tmp;
                } // else
                    // zero <--> zero
            }
    }
}

template<class Vecteur, class D>
void FaireElimination( Vecteur& lignecourante,
                       const Vecteur& lignepivot,
                       const Indice& indcol,
                       const Indice& indpermut,
                       D& columns) {
    
    typedef typename Vecteur::coefficientSpace F;
    typedef typename Vecteur::elements E;

    Indice k = indcol - 1;
    Indice nj =  lignecourante.size() ;
    if (nj) {
       Indice j_head(0);
       for(; j_head<nj; ++j_head)
           if (lignecourante[j_head].j() >= indpermut) break;
       Indice bjh(j_head-1);
       if ((j_head<nj) && (lignecourante[j_head].j() == indpermut)) {
                // -------------------------------------------
                // Permutation
            if (indpermut != k) {
                if (lignecourante[0].j() == k) {     
                        // non zero  <--> non zero
                    F tmp = lignecourante[0].getvalue() ;
                    lignecourante[0].affect(lignecourante[j_head].getvalue() );
                    lignecourante[j_head].affect(tmp);
                } else {
                        // zero <--> non zero
                    E tmp = lignecourante[j_head];
                    columns.decr(tmp.j());
                    columns.incr(k);
                    tmp.change_j(k);
                    for(Indice l=j_head; l>0; l--)
                        lignecourante[l] = lignecourante[l-1];
                    lignecourante[0] = tmp;
                } 
                 j_head = 0;
            }
                // -------------------------------------------
                // Elimination
            Indice npiv = lignepivot.size();
            Vecteur construit(nj + npiv);
                // construit : <-- j
                // courante  : <-- m
                // pivot     : <-- l
            Indice j=0;
            Indice m=j_head+1;
                // A[i,k] <-- - A[i,k] / A[k,k]

            lignecourante[j_head].affect(  -lignecourante[j_head].getvalue() / lignepivot[0].getvalue() ) ;
            F headcoeff = lignecourante[j_head].getvalue() ;
            columns.decr(lignecourante[j_head].j());
        
                // if A[k,j]=0, then A[i,j] <-- A[i,j]
            while(j < j_head) {
                construit[j] = lignecourante[j];
                j++;
            }
            Indice bef_piv = 0, j_piv;

            Indice l(0);
            for(; l<npiv; l++)
                if (lignepivot[l].j() > k) break;
                    // for all j such that (j>k) and A[k,j]!=0
                while(l < npiv) {
                    j_piv = lignepivot[l].j();
                        // if A[k,j]=0, then A[i,j] <-- A[i,j]
                     while ((m<nj) && (lignecourante[m].j() < j_piv))
                        construit[j++] = lignecourante[m++];
                        // if A[i,j]!=0, then A[i,j] <-- A[i,j] - A[i,k]*A[k,j]
                    if ((m<nj) && (lignecourante[m].j() == j_piv)) {
                        F tmp; tmp.muladd(headcoeff,lignepivot[l].getvalue(),lignecourante[m].getvalue());
                        if (! iszero(tmp)) {
                            lignecourante[m].affect( tmp );
                            construit[j++] = lignecourante[m++];
                        } else
                            columns.decr(lignecourante[m++].j());
//                             m++;
                    
                    } else {
                        F tmp = headcoeff * lignepivot[l].getvalue();
                        if (! iszero(tmp)) {
                            columns.incr(j_piv);
                            construit[j++] = E(j_piv, tmp );
                        }
                    }
                    l++;
                }
        
                // if A[k,j]=0, then A[i,j] <-- A[i,j] 
            while (m<nj)
                construit[j++] = lignecourante[m++];
        
            construit.resize(j);
            lignecourante = construit;
        } else
                // -------------------------------------------
                // Permutation
            if (indpermut != k) {
                Indice l(0);
                for(; l<nj; ++l)
                    if (lignecourante[l].j() >= k) break;
                if ((l<nj) && (lignecourante[l].j() == k))  {
                        // non zero <--> zero
                    E tmp = lignecourante[l];
                    columns.decr(tmp.j());
                    columns.incr(indpermut);
                    tmp.change_j(indpermut);
                    for(;l<bjh;l++)
                        lignecourante[l] = lignecourante[l+1];
                    lignecourante[bjh] = tmp;
                } // else
                    // zero <--> zero
            }
    }
}

// ------------------------------------------------------
// Rank calculators, defining row strategy
// ------------------------------------------------------

template<class BB>
void modtwo_gauss_rank (const BB& SLA, UIndice& rank) {
    UIndice Ni = SLA.n_row();
    typedef UIndice T;
    typedef vector< T > Vecteur;
    
    
    Vecteur * LigneA = new Vecteur[Ni];
    Indice jj;
    for(jj=0; jj<Ni; jj++) {
        Vecteur toto((SLA[jj]).size());
        Indice rs=0;
	Indice k=0;
        for(; k<(SLA[jj]).size(); k++)
            if ((SLA[jj])[k].getvalue() & 0x1)
                toto[rs++] = (SLA[jj])[k].j(); 
        LigneA[jj] = Vecteur(rs);
        for(k=0; k<rs; k++)
            LigneA[jj][k] = toto[k];
    }

    Indice last = Ni-1;
    Indice c;
    Indice indcol(0);
    
    
//     for (Indice k=0; k<last;k++) {
    for (Indice k=0; k<last;++k) {
        Indice l,p=k,s=LigneA[k].size(),sl;
        if (s) {
            for(l=k+1; l < Ni; ++l)
                if (((sl=LigneA[l].size()) < s) && (sl)) {
                    s = sl;
                    p = l;
                }
            if (p != k) {
                Vecteur vtm = LigneA[k];
                LigneA[k] = LigneA[p];
                LigneA[p] = vtm;
            }       
            modtwo_CherchePivot( LigneA[k], indcol, c ) ;
            if (c != -1)
                for(Indice l=k + 1; l < Ni; l++)
                    modtwo_FaireElimination(LigneA[l], LigneA[k], indcol, c);
        }
    }
    modtwo_CherchePivot( LigneA[last], indcol, c );
    
    rank = indcol;
}

template<class BB>
void modtwo_gauss_rank2 (const BB& SLA, UIndice& rank) {
    UIndice Ni = SLA.n_row();
    typedef UIndice T;
    typedef vector< T > Vecteur;
    
    
    Vecteur * LigneA = new Vecteur[Ni];
    Indice jj;
    for(jj=0; jj<Ni; jj++) {
        Vecteur toto((SLA[jj]).size());
        Indice rs=0;
	Indice k=0;
        for(; k<(SLA[jj]).size(); k++)
            if ((SLA[jj])[k].getvalue() & 0x1)
                toto[rs++] = (SLA[jj])[k].j(); 
        LigneA[jj] = Vecteur(rs);
        for(k=0; k<rs; k++)
            LigneA[jj][k] = toto[k];
    }

    Indice last = Ni-1;
    Indice c;
    Indice indcol(0);
    
    
    for (Indice k=0; k<last;k++) {
        modtwo_CherchePivot( LigneA[k], indcol, c ) ;
        if (c != -1)
            for(Indice l=k + 1; l < Ni; l++)
                modtwo_FaireElimination(LigneA[l], LigneA[k], indcol, c);
    }
    modtwo_CherchePivot( LigneA[last], indcol, c );
    rank = indcol;
}

template<class BB, class De>
void modtwo_gauss_rank (const BB& SLA, UIndice& rank, const De& density_trait) {
    UIndice Ni = SLA.n_row();
    typedef UIndice T;
    typedef vector< T > Vecteur;
    char * UT = new char[30]; sprintf(UT,"Elimination of %ld rows.", Ni);
    startActivity(GlobalCommentator, "Gauss", UT, LVL_BLABLA, INTERNAL_DESCRIPTION ); 
    
    De col_density(SLA.n_col());

    Vecteur * LigneA = new Vecteur[Ni];
    Indice jj;
    Vecteur toto;
    for(jj=0; jj<Ni; jj++) {
//         Vecteur toto((SLA[jj]).size());
        toto.resize((SLA[jj]).size());
        Indice rs=0;
	Indice k=0;
        for(; k<(SLA[jj]).size(); k++)
            if ((SLA[jj])[k].getvalue() & 0x1){
                col_density.incr((SLA[jj])[k].j());
                toto[rs++] = (SLA[jj])[k].j(); 
            }
//         LigneA[jj] = Vecteur(rs);
        LigneA[jj].resize(rs);
        for(k=0; k<rs; k++)
            LigneA[jj][k] = toto[k];
    }

    Indice last = Ni-1;
    Indice c;
    Indice indcol(0);
    
    
    for (Indice k=0; k<last;++k) {
        if (! (k % 1000) )
            progressReport(GlobalCommentator, "rows", LVL_NORMAL, k, last);
        Indice l,p=k,s=LigneA[k].size(),sl;
        if (s) {
            for(l=k+1; l < Ni; ++l)
                if (((sl=LigneA[l].size()) < s) && (sl)) {
                    s = sl;
                    p = l;
                }
            if (p != k) {
                Vecteur vtm = LigneA[k];
                LigneA[k] = LigneA[p];
                LigneA[p] = vtm;
            }       
            modtwo_CherchePivot( LigneA[k], indcol, c, col_density ) ;
            if (c != -1)
                for(Indice l=k + 1; l < Ni; l++)
                    modtwo_FaireElimination(LigneA[l], LigneA[k], indcol, c, col_density);
        }
    }
    modtwo_CherchePivot( LigneA[last], indcol, c );
    
    rank = indcol;
    stopActivity(GlobalCommentator, "", LVL_NEVER, INTERNAL_DESCRIPTION ); 
}


template<class Modulo, class BB>
void gauss_rank (Modulo MOD, const BB& SLA, UIndice& rank) {
    Indice Ni = SLA.n_row();
    typedef typename BB::coefficientSpace K;
    typedef Sparse_Vector< K > KVecteur;
    typedef Sparse_Vector< Modulo > Vecteur;
    typedef Modulo F;
    typedef typename Vecteur::elements E;    
    
    K KMOD(MOD);
    Vecteur * LigneA = new Vecteur[Ni];
    Indice jj;
    for(jj=0; jj<Ni; jj++) {
        const KVecteur& tmp = SLA[jj];
        Vecteur toto(tmp.size());
        Indice rs=0;
	Indice k=0;
        for(; k<tmp.size(); k++) {
            K r = tmp[k].getvalue();
            if ((r <0) || (r >= KMOD)) r = r % KMOD ;
            if (r <0) r = r + KMOD ;
            if (! iszero(r))
                toto[rs++] = E(tmp[k].j(), F(r)); 
        }
        LigneA[jj] = Vecteur(rs);
        for(k=0; k<rs; k++)
            LigneA[jj][k] = toto[k];
    }
    
    Indice last = Ni-1;
    Indice c;
    Indice indcol(0);
    
    
    for (Indice k=0; k<last;k++) {
        CherchePivot( LigneA[k], indcol, c ) ;
        if (c != -1)
            for(Indice l=k + 1; l < Ni; l++)
                FaireElimination(MOD, LigneA[l], LigneA[k], indcol, c);
    }
    CherchePivot( LigneA[last], indcol, c );
    
    delete [] LigneA;
        
    rank = indcol;
}

template<class Modulo, class BB>
void prime_rank (Modulo MOD, const BB& SLA, UIndice& rank) {
    gauss_rank(MOD,SLA,rank);
}

template<class Modulo, class BB, class De>
void gauss_rank (Modulo MOD, const BB& SLA, UIndice& rank, const De& density_trait) {
    Indice Ni = SLA.n_row();
    typedef typename BB::coefficientSpace K;
    typedef Sparse_Vector< K > KVecteur;
    typedef Sparse_Vector< Modulo > Vecteur;
    typedef Modulo F;
    typedef typename Vecteur::elements E;    
    
    char * UT = new char[30]; sprintf(UT,"Elimination of %ld rows.", Ni);
    startActivity(GlobalCommentator, "Gauss", UT, LVL_BLABLA, INTERNAL_DESCRIPTION ); 
    De col_density(SLA.n_col());
    K KMOD(MOD);
    Vecteur * LigneA = new Vecteur[Ni];
    Indice jj;
    Vecteur toto;
    for(jj=0; jj<Ni; jj++) {
        const KVecteur& tmp = SLA[jj];
//         Vecteur toto(tmp.size());
        toto.resize(tmp.size());
        Indice rs=0;
	Indice k=0;
        for(; k<tmp.size(); k++) {
            K r = tmp[k].getvalue();
            if ((r <0) || (r >= KMOD)) r = r % KMOD ;
            if (r <0) r = r + KMOD ;
            if (! iszero(r)){
                col_density.incr(tmp[k].j());
                toto[rs++] = E(tmp[k].j(), F(r)); 
            }
        }
//         LigneA[jj] = Vecteur(rs);
        LigneA[jj].resize(rs);
        for(k=0; k<rs; k++)
            LigneA[jj][k] = toto[k];
    }
    
    Indice last = Ni-1;
    Indice c;
    Indice indcol(0);
    
    for (Indice k=0; k<last;++k) {
        if (! (k % 1000) )
            progressReport(GlobalCommentator, "rows", LVL_NORMAL, k, last);
        Indice l,p=k,s=LigneA[k].size(),sl;
        if (s) {
            for(l=k+1; l < Ni; ++l)
                if (((sl=LigneA[l].size()) < s) && (sl)) {
                    s = sl;
                    p = l;
                }
            if (p != k) {
                Vecteur vtm = LigneA[k];
                LigneA[k] = LigneA[p];
                LigneA[p] = vtm;
            }       
            CherchePivot( LigneA[k], indcol, c , col_density) ;
            if (c != -1)
                for(l=k + 1; l < Ni; ++l)
                    FaireElimination(MOD, LigneA[l], LigneA[k], indcol, c, col_density);
        }
    }
    CherchePivot( LigneA[last], indcol, c );
    
    delete [] LigneA;
        
    rank = indcol;
    stopActivity(GlobalCommentator, "", LVL_NEVER, INTERNAL_DESCRIPTION ); 
}

template<class Modulo, class BB>
void gauss_rank (Modulo FMOD, Modulo PRIME, const BB& SLA, UIndice& rank){
    Indice Ni = SLA.n_row();
    typedef typename BB::coefficientSpace K;
    typedef Sparse_Vector< K > KVecteur;
    typedef Sparse_Vector< Modulo > Vecteur;
    typedef Modulo F;
    typedef typename Vecteur::elements E;
    
    Modulo MOD = FMOD;
    K KMOD(MOD);
    Vecteur * LigneA = new Vecteur[Ni];
    Indice jj;
    for(jj=0; jj<Ni; jj++) {
        const KVecteur& tmp = SLA[jj];
        Vecteur toto(tmp.size());
        Indice rs=0;
	Indice k=0;
        for(; k<tmp.size(); k++) {
            K r = tmp[k].getvalue();
            if ((r <0) || (r >= KMOD)) r = r % KMOD ;
            if (r <0) r = r + KMOD ;
            if (! iszero(r)) 
                toto[rs++] = E(tmp[k].j(),F(r));            
        }
        LigneA[jj] = Vecteur(rs);
        for(k=0; k<rs; k++)
            LigneA[jj][k] = toto[k];
    }
    
    Indice last = Ni-1;
    Indice c;
    Indice indcol(0);
    
    for (Indice k=0; k<last;k++) {
        PivotTotal(FMOD, MOD, PRIME, LigneA, Ni, k, indcol, c ) ;
        if (c != -1)
            for(Indice l=k + 1; l < Ni; l++)
                FaireElimination(MOD, LigneA[l], LigneA[k], indcol, c);
    }
    PivotTotal(FMOD, MOD, PRIME, LigneA, Ni, last, indcol, c );

    delete [] LigneA;
        
    rank = indcol;
}

template<class Modulo, class BB>
void prime_power_rank (Modulo FMOD, Modulo PRIME, const BB& SLA, UIndice& rank){
    gauss_rank(FMOD,PRIME,SLA,rank);
}

template<class Modulo, class BB, class D>
void gauss_rank(Modulo FMOD, Modulo PRIME, const BB& SLA, UIndice& rank, const D& density_trait) {
    Indice Ni = SLA.n_row(), Nj = SLA.n_col();
    typedef typename BB::coefficientSpace K;
    typedef Sparse_Vector< K > KVecteur;
    typedef Sparse_Vector< Modulo > Vecteur;
    typedef Modulo F;
    typedef typename Vecteur::elements E;
    
    char * UT = new char[80]; sprintf(UT,"Elimination of %ld rows.", Ni);
    startActivity(GlobalCommentator, "PP Gauss", UT, LVL_BLABLA, INTERNAL_DESCRIPTION ); 
    Modulo MOD = FMOD;
    K KMOD(MOD);
    
    D col_density(Nj);

    Vecteur * LigneA = new Vecteur[Ni];
//     vector< Vecteur > LigneA(Ni);
    Indice jj;
    Vecteur toto;
    for(jj=0; jj<Ni; jj++) {
        const KVecteur& tmp = SLA[jj];
//         Vecteur toto(tmp.size());
        toto.resize(tmp.size());
        Indice rs=0;
	Indice k=0;
        for(; k<tmp.size(); k++) {
            K r = tmp[k].getvalue();
            if ((r <0) || (r >= KMOD)) r = r % KMOD ;
            if (r <0) r = r + KMOD ;
            if (! iszero(r)) {
                col_density.incr(tmp[k].j());
                toto[rs++] = E(tmp[k].j(),F(r)); 
            }
        }
        toto.resize(rs);
        LigneA[jj] = toto;
    }

    Indice last = Ni-1;
    Indice c;
    Indice indcol(0);

    for (Indice k=0; k<last;++k) {
        if (! (k % 1000) )
            progressReport(GlobalCommentator, "rows", LVL_NORMAL, k, last);
        Indice p=k;
        for(;;) {
            Indice M=Ni,s,sl;
            for(p=k,s=LigneA[k].size(); M > k;) {
                if (s) {
                    for(Indice l=k+1; l < M; ++l)
                        if (((sl=LigneA[l].size()) < s) && (sl)) {
                            s = sl;
                            p = l;
                        }
                }
                CherchePivot( PRIME, LigneA[p], indcol, c , col_density) ;
                if (c > -2) break;
                else {
                    Vecteur vtm = LigneA[p];
                    LigneA[p] = LigneA[--M];
                    LigneA[M] = vtm;
                }
            }
            if (c > -2) break;
            for(Indice ii=k;ii<Ni;++ii)
                for(Indice jj=LigneA[ii].size();jj--;)
                    LigneA[ii][jj].affect( LigneA[ii][jj].getvalue() / PRIME);
            MOD = MOD / PRIME;
            sprintf(UT,"Rank mod %ld : %ld", FMOD/MOD, indcol);
            activityReport(GlobalCommentator, UT, LVL_IMP, PARTIAL_RESULT );
//             if (MOD == 1) cerr << "wattadayada inhere ?" << endl;
        }
        if (p != k) {
            Vecteur vtm = LigneA[k];
            LigneA[k] = LigneA[p];
            LigneA[p] = vtm;
        }
        if (c != -1)
            for(Indice l=k + 1; l < Ni; ++l)
                FaireElimination(MOD, LigneA[l], LigneA[k], indcol, c, col_density);
    }
    CherchePivot( LigneA[last], indcol, c );
    
    delete [] LigneA;
        
    rank = indcol;
    stopActivity(GlobalCommentator, "", LVL_NEVER, INTERNAL_DESCRIPTION ); 
}


template<class BB>
void gauss_rank (UIndice& rank, const BB& SLA) {
    Indice Ni = SLA.n_row();
    typedef typename BB::coefficientSpace K;
    typedef Sparse_Vector< K > Vecteur;
    typedef typename Vecteur::elements E;    
    
    Vecteur * LigneA = new Vecteur[Ni];
    Indice jj;
    for(jj=0; jj<Ni; jj++) {
        const Vecteur& tmp = SLA[jj];
        Vecteur toto(tmp.size());
        Indice rs=0;
	Indice k=0;
        for(; k<tmp.size(); k++) {
            K r = tmp[k].getvalue();
            if (! iszero(r))
                toto[rs++] = E(tmp[k].j(), r); 
        }
        LigneA[jj] = Vecteur(rs);
        for(k=0; k<rs; k++)
            LigneA[jj][k] = toto[k];
    }
    
    Indice last = Ni-1;
    Indice c;
    Indice indcol(0);
    
    
    for (Indice k=0; k<last;k++) {
        CherchePivot( LigneA[k], indcol, c ) ;
        if (c != -1)
            for(Indice l=k + 1; l < Ni; l++)
                FaireElimination(LigneA[l], LigneA[k], indcol, c);
    }
    CherchePivot( LigneA[last], indcol, c );
    
    delete [] LigneA;
        
    rank = indcol;
}




template<class BB, class D>
void gauss_rank(UIndice& rank, const BB& SLA, const D& density_trait) {
    Indice Ni = SLA.n_row(), Nj = SLA.n_col();
    typedef typename BB::coefficientSpace K;
    typedef Sparse_Vector< K > Vecteur;
    typedef typename Vecteur::elements E;    
    
    D col_density(Nj);

    Vecteur * LigneA = new Vecteur[Ni];
    Indice jj;
    for(jj=0; jj<Ni; jj++) {
        const Vecteur& tmp = SLA[jj];
        Vecteur toto(tmp.size());
        Indice rs=0;
	Indice k=0;
        for(; k<tmp.size(); k++) {
            K r = tmp[k].getvalue();
            if (! iszero(r)) {
                col_density.incr(tmp[k].j());
                toto[rs++] = E(tmp[k].j(), r); 
            }
        }
        toto.resize(rs);
        LigneA[jj] = toto;
    }

    Indice last = Ni-1;
    Indice c;
    Indice indcol(0);
    
    for (Indice k=0; k<last;++k) {
        Indice l,p=k,s=LigneA[k].size(),sl;
        if (s) {
            for(l=k+1; l < Ni; ++l)
                if (((sl=LigneA[l].size()) < s) && (sl)) {
                    s = sl;
                    p = l;
                }
            if (p != k) {
                Vecteur vtm = LigneA[k];
                LigneA[k] = LigneA[p];
                LigneA[p] = vtm;
            }       
            CherchePivot( LigneA[k], indcol, c , col_density) ;
            if (c != -1)
                for(l=k + 1; l < Ni; ++l)
                    FaireElimination(LigneA[l], LigneA[k], indcol, c, col_density);
        }
    }
    CherchePivot( LigneA[last], indcol, c );
    
    delete [] LigneA;
        
    rank = indcol;
}

#endif
