#ifndef __LINBOX_INTEGER_SMITH_NORMAL_FORM_C__
#define __LINBOX_INTEGER_SMITH_NORMAL_FORM_C__
// ======================================================================= //
// Copyright (C) LinBox project 1999
// Givaro / Athapascan-1
// Smith form
// Time-stamp: <04 Aug 01 11:27:20 Jean-Guillaume.Dumas@imag.fr> 
// ======================================================================= //

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <timer.h>
#include <list>
using std::list;

#include "intfactor.C"
#include "inttcr.C"

#include <givtype.h>
#include <integer.h>
#include <intnumtheo.h>

#include "lin_valence.C"
#include "lin_gauss.C"
#include "givdensity.h"
#include "givinvariants.h"

template<class BBInt, class BBZpz>
void seq_Integer_smith(Invariants<ulong>& Inv, BBInt& TheMat, BBZpz& A, double OCB, double& wrong, double error = 0) {
    startActivity(GlobalCommentator, "Smith", "Smith Form Computation", LVL_NORMAL, INTERNAL_DESCRIPTION); 

//     Timer glob;glob.clear();
//     Timer loc; 

    UIndice rank;
    Integer valence;
    ulong nbfactor;
    list<Integer> Lm;
    list<ulong> Le;
    
// loc.clear();loc.start();
    seq_Integer_valence(valence, rank, TheMat, A, OCB, wrong, error);
    nbfactor = factorset(valence,Lm,Le);

	if (GlobalCommentator.printed(LVL_IMP, PARTIAL_RESULT)) {
		list<Integer>::const_iterator iLm = Lm.begin();
		list<ulong>::const_iterator iLe = Le.begin();
		cerr << "#I:R         Smith: Valence is ";
		for(; 1 ; ++iLm) {
			cerr << *iLm;
			if (*iLe > 1) cerr << "^" << *iLe;
			if (++iLe != Le.end() ) cerr << " * ";
			else break;
		}
		cerr << endl;
	} 
// loc.stop(); glob = loc;
// cerr << glob.usertime() << "\t| " << loc.usertime() << "\t: ";

    list<Integer>::iterator im=Lm.begin();
    list<ulong>::iterator ie=Le.begin();
//     if (GIVABS(valence) > 3) {
//         cerr << *im << "^" << *ie;
//         for(++im,++ie;im != Lm.end(); ++im, ++ie)
//             cerr << " * " << *im << "^" << *ie;
//         cerr << endl ;
//     } else
//         cerr << valence << endl;
    


    UIndice * r = new UIndice[nbfactor];
    UIndice ir;
    ulong ulim;
    Indice k, minr= GIVMIN(TheMat.n_row(),TheMat.n_col());
    Integer pr=1;
    for( k=0,im=Lm.begin(),ie=Le.begin();im != Lm.end(); ++im, ++ie,++k)
//        if (*ie > 1) {
        if (*ie > 0) {
// loc.clear();loc.start();        
            pr *= *im;
            if ( (ulim = *im%(ulong)MAXPRIMESIZE) == 2)
                modtwo_gauss_rank(TheMat,r[k],Density());
            else
                gauss_rank(ulim,TheMat,r[k],Density());
            minr = GIVMIN(minr,r[k]);
// loc.stop(); glob = glob + loc;
// cerr << glob.usertime() << "\t| " << loc.usertime() << "\t: ";
            char * UT = new char[70]; sprintf(UT,"rank mod %ld : %ld", (*im).Int2long(), r[k]);
            activityReport(GlobalCommentator, UT, LVL_NORMAL, PARTIAL_RESULT);
//             cerr << "rank mod " << *im << "\t: " << r[k]  << endl ;
        }
    
//    Integer p(2);
//    while ((iszero(pr % p)) && (! isprime(p)))
//        p = nextprime(p);
//    if ( (ulim = p.Int2long()) == 2)
//        modtwo_gauss_rank(TheMat,ir);
//    else
//        gauss_rank(p.Int2long(),TheMat,ir,Density());
    Integer p(2);
    while ((iszero(pr % p)) || (! isprime(p)))
        p = nextprime(p);
    if ( (ulim = p.Int2long()) == 2)
        modtwo_gauss_rank(TheMat,ir,Density());
    else
        gauss_rank(p.Int2long(),TheMat,ir,Density());
    char * UT = new char[150]; sprintf(UT,"Integer rank (mod %ld) : %ld, correct with probability at least 1 - %e", (p).Int2long(), ir, wrong);
    activityReport(GlobalCommentator, UT, LVL_NORMAL, PARTIAL_RESULT);

    if (ir > minr) {
        Indice ssize = ir-minr;
        ulong * Smith = new ulong[ssize];
        Indice ll=ssize;
        for (;ll--;)
            Smith[ll] = 1;
    

        UIndice l=0,l1=0;
        for( k=0,im=Lm.begin(),ie=Le.begin();im != Lm.end(); ++im, ++ie,++k)
//            if (*ie > 1) {
            if (*ie > 0) {
                l = r[k];
                ulong p = *im%(ulong)MAXPRIMESIZE;
                ulong pp = p;
                while (l < ir) {
// loc.clear();loc.start();
                    for(Indice lk=ir-l;lk;--lk)
                        Smith[ssize-lk] *= p;
                    pp *= p;
                    gauss_rank(pp,p,TheMat,l,Density());
// loc.stop(); glob = glob + loc;
// cerr << glob.usertime() << "\t| " << loc.usertime() << "\t: ";
            char * UT = new char[70]; sprintf(UT,"rank mod %ld : %ld", pp, l);
            activityReport(GlobalCommentator, UT, LVL_NORMAL, PARTIAL_RESULT);
//                     cerr << "rank mod " << pp << "\t: " << l  << endl ;
                }
            }

//         cerr << endl;
        Inv.PushInvariant(1,minr);
//         cerr << "Integer Smith form : " << 1 << ":" << minr;
        Indice lj=0;
        for(ll=0;ll<ssize;) {
//             cerr << " & " << Smith[ll] << ":";
            while ((++lj<ssize) && (Smith[lj] == Smith[ll])) {};
//             cerr << lj-ll;
            Inv.PushInvariant(Smith[ll],lj-ll);
            ll = lj;
        }
        delete [] Smith;
    
    } else {
        Inv.PushInvariant(1,ir);
//         cerr << "\nInteger Smith form : " << 1 << ":" << ir;
//         cerr << "Integer Smith form : " << 1 << ":" << ir;
    }
    
        
//     cerr << endl;
    delete [] r;
    stopActivity(GlobalCommentator, "Done.", LVL_NORMAL, INTERNAL_DESCRIPTION ); 
};


template<class BBInt, class BBZpz>
void gauss_rank_N_torsion(Invariants<ulong>& Inv, BBInt& TheMat, BBZpz& A, Integer IN) {
    startActivity(GlobalCommentator, "N-Torsion", "Rank N-Torsion Computation", LVL_NORMAL, INTERNAL_DESCRIPTION); 

    UIndice rank;
    Integer valence;
    ulong nbfactor;
    list<Integer> Lm;
    list<ulong> Le;
    
    nbfactor = factorset(IN,Lm,Le);

    list<Integer>::iterator im=Lm.begin();
    list<ulong>::iterator ie=Le.begin();


    UIndice * r = new UIndice[nbfactor];
    UIndice ir;
    ulong ulim;
    Indice k, dimr = GIVMIN(TheMat.n_row(),TheMat.n_col());
    Indice minr = dimr, maxr = 0;
    Integer pr=1;
    for( k=0,im=Lm.begin(),ie=Le.begin();im != Lm.end(); ++im, ++ie,++k)
        if (*ie > 0) {
            pr *= *im;
            if ( (ulim = *im%(ulong)MAXPRIMESIZE) == 2)
                modtwo_gauss_rank(TheMat,r[k],Density());
            else
                gauss_rank(ulim,TheMat,r[k],Density());
            minr = GIVMIN(minr,r[k]);
            maxr = GIVMAX(maxr,r[k]);
            char * UT = new char[70]; sprintf(UT,"rank mod %ld : %ld", (*im).Int2long(), r[k]);
            activityReport(GlobalCommentator, UT, LVL_NORMAL, PARTIAL_RESULT);
        }
        /*
    Integer p(2);
    while ((iszero(pr % p)) && (! isprime(p)))
        p = nextprime(p);
    if ( (ulim = p.Int2long()) == 2)
        modtwo_gauss_rank(TheMat,ir);
    else
        gauss_rank(p.Int2long(),TheMat,ir,Density());
    char * UT = new char[70]; sprintf(UT,"Integer rank (mod %ld) : %ld", (p).Int2long(), ir);
    activityReport(GlobalCommentator, UT, LVL_NORMAL, PARTIAL_RESULT);
        */
    ir = dimr;
    if (ir > minr) {
        Indice ssize = ir-minr;
        ulong * Smith = new ulong[ssize];
        Indice ll=ssize;
        for (;ll--;)
            Smith[ll] = 1;
    

        UIndice l=0,l1=0, count_ie;
        for( k=0,im=Lm.begin(),ie=Le.begin();im != Lm.end(); ++im, ++ie,++k)
            if (*ie > 0) {
                count_ie = 0;
                l = r[k];
                ulong p = *im%(ulong)MAXPRIMESIZE;
                ulong pp = p;
//                 while (l < ir) {
                while (1) {
                    for(Indice lk=ir-l;lk;--lk)
                        Smith[ssize-lk] *= p;
                    pp *= p;
                    if (++count_ie < *ie) {
                        gauss_rank(pp,p,TheMat,l,Density());
                        maxr = GIVMAX(maxr, l);
                    } else
                        break;
                    char * UT = new char[70]; sprintf(UT,"rank mod %ld : %ld", pp, l);
                    activityReport(GlobalCommentator, UT, LVL_NORMAL, PARTIAL_RESULT);
                }
            }

        Inv.PushInvariant(1,minr);
        Indice lj=0;
        ssize = GIVMIN(maxr, dimr) - minr;
        for(ll=0;ll<ssize;) {
            while ((++lj<ssize) && (Smith[lj] == Smith[ll])) {};
            Inv.PushInvariant(Smith[ll],lj-ll);
            ll = lj;
        }
        delete [] Smith;
    
    } else {
        Inv.PushInvariant(1,ir);
    }
    
        
    delete [] r;
    stopActivity(GlobalCommentator, "Done.", LVL_BLABLA, INTERNAL_DESCRIPTION ); 
};
#endif
