// ======================================================================= //
// Copyright (C)  1999, Linbox project
// Givaro / Athapascan-1
// Valence computation
// Time-stamp: <04 Aug 01 11:09:35 Jean-Guillaume.Dumas@imag.fr> 
// ======================================================================= //
#ifndef __LINBOX_VALENCE_C__
#define __LINBOX_VALENCE_C__

#include <math.h>
#define MAXPRIMESIZE 65536
#define MINLOWERSIZE 5000
#define valBOUND 5

// #include "../LinBox/Algorithms/LinAlgs/lin_algebra.h"
#include "lin_b_b_trans.h"      // Transposition of BlackBox
#include "lin_b_b_comp.h"       // Composition of BlackBox
#include "lin_wiedemann.C"      // Symetric Wiedemann

#include <list>
using std::list;
#include <intnumtheo.h>
#include <inttcr.C>


// template<class BB> typename BB::coefficientSpace& absolute_bound(typename BB::coefficientSpace& b, const BB& lA) {
//     startActivity(GlobalCommentator, "Gersh", "Cassini Ovals method, to create a bound on the coefficients of A^t . A", LVL_NORMAL, INTERNAL_DESCRIPTION ); 
//     BB abslA;
//     abslA.abshomomorphic(lA);
//     BBGershgorin< BB > lAdata(&abslA);
//     lAdata.makedata();
//     char * UT = new char[70]; sprintf(UT,"Cassini Ovals eigen value bound : %ld", lAdata.getbound()); 
//     stopActivity(GlobalCommentator, UT, LVL_IMP, PARTIAL_RESULT ); 
//     return b = lAdata.getbound();
// };    

template<class BBlong, class BBZpz>
void seq_Integer_valence(Integer& Valence, UIndice& rank, BBlong& lA, BBZpz& A, double bound, double& wrong, double error = 0) {

    startActivity(GlobalCommentator, "Int Vale", "Integer Valence", LVL_NORMAL, INTERNAL_DESCRIPTION ); 
    typedef typename BBZpz::coefficientSpace K;
    typedef typename BBZpz::PreferredInMatrix Right;
    typedef typename BBZpz::PreferredOutMatrix Left;
    typedef typename BBlong::coefficientSpace VField;

//    srand48(BaseTimer::seed());


//     VField bound;
//     absolute_bound(bound, lA);
//     cerr << "Gershgorin eigen value bound : " << bound << endl ;

    double ulogMM = 1/log(MAXPRIMESIZE - MINLOWERSIZE);
    long valence;
    K v1,v;
    UIndice r;

    unsigned long p1;
    do 
        p1 = prevprime(MAXPRIMESIZE - (ulong)lrand48()%(ulong)MINLOWERSIZE) % (ulong)MAXPRIMESIZE;
    while (! isprime(p1));
    
    NEWFIELD(p1,1);
    A.firsthomomorphic(lA);

    double bad;
    unsigned long nbloops;
    
    BBTransposition< BBZpz > TheTransposed(&A);
    list<Integer> Lr, Lm;
    Integer V0, prod;
    int diff_bound = 0;
    if (A.n_row() > A.n_col()) {
        Right VY(A.n_col());
        for(UIndice k=0; k<VY.size(); k++)
            GIVRANDOM(VY[k]);
        Left inter(A.n_row());
        BBComposition< BBTransposition< BBZpz >, BBZpz, Right > TheComposed(&TheTransposed,&A,&inter);
        valence_ssw(v1, rank, TheComposed, VY);
        char * UT = new char[70]; sprintf(UT,"Min Poly Degree : %ld", rank);
        activityReport(GlobalCommentator, UT, LVL_IMP, PARTIAL_RESULT);
// 	cerr << "Min Poly Degree : " << rank << endl;
        valence = v1.residu();
        Lr.push_back(valence);
        Lm.push_back(p1);
        double maxsize = (double)rank * log(bound) * ulogMM;
	bad = (double)(rank+1)*double(rank+1) * log(bound) / log(MINLOWERSIZE) / 2.0;
	if ((error > 0) && (error < 1)) {
	    double maxerror = log ( error ) / log( bad / 3030 );
//cerr << "maxsize: " << maxsize << ", maxerror: " << maxerror << endl;
	    if (maxerror > maxsize) { 
	    	activityReport(GlobalCommentator, "Taking probability into account...", LVL_NORMAL, INTERNAL_DESCRIPTION); 
		maxsize = maxerror;
	    }
	}
        nbloops = (unsigned long)ceil(maxsize);
        if (nbloops > 6103) {
            Valence = Integer(1);
            activityReport(GlobalCommentator, "More than 6103 primes needed. Wanted Probability too high.", LVL_IMP, INTERNAL_ERROR);
        }

        if (rank & 0x1) {
            if (valence > 0)
                valence = valence - p1;
        } else
            if (valence < 0)
                valence = valence + p1;
        Valence = Integer(valence);

        for(int l=nbloops-1;l>0;--l) {
            progressReport(GlobalCommentator, "modular valence", LVL_NORMAL, nbloops-l, nbloops);
            do 
                p1=prevprime(p1) % (ulong)MAXPRIMESIZE;
            while (! isprime(p1));
            NEWFIELD(p1,1);
            A.homomorphic(lA);
            for(UIndice k=0; k<VY.size(); k++)
                GIVRANDOM(VY[k]);
            valence_ssw(v, r, TheComposed, VY);
            Lr.push_back(Integer(v.residu()));
            Lm.push_back(Integer(p1));
            tcr(V0, prod, Lr,Lm);
            if (rank & 0x1) {
                if (V0 > 0)
                    V0 = V0 - prod;
            } else
                if (V0 < 0)
                    V0 = V0 + prod;
//            if (V0 == Valence) { if (++diff_bound > valBOUND) break; }
//		    else diff_bound = 0;
            Valence = V0;
        }
    } else {
        Left VY(A.n_row());
        for(UIndice k=0; k<VY.size(); k++)
            GIVRANDOM(VY[k]);
        Right inter(A.n_col());

        BBComposition< BBZpz, BBTransposition< BBZpz >, Right > TheComposed(&A,&TheTransposed,&inter);
    
        valence_ssw(v1, rank, TheComposed, VY);
        char * UT = new char[70]; sprintf(UT,"Min Poly Degree : %ld", rank);
        activityReport(GlobalCommentator, UT, LVL_IMP, PARTIAL_RESULT);
// 	cerr << "Min Poly Degree : " << rank << endl;
        valence = v1.residu();
        Lr.push_back(valence);
        Lm.push_back(p1);
        double maxsize = (double)rank * log(bound) * ulogMM;
	bad = (double)(rank+1)*double(rank+1) * log(bound) / log(MINLOWERSIZE) / 2.0;
	if ((error > 0) && (error < 1)) {
	    double maxerror = log ( error )/ log( bad / 3030 );
//cerr << "maxsize: " << maxsize << ", maxerror: " << maxerror << endl;
	    if (maxerror > maxsize) {
	        activityReport(GlobalCommentator, "Taking probability into account...", LVL_NORMAL, INTERNAL_DESCRIPTION); 
		maxsize = maxerror;
	    }
	}
        nbloops = (unsigned long)ceil(maxsize);
        if (nbloops > 6103) {
            Valence = Integer(1);
            activityReport(GlobalCommentator, "More than 6103 primes needed. Wanted Probability too high.", LVL_IMP, INTERNAL_ERROR);
        }


        if (rank & 0x1) {
            if (valence > 0)
                valence = valence - p1;
        } else
            if (valence < 0)
                valence = valence + p1;
        Valence = Integer(valence);

        for(int l=nbloops-1;l>0;--l) {
            progressReport(GlobalCommentator, "modular valence", LVL_NORMAL, nbloops-l, nbloops);
            
            do 
                p1=prevprime(p1) % MAXPRIMESIZE;
            while (! isprime(p1));
            NEWFIELD(p1,1);
            A.homomorphic(lA);
            for(UIndice k=0; k<VY.size(); k++)
                GIVRANDOM(VY[k]);
            valence_ssw(v, r, TheComposed, VY);
            Lr.push_back(Integer(v.residu()));
            Lm.push_back(Integer(p1));
            tcr(V0, prod, Lr,Lm);
            if (rank & 0x1) {
                if (V0 > 0)
                    V0 = V0 - prod;
            } else
                if (V0 < 0)
                    V0 = V0 + prod;
//            if (V0 == Valence) if (++diff_bound > valBOUND) break;
            Valence = V0;
        }
   }
    wrong = pow( bad/3030, nbloops);
    char * UT= new char[500]; sprintf(UT,"Valence is %s, correct with probability at least 1 - %e", Valence.Int2str(), wrong ); 
    stopActivity( GlobalCommentator, UT, LVL_NORMAL, INTERNAL_DESCRIPTION) ;
    
};

// template<class BBlong, class BBZpz>
// void ata_Integer_minpoly(Poly1<Integer>& Phi, BBlong& lA, BBZpz& A) {

//     typedef typename BBZpz::coefficientSpace K;
//     typedef typename BBZpz::PreferredInMatrix Right;
//     typedef typename BBZpz::PreferredOutMatrix Left;
//     typedef typename BBlong::coefficientSpace VField;

// //    if (A.n_row() > A.n_col()) {
//         Left inter(A.n_row());
//         BBTransposition< BBZpz > TheTransposed(&A);
//         BBComposition< BBTransposition< BBZpz >, BBZpz, Right > TheComposed(&TheTransposed,&A,&inter);


//         VField bound;
//         absolute_bound(bound, lA);
//         cerr << "Gershgorin eigen value bound : " << bound << endl ;


// //        srand48(BaseTimer::seed());

//         double ulogMM = 1/log(MAXPRIMESIZE - MINLOWERSIZE);

//         vector<long> phil;Poly1<K> phiz;
//         K v1,v;
//         UIndice r,rank;

//         unsigned long p1;
//         do 
//             p1 = prevprime(MAXPRIMESIZE - (ulong)lrand48()%(ulong)MINLOWERSIZE) % (ulong)MAXPRIMESIZE;
//         while (! isprime(p1));
    
//         NEWFIELD(p1,1);
//         A.firsthomomorphic(lA);
    
//         Integer prod;
    
//         Right VY(A.n_col());
//         for(UIndice k=0; k<VY.size(); k++)
//             GIVRANDOM(VY[k]);
//         minpoly_ssw(phiz, TheComposed, VY);
// 	cerr << "Min Poly Degree : " << (rank=phiz.degree()) << endl;
//         phil = vector<long>(rank+1);
//         list< Integer > Lm;
//         Lm.push_back(p1);
//         double maxsize = (double)rank * log(bound) * ulogMM;
//         unsigned long nbloops = (unsigned long)ceil(maxsize);
// //         cerr << "maxsize: " << maxsize << "  nbloops:" << nbloops << endl;
//         Phi = Poly1<Integer>(phiz.indeter(),Degree(rank));

//         for(Indice ui=rank;ui>=0;--ui) {
//             phil[ui] = phiz[ui].residu();
//             if ((rank+ui) & 0x1) {
//                 if (phil[ui] > 0)
//                     phil[ui] = phil[ui] - p1;
//             } else
//                 if (phil[ui] < 0)
//                     phil[ui] = phil[ui] + p1;
//             Phi[ui] = Integer( phil[ui] );
//         }
        
//         if (nbloops > 1) {

//             vector< list<Integer> > Lr(rank+1);
//             for(Indice ui=rank;ui>=0;--ui) Lr[ui].push_back(Phi[ui]);
//             Poly1<Integer> V0(phiz.indeter(),Degree(rank));
// //            for(int l=nbloops-1;l>0;--l) {
//             for(int l=0;l<(nbloops-1);++l) {
//                 do 
//                     p1=prevprime(p1) % (ulong)MAXPRIMESIZE;
//                 while (! isprime(p1));
//                 NEWFIELD(p1,1);
//                 A.homomorphic(lA);
//                 for(UIndice k=0; k<VY.size(); k++)
//                     GIVRANDOM(VY[k]);
//                 minpoly_ssw(phiz, TheComposed, VY);
//                 Lm.push_back(Integer(p1));
// //                 cerr << "primes : ";
// //                 list<Integer>::iterator le=Lm.begin();
// //                 for(;le!=Lm.end();++le) 
// //                     cerr << *le << " "; 
// //                 cerr << endl;
//                 for(Indice ui=rank;ui>=0;--ui) {
//                     Lr[ui].push_back(Integer(phiz[ui].residu()));
// //                     list<Integer>::iterator ll=Lr[ui].begin();
// //                     for(;ll!=Lr[ui].end();++ll) 
// //                         cerr << *ll << " ";
//                     tcr(V0[ui], prod, Lr[ui],Lm);
// //                     cerr << " ---> " << V0[ui] << " mod " << prod << endl;
//                     if ((rank+ui) & 0x1) {
//                         if (V0[ui] > 0)
//                             V0[ui] = V0[ui] - prod;
//                     } else
//                         if (V0[ui] < 0)
//                             V0[ui] = V0[ui] + prod;
//                 }
//             }
//             for(Indice ui=rank;ui>=0;--ui) Phi[ui] = V0[ui];
//        }
        


// //     } else {
// //         Left VY(A.n_row());
// //         for(UIndice k=0; k<VY.size(); k++)
// //             GIVRANDOM(VY[k]);
// //         Right inter(A.n_col());

// //         BBComposition< BBZpz, BBTransposition< BBZpz >, Right > TheComposed(&A,&TheTransposed,&inter);
    
// //         valence_ssw(v1, rank, TheComposed, VY);
// // 	cerr << "Min Poly Degree : " << rank << endl;
// //         valence = v1.residu();
// //         Lr.push_back(valence);
// //         Lm.push_back(p1);
// //         double maxsize = (double)rank * log(lAdata.getbound()) * ulogMM;
// //         unsigned long nbloops = (unsigned long)ceil(maxsize);

// //         if (rank & 0x1) {
// //             if (valence > 0)
// //                 valence = valence - p1;
// //         } else
// //             if (valence < 0)
// //                 valence = valence + p1;
// //         Valence = Integer(valence);

// //         for(int l=nbloops-1;l>0;--l) {
// //             do 
// //                 p1=prevprime(p1) % MAXPRIMESIZE;
// //             while (! isprime(p1));
// //             NEWFIELD(p1,1);
// //             A.homomorphic(lA);
// //             for(UIndice k=0; k<VY.size(); k++)
// //                 GIVRANDOM(VY[k]);
// //             valence_ssw(v, r, TheComposed, VY);
// //             Lr.push_back(Integer(v.residu()));
// //             Lm.push_back(Integer(p1));
// //             tcr(V0, prod, Lr,Lm);
// //             if (rank & 0x1) {
// //                 if (V0 > 0)
// //                     V0 = V0 - prod;
// //             } else
// //                 if (V0 < 0)
// //                     V0 = V0 + prod;
// //             if (V0 == Valence) if (++diff_bound > valBOUND) break;
// //             Valence = V0;
// //         }
// //    }
// };



template<class BBZpz>
void seq_Integer_valence(char * File_Name, Integer& Valence, UIndice& rank, BBZpz& A, double bound, double& wrong, double error = 0) {

    startActivity(GlobalCommentator, "Int Vale", "Integer Valence", LVL_NORMAL, INTERNAL_DESCRIPTION ); 
    typedef typename BBZpz::coefficientSpace K;
    typedef typename BBZpz::PreferredInMatrix Right;
    typedef typename BBZpz::PreferredOutMatrix Left;


    double ulogMM = 1/log(MAXPRIMESIZE - MINLOWERSIZE);
    long valence;
    K v1,v;
    UIndice r;

    unsigned long p1;
    do 
        p1 = prevprime(MAXPRIMESIZE - (ulong)lrand48()%(ulong)MINLOWERSIZE) % (ulong)MAXPRIMESIZE;
    while (! isprime(p1));
    
    NEWFIELD(p1,1);
    long ni = A.get_n_row(File_Name);
    long nj = A.n_col();
    if (ni > nj)
        A.fileset_transpose(File_Name);
    else
        A.fileset(File_Name);

    double bad;
    unsigned long nbloops;
    
    BBTransposition< BBZpz > TheTransposed(&A);
    list<Integer> Lr, Lm;
    Integer V0, prod;
    int diff_bound = 0;
    if (A.n_row() > A.n_col()) {
        Right VY(A.n_col());
        for(UIndice k=0; k<VY.size(); k++)
            GIVRANDOM(VY[k]);
        Left inter(A.n_row());
        BBComposition< BBTransposition< BBZpz >, BBZpz, Right > TheComposed(&TheTransposed,&A,&inter);
        valence_ssw(v1, rank, TheComposed, VY);
        char * UT = new char[70]; sprintf(UT,"Min Poly Degree : %ld", rank);
        activityReport(GlobalCommentator, UT, LVL_IMP, PARTIAL_RESULT);
// 	cerr << "Min Poly Degree : " << rank << endl;
        valence = v1.residu();
        Lr.push_back(valence);
        Lm.push_back(p1);
        double maxsize = (double)rank * log(bound) * ulogMM;
	bad = (double)(rank+1)*double(rank+1) * log(bound) / log(MINLOWERSIZE) / 2.0;
	if ((error > 0) && (error < 1)) {
	    double maxerror = log ( error ) / log( bad / 3030 );
//cerr << "maxsize: " << maxsize << ", maxerror: " << maxerror << endl;
	    if (maxerror > maxsize) { 
	    	activityReport(GlobalCommentator, "Taking probability into account...", LVL_NORMAL, INTERNAL_DESCRIPTION); 
		maxsize = maxerror;
	    }
	}
        nbloops = (unsigned long)ceil(maxsize);
        if (nbloops > 6103) {
            Valence = Integer(1);
            activityReport(GlobalCommentator, "More than 6103 primes needed. Wanted Probability too high.", LVL_IMP, INTERNAL_ERROR);
        }

        if (rank & 0x1) {
            if (valence > 0)
                valence = valence - p1;
        } else
            if (valence < 0)
                valence = valence + p1;
        Valence = Integer(valence);

        for(int l=nbloops-1;l>0;--l) {
            progressReport(GlobalCommentator, "modular valence", LVL_NORMAL, nbloops-l, nbloops);
            do 
                p1=prevprime(p1) % (ulong)MAXPRIMESIZE;
            while (! isprime(p1));
            NEWFIELD(p1,1);
            A.fileset_transpose(File_Name);
//             A.homomorphic(lA);
            for(UIndice k=0; k<VY.size(); k++)
                GIVRANDOM(VY[k]);
            valence_ssw(v, r, TheComposed, VY);
            Lr.push_back(Integer(v.residu()));
            Lm.push_back(Integer(p1));
            tcr(V0, prod, Lr,Lm);
            if (rank & 0x1) {
                if (V0 > 0)
                    V0 = V0 - prod;
            } else
                if (V0 < 0)
                    V0 = V0 + prod;
//            if (V0 == Valence) if (++diff_bound > valBOUND) break;
            Valence = V0;
        }
    } else {
        Left VY(A.n_row());
        for(UIndice k=0; k<VY.size(); k++)
            GIVRANDOM(VY[k]);
        Right inter(A.n_col());

        BBComposition< BBZpz, BBTransposition< BBZpz >, Right > TheComposed(&A,&TheTransposed,&inter);
    
        valence_ssw(v1, rank, TheComposed, VY);
        char * UT = new char[70]; sprintf(UT,"Min Poly Degree : %ld", rank);
        activityReport(GlobalCommentator, UT, LVL_IMP, PARTIAL_RESULT);
// 	cerr << "Min Poly Degree : " << rank << endl;
        valence = v1.residu();
        Lr.push_back(valence);
        Lm.push_back(p1);
        double maxsize = (double)rank * log(bound) * ulogMM;
	bad = (double)(rank+1)*double(rank+1) * log(bound) / log(MINLOWERSIZE) / 2.0;
	if ((error > 0) && (error < 1)) {
	    double maxerror = log ( error )/ log( bad / 3030 );
//cerr << "maxsize: " << maxsize << ", maxerror: " << maxerror << endl;
	    if (maxerror > maxsize) {
	        activityReport(GlobalCommentator, "Taking probability into account...", LVL_NORMAL, INTERNAL_DESCRIPTION); 
		maxsize = maxerror;
	    }
	}
        nbloops = (unsigned long)ceil(maxsize);
        if (nbloops > 6103) {
            Valence = Integer(1);
            activityReport(GlobalCommentator, "More than 6103 primes needed. Wanted Probability too high.", LVL_IMP, INTERNAL_ERROR);
        }


        if (rank & 0x1) {
            if (valence > 0)
                valence = valence - p1;
        } else
            if (valence < 0)
                valence = valence + p1;
        Valence = Integer(valence);

        for(int l=nbloops-1;l>0;--l) {
            progressReport(GlobalCommentator, "modular valence", LVL_NORMAL, nbloops-l, nbloops);
            
            do 
                p1=prevprime(p1) % MAXPRIMESIZE;
            while (! isprime(p1));
            NEWFIELD(p1,1);
            A.fileset(File_Name);
//             A.homomorphic(lA);
            for(UIndice k=0; k<VY.size(); k++)
                GIVRANDOM(VY[k]);
            valence_ssw(v, r, TheComposed, VY);
            Lr.push_back(Integer(v.residu()));
            Lm.push_back(Integer(p1));
            tcr(V0, prod, Lr,Lm);
            if (rank & 0x1) {
                if (V0 > 0)
                    V0 = V0 - prod;
            } else
                if (V0 < 0)
                    V0 = V0 + prod;
//            if (V0 == Valence) if (++diff_bound > valBOUND) break;
            Valence = V0;
        }
   }
    wrong = pow( bad/3030, nbloops);
    char * UT= new char[500]; sprintf(UT,"Valence is %s, correct with probability at least 1 - %e", Valence.Int2str(), wrong ); 
    stopActivity( GlobalCommentator, UT, LVL_NORMAL, INTERNAL_DESCRIPTION) ;
    
};

#endif
