// ======================================================================= //
// Copyright (C)  Linbox project 1999
// Wiedemann
// Multiply computations are stopped when the polynomials remains the same
// for more than EARLY_TERM_THRESOLD
// Time-stamp: <02 Aug 00 11:04:01 Jean-Guillaume.Dumas@imag.fr> 
// ======================================================================= //
#ifndef _LIN_WIEDEMANN_C_
#define _LIN_WIEDEMANN_C_

#include <poly1.h>
#include <givrandom.h>

#define EARLY_TERM_THRESHOLD 20

// -----------------------------------------------
// dot product
// Preconditions : u.size() > 0, v.size() > 0 ....
// -----------------------------------------------
template <class T, class A1, class A2>
void DOTPROD(T& coeff,const A1& u, const A2& v) {
                coeff = u[0] * v[0]; 
                for (Indice k=(u).size()-1 ; k>0; --k)
                    coeff.automuladd( u[k], v[k]);
}

template<class Polynome, class BB>
void massey(Polynome& C, BB& A, typename BB::PreferredInMatrix& u, bool is_symetric) { 
    startActivity(GlobalCommentator, "Massey", "Wiedemann's algorithm", LVL_BLABLA, INTERNAL_DESCRIPTION ); 
    typedef typename BB::PreferredOutMatrix PrefVect;
    typedef typename BB::coefficientSpace T; T Zero(zero); T One(one);

    const Indice ni = A.n_row(), nj = u.size();
    const Indice n = GIVMIN(ni,nj);
    const Indice END = 2*n;

// ====================================================
// Sequence initialization
    PrefVect S(END+1);
    T dots;
    PrefVect v(ni);

    Indice s = -1;

    
    DOTPROD(dots,u,u); S[0] = dots;
    s = 0;
    if (iszero(S[0])) {
        activityReport(GlobalCommentator, "Wiedemann's algorithm Failure : self-orthogonal vector, please try again.", LVL_IMP, INTERNAL_ERROR);
//         cerr << "Wiedemann's algorithm Failure : self-orthogonal vector, please try again." << endl;
        return;
    }
// ====================================================
    

    Indeter X("X");
    C = Polynome(X, Degree(0), T(one));
    Polynome B(X, Degree(0), T(one));

    Indice L=0;
    T b = One, d, Ds;
    Degree x = 1, b_deg = 0, c_deg = 0, l_deg;

    for (Indice N=0; N<END; ++N) {
        if (! (N % 1000) )
            progressReport(GlobalCommentator, "matrix-vector products", LVL_NORMAL, N, END);
        d = S[N];
        for(Indice i=GIVMIN(L,c_deg);i;--i) d += C[i]*S[N-i];
        if (iszero(d)) {
            ++x;
        } else {
            if (L > (N>>1)) {
//                 C = C + (Polynome(X,x,-d/b) * B);
                Ds = -d/b;
                l_deg = GIVMAX(c_deg,x+b_deg);
                if (l_deg > c_deg) {
                    Polynome TD(X,l_deg);
                    Indice i = l_deg;
                    if (x>c_deg) {
                        for(;i>=x;--i)
                            TD[i] = Ds * B[i-x];
                        for(;i>c_deg;--i)
                            TD[i] = Zero;
                    } else {
                        for(;i>c_deg;--i)
                            TD[i] = Ds * B[i-x];
                        for(;i>=x;--i)
                            TD[i].muladd(Ds,B[i-x],C[i]);
                    }
                    for(;i>=0;--i)
                        TD[i] = C[i];
                    C = TD;
                } else {
                    for(Indice i=b_deg;i>=0;--i)
                        C[i+x].automuladd(Ds,B[i]);
                }
                c_deg = C.degree();
                ++x;
            } else {
                Polynome TC(X,c_deg,C[c_deg]);
                for (Indice i=c_deg;i--;) TC[i] = C[i];
//                 C = C + (Polynome(X,x,-d/b) * B);
                Ds = -d/b;
                l_deg = GIVMAX(c_deg,x+b_deg);
                if (l_deg > c_deg) {
                    Polynome TD(X,l_deg);
                    Indice i = l_deg;
                    if (x>c_deg) {
                        for(;i>=x;--i)
                            TD[i] = Ds * B[i-x];
                        for(;i>c_deg;--i)
                            TD[i] = Zero;
                    } else {
                    for(;i>c_deg;--i)
                        TD[i] = Ds * B[i-x];
                    for(;i>=x;--i)
                        TD[i].muladd(Ds,B[i-x],C[i]);
                    }
                    for(;i>=0;--i)
                        TD[i] = C[i];
                    C = TD; 
                } else {
                    for(Indice i=b_deg;i>=0;--i)
                        C[i+x].automuladd(Ds,B[i]);
                }
                L = N+1-L;
                B = TC; 
                b_deg = c_deg;
                c_deg = C.degree();
                b = d;
                x = 1;
            }
        }
        if (x > EARLY_TERM_THRESHOLD) break; 
// ====================================================
// Next coefficient
        if (is_symetric) {
            if (N & 0x1) {
                u = v;
                DOTPROD(dots,u,u); S[++s] = dots;
            } else {
                A.Apply(v,u);
                DOTPROD(dots,v,u); S[++s] = dots;
            }
        } else {
            if (N & 0x1) {
                A.ApplyTrans( u, v); 
                DOTPROD(dots,u,u); S[++s] = dots;
            } else {
                A.Apply(v,u);
                DOTPROD(dots,v,v); S[++s] = dots;
            }
        }
// ====================================================
    }
    stopActivity(GlobalCommentator, "", LVL_NEVER, INTERNAL_DESCRIPTION ); 
}



// ---------------------------------------------
// Symetric Wiedemann

template<class BB>
void rank_ssw(ulong& rank, BB& A, typename BB::PreferredInMatrix& u) {
    typedef typename BB::coefficientSpace T;
    Poly1<T> phi;
    massey(phi, A, u, 1);
    rank = phi.degree() - phi.val();
}

template<class BB>
void seq_rank_ssw(ulong& rank, BB& A) {
    typedef typename BB::coefficientSpace T;
    typedef typename BB::PreferredInMatrix Right;
//    srand48(BaseTimer::seed());
    Right VY(A.n_col());
    for(Indice k=0; k<VY.size(); k++)
        GIVRANDOM(VY[k]);
    Poly1<T> phi;
    massey(phi, A, VY, 1);
    rank = phi.degree() - phi.val();
}
template<class BB>
void valence_ssw(typename BB::coefficientSpace& valence, ulong& rank, BB& A, typename BB::PreferredInMatrix& u) {
    typedef typename BB::coefficientSpace T;
    Poly1<T>  phi;
    massey(phi, A, u, 1);
    valence = phi.leadcoef() / phi.tcoef() ;
    rank = phi.degree() - phi.val();
}

template<class BB>
void minpoly_ssw(Poly1< typename BB::coefficientSpace > & phi, BB& A, typename BB::PreferredInMatrix& u) {
    massey(phi, A, u, 1);
    if (! iszero(phi)) {
        typedef typename BB::coefficientSpace T;
        T trail = phi.tcoef();
        Degree dp = phi.degree();
        for(Degree i = dp >> 1;i > 0; --i) {
            phi[0] = phi[i] / trail;
            phi[i] = phi[dp-i] / trail;
            phi[dp-i] = phi[0];
        }
        phi[0] = phi[dp] / trail;
        phi[dp] = T(one);
    }
}

    
// ---------------------------------------------
// Wiedemann


template<class BB>
void rank_sw(ulong& rank, BB& A, typename BB::PreferredInMatrix& u) {
    typedef typename BB::coefficientSpace T;
    Poly1<T> phi;
    massey(phi, A, u, 0);
    rank = phi.degree() - phi.val();
}

template<class BB>
void valence_sw(typename BB::coefficientSpace& valence, ulong& rank, BB& A, typename BB::PreferredInMatrix& u) {
    typedef typename BB::coefficientSpace T;
    Poly1<T> phi;
    massey(phi, A, u, 0);
    valence = phi.leadcoef() / phi.tcoef() ;
    rank = phi.degree() - phi.val();
}

template<class BB>
void minpoly_sw(Poly1< typename BB::coefficientSpace >& phi, BB& A, typename BB::PreferredInMatrix& u) {
    massey(phi, A, u, 0);
    typedef typename BB::coefficientSpace T;
    T trail = phi.tcoef();
    Degree dp = phi.degree();
    for(Degree i = dp >> 1;i > 0; --i) {
        phi[0] = phi[i] / trail;
        phi[i] = phi[dp-i] / trail;
        phi[dp-i] = phi[0];
    }
    phi[0] = phi[dp] / trail;
    phi[dp] = T(one);
}

template<class BB>
void precondition(BB& A) {
    typedef typename BB::PreferredInMatrix Right;
    typedef typename BB::PreferredOutMatrix Left;
    Indice ll=0;
//    srand48(BaseTimer::seed());
    Left diag_left(A.n_row());
    for(; ll<A.n_row(); ll++)
        NONZEROGIVRANDOM(diag_left[ll]);
 
    Right diag_right(A.n_col());
    for(ll=0;ll<A.n_col(); ll++)
        NONZEROGIVRANDOM(diag_right[ll]);

    A.rank_precondition(diag_left,diag_right);
}

template<class BB>
void seq_rank_sw(UIndice& rank, BB& A) {
    typedef typename BB::PreferredInMatrix Right;
//    srand48(BaseTimer::seed());
    Right VY(A.n_col());
    for(Indice k=0; k<VY.size(); k++)
        GIVRANDOM(VY[k]);
    rank_sw(rank, A, VY);
};

    
#endif
