#ifndef __B_B_COMPOSITION_H__
#define __B_B_COMPOSITION_H__

// ======================================================
// Copyright (C)  LinBox Project 1999
//
// Black Blox Composition
// An example of generic algorithm
// file : lin_b_b_comp.h
// ======================================================

template<class BBA, class BBB, class Intermediate>
class BBComposition {
private:
        // Composition is  BBA x BBB
    BBA * _amat;
    BBB * _bmat;
    Intermediate * _vect;

public:
    BBComposition() {};
    
    BBComposition(BBA* a, BBB* b, Intermediate* v) : _amat(a), _bmat(b), _vect(v) {
            // Check sizes
        if ((_amat->n_col() != _bmat->n_row()) || (_amat->n_col() != _vect->size()))
            cerr << "Exception : matrices are not of same size\n"
                 << _amat->n_row() << "x" << _amat->n_col() << " times " <<  _vect->size() << " times " << _bmat->n_row() << "x" << _bmat->n_col() << endl;
            // Check coefficient spaces
        typename BBA::coefficientSpace xxx( typename BBB::coefficientSpace() );
    };
    
    long n_row() const { return _amat->n_row();};
    long n_col() const { return _bmat->n_col();};

    typedef typename BBB::PreferredInMatrix PreferredInMatrix;
    typedef typename BBA::PreferredOutMatrix PreferredOutMatrix;
        // Coefficient space, BBA's for instance
    typedef typename BBA::coefficientSpace coefficientSpace;

    template<class OutMatrix, class InMatrix>
    inline OutMatrix& Apply(OutMatrix& outM,  const InMatrix& inM ) {
            // Need the preferred intermediate matrix type 
            // allocate intermediate with the good size
        return _amat->Apply (outM, _bmat->Apply ( *_vect, inM) );
    }
    
    template<class OutMatrix, class InMatrix>
    inline OutMatrix& ApplyTrans(OutMatrix& outM,  const InMatrix& inM ) {
            // Compiles if BBA and BBB have ApplyTrans, or if no call to ApplyTrans is made
//         return ApplyTrans_specialist( typename BBA::BB_info (), typename BBB::BB_info (), outM, inM);
//     }
    
// private:
//     template<class OutMatrix, class InMatrix>
//     inline OutMatrix& ApplyTrans_specialist(BB_with_transpose bba_info, BB_with_transpose bbb_info, OutMatrix& outM,  const InMatrix& inM ) {
//             // Need the preferred intermediate matrix type 
        return _bmat->ApplyTrans (outM, _amat->ApplyTrans ( *_vect, inM) ) ;
    }

    template<class Field, class OutMatrix, class InMatrix>
    inline OutMatrix& Apply(const Field& F, OutMatrix& outM,  const InMatrix& inM ) {
            // Need the preferred intermediate matrix type 
            // allocate intermediate with the good size
        return _amat->Apply (F, outM, _bmat->Apply (F, *_vect, inM) );
    }
    
    template<class Field, class OutMatrix, class InMatrix>
    inline OutMatrix& ApplyTrans(const Field& F, OutMatrix& outM,  const InMatrix& inM ) {
        return _bmat->ApplyTrans (F, outM, _amat->ApplyTrans (F, *_vect, inM) ) ;
    }

    template<class OutMatrix, class InMatrix, class M>
    inline OutMatrix& Apply(OutMatrix& outM, const InMatrix& inM, M m, M p ) {
            // Need the preferred intermediate matrix type 
            // allocate intermediate with the good size
        return _amat->Apply (outM, _bmat->Apply (*_vect, inM, m, p), m, p );
    }
    
    template<class OutMatrix, class InMatrix, class M>
    inline OutMatrix& ApplyTrans(OutMatrix& outM, const InMatrix& inM, M m, M p ) {
        return _bmat->ApplyTrans (outM, _amat->ApplyTrans (*_vect, inM, m, p), m, p ) ;
    }

};



/***************************************************
 * Specialisation of BlackBox for BBComposition
 * As a BBComposition matches BlackBox requirements
 * this is just to provide BlackBox< BBComposition >
 ***************************************************/
#include "lin_b_b.h"

template<class BBA, class BBB, class III>
class BlackBox< BBComposition<BBA,BBB, III> > : public BBComposition<BBA,BBB, III> {
    BlackBox() {}
    BlackBox(BBA* a, BBB* b, III* v) : BBComposition<BBA,BBB,III>(a,b,v) {}
};


#endif

    
            
