#ifndef __B_B_TRANSPOSE_H_
#define __B_B_TRANSPOSE_H_

// ======================================================
// Copyright (C)  LinBox Project 1999
//
// Black Blox Transposition
// An example of generic algorithm
// file : lin_b_b_comp.h
// ======================================================

template<class BBA>
class BBTransposition {
private:
    BBA * _amat;

public:
    BBTransposition() {};
    
    BBTransposition(BBA* a) : _amat(a){};
    
    long n_row() const { return _amat->n_col();};
    long n_col() const { return _amat->n_row();};

    typedef typename BBA::PreferredOutMatrix PreferredInMatrix;
    typedef typename BBA::PreferredInMatrix PreferredOutMatrix;
    typedef typename BBA::coefficientSpace coefficientSpace;

    template<class OutMatrix, class InMatrix>
    inline OutMatrix& Apply(OutMatrix& outM,  const InMatrix& inM ) {
        return _amat->ApplyTrans (outM, inM);
    }
    
    template<class OutMatrix, class InMatrix>
    inline OutMatrix& ApplyTrans(OutMatrix& outM,  const InMatrix& inM ) {
        return _amat->Apply (outM, inM);
    }

    template<class OutMatrix, class InMatrix, class M>
    inline OutMatrix& Apply(OutMatrix& outM,  const InMatrix& inM, M m, M p ) {
        return _amat->ApplyTrans (outM, inM, m, p);
    }
    
    template<class OutMatrix, class InMatrix, class M>
    inline OutMatrix& ApplyTrans(OutMatrix& outM,  const InMatrix& inM, M m, M p) {
        return _amat->Apply (outM, inM, m, p);
    }

    template<class Field, class OutMatrix, class InMatrix>
    inline OutMatrix& Apply(const Field& F, OutMatrix& outM,  const InMatrix& inM ) {
        return _amat->ApplyTrans (F, outM, inM);
    }
    
    template<class Field, class OutMatrix, class InMatrix>
    inline OutMatrix& ApplyTrans(const Field& F, OutMatrix& outM,  const InMatrix& inM ) {
        return _amat->Apply (F, outM, inM);
    }
};



/***************************************************
 * Specialisation of BlackBox for BBTransposition
 * As a BBTransposition matches BlackBox requirements
 * this is just to provide BlackBox< BBTransposition >
 ***************************************************/
#include "lin_b_b.h"

template<class BBA>
class BlackBox< BBTransposition<BBA> > : public BBTransposition<BBA> {
    BlackBox() {}
    BlackBox(BBA* a) : BBTransposition<BBA>(a) {}
};


#endif
