#ifndef __SPARSE_B_B_DOMAIN_H__
#define __SPARSE_B_B_DOMAIN_H__

// =========================================================
// Copyright (C)  LinBox Project 1999
//
// Common object Givaro Black Blox Matrices
// Linbox wrapper for Givaro sparse vectors
// file : lin_sparse_b_b.h
// =========================================================

#include <fstream.h> 
#include <sparse_vector.h>      // Givaro sparse vectors
#include <list>               // STL lists
using std::list;
#include <lin_b_b.h>            // BlackBox interface, BB_info types

// #include <g_vector.h>
// #define _IBB_VECTOR_ Vector

#include <vector>
using std::vector;
#ifndef _IBB_VECTOR_
#define _IBB_VECTOR_ std::vector
#endif

/*************************************************************
 * BB Givaro sparse matrices wrapper
 * Specialisation of BlackBox for _IBB_VECTOR_<Sparse_Vector<K> >
 *************************************************************/
template <class K>
class SparseBlackBoxStorage : public _IBB_VECTOR_ < Sparse_Vector < K > > {
private:
    Indice _col_dim;
public:
    typedef             K                               coefficientSpace_t;
    typedef             K                               Type_t;
    typedef             _IBB_VECTOR_<K>                 PreferredInMatrix_t;
    typedef             _IBB_VECTOR_<K>                 PreferredOutMatrix_t;

    typedef typename    Sparse_Vector<K>::elements      Element_t;
    typedef             Sparse_Vector<K>                Row_t;
    typedef             SparseBlackBoxStorage< K >      Self_t;

    SparseBlackBoxStorage() {}
    SparseBlackBoxStorage(Indice n) 
            : _IBB_VECTOR_ < Sparse_Vector < K > >(n),_col_dim(n) {}
    SparseBlackBoxStorage(Indice ni, Indice nj)
            : _IBB_VECTOR_ < Sparse_Vector < K > >(ni),_col_dim(nj) {}

    void init(char * P_File_Name) { fileset(P_File_Name); }
    void write(char * O_File_Name) const {
        FILE* FileDes = fopen(O_File_Name, "w");
        if (FileDes != 0) {
           fprintf(FileDes,"%ld %ld M\n",n_row(),n_col());      
           Element_t _entry;
           for (long i=0; i<n_row(); i++) 
             for (long j=0; j< (*this)[i].size(); j++) {            
               _entry=(*this)[i][j];    
               fprintf(FileDes,"%ld %ld %ld\n",i+1,_entry.getindex() +1,_entry.getvalue());
             }
           fprintf(FileDes,"%ld %ld %ld\n",0,0,0);           
        }
        fclose(FileDes);
    }
    
    void init_transpose(char * P_File_Name) { fileset_transpose(P_File_Name); } 
    
    long n_row() const { return size(); }
    long n_col() const { return _col_dim; }// don't know ...


    Indice get_n_row(char * P_File_Name) {
        Indice ni,nj;
	char *UT, *File_Name;
        int is_gzipped = 0;
	size_t s = strlen(P_File_Name);
	if ((P_File_Name[--s] == 'z') && (P_File_Name[--s] == 'g') && (P_File_Name[--s] == '.')) {
            is_gzipped = 1;
            File_Name = tempnam("/tmp","bbx_");
            UT = new char[s+34+strlen(File_Name)];
            sprintf(UT,"gunzip -c %s > %s", P_File_Name, File_Name);
            system(UT);
            sprintf(UT,"\\rm %s", File_Name);
	} else
            File_Name = P_File_Name;
        
        FILE* FileDes = fopen(File_Name, "r");
        if (FileDes != NULL) {
            fscanf(FileDes,"%ld %ld M\n",&ni, &nj) ;
        }
        fclose(FileDes);
        if (is_gzipped) {
            system(UT);
        }
        _col_dim = nj;
        return ni;
    }

private:
    void fileset(char * P_File_Name) {
	char *UT, *File_Name;
        int is_gzipped = 0;
	size_t s = strlen(P_File_Name);
	if ((P_File_Name[--s] == 'z') && (P_File_Name[--s] == 'g') && (P_File_Name[--s] == '.')) {
            is_gzipped = 1;
            File_Name = tempnam("/tmp","bbx_");
            UT = new char[s+34+strlen(File_Name)];
            sprintf(UT,"gunzip -c %s > %s", P_File_Name, File_Name);
            system(UT);
            sprintf(UT,"\\rm %s", File_Name);
	} else
            File_Name = P_File_Name;
        
        FILE* FileDes = fopen(File_Name, "r");
        if (FileDes != NULL) {
            Indice ni,nj;
  
            fscanf(FileDes,"%ld %ld M\n",&ni, &nj) ;
            *this = Self_t(ni,nj);
                // vecteur temporaire de taille nj trop grande ...
            Row_t v(nj);

            long i,j, val;
            fscanf(FileDes,"%ld %ld %ld\n",&i, &j, &val) ;
           
            for(Indice ii=0; ii<ni; ii++) {
                Indice k=0;
                v.resize(nj);
                while (i == (ii+1)) {
                    v[k++] = Element_t(j-1, Type_t(val) );
                    fscanf(FileDes,"%ld %ld %ld\n",&i, &j, &val) ;
                }
                v.resize(k);
                (*this)[ii] = v;
            }

            fclose(FileDes);
            if (is_gzipped) {
                system(UT);
            }
        }
    }

    void fileset_transpose(char * P_File_Name) {
	char *UT, *File_Name;
        int is_gzipped = 0;
	size_t s = strlen(P_File_Name);
	if ((P_File_Name[--s] == 'z') && (P_File_Name[--s] == 'g') && (P_File_Name[--s] == '.')) {
            is_gzipped = 1;
            File_Name = tempnam("/tmp","bbx_");
            UT = new char[s+34+strlen(File_Name)];
            sprintf(UT,"gunzip -c %s > %s", P_File_Name, File_Name);
            system(UT);
            sprintf(UT,"\\rm %s", File_Name);
	} else
            File_Name = P_File_Name;
        
        FILE* FileDes = fopen(File_Name, "r");
        if (FileDes != NULL) {
        
            Indice ni,nj;
  
            fscanf(FileDes,"%ld %ld M\n",&ni, &nj) ;
      
            list<Element_t> * TheT = new list<Element_t>[nj];
            
            long i,j,val;
            fscanf(FileDes,"%ld %ld %ld\n",&i, &j, &val) ;
            
            for(Indice ii=0; ii<ni; ii++)
                while (i == (ii+1)) {
                        TheT[j-1].push_back( Element_t(i-1, Type_t(val) ) );
                    fscanf(FileDes,"%ld %ld %ld\n",&i, &j, &val) ;
                }     
            
            *this = Self_t(nj,ni);
            for(Indice k=0; k<nj;k++) {
                Row_t w(TheT[k].size());
                list<Element_t>::iterator l=TheT[k].begin();
                Indice i=0;
                for (; l!=TheT[k].end(); ++l,++i)
                    w[i] = (*l);
                (*this)[k] = w;
            }
        }

        fclose(FileDes);
        if (is_gzipped) {
            system(UT);
        }
    }

};


template <class Domain>
class SparseBlackBoxDom {
public:
    typedef          Domain                                     Domain_t;
    typedef typename Domain::Rep                                Type_t;
    typedef          SparseBlackBoxStorage< Type_t >            Storage_t;
    typedef          Storage_t                                  Rep;
    typedef          SparseBlackBoxDom< Domain >                Self_t;
protected:
    typedef typename Storage_t::PreferredInMatrix_t             PreferredInMatrix_t;
    typedef typename Storage_t::PreferredOutMatrix_t            PreferredOutMatrix_t;
    typedef typename Storage_t::Element_t                       Element_t;
    typedef typename Storage_t::Row_t                           Row_t;

    Domain _domain;
public:
        // Transpose information
    typedef BB_with_transpose BB_info;

    
   
        //-- Default cstors:
    SparseBlackBoxDom() : _domain() {};
    SparseBlackBoxDom(const Domain& D) : _domain(D) {}
    
        //-- Cstor of recopy: compiler's generated
    SparseBlackBoxDom(const Self_t& M) : _domain(M._domain) {}

        //-- Usefull to use the same domain to perform other operations
    Domain getdomain() const { return _domain; }

        //-- initialization of a new object, with memory allocation
    void init(Rep& r) const {
        r = Storage_t();
    }
    
    void init(Rep& r, Indice n) const {
        r = Storage_t(n);
    }
    
    void init(Rep& r, Indice ni, Indice nj) const {
        r = Storage_t(ni,nj);
    }
    
    void init(Rep& r, char * P_File_Name) const {
        r.init(P_File_Name);
        assign(r);
    }
    
    void init_transpose(Rep& r, char * P_File_Name) const {
        r.init_transpose(P_File_Name);
        assign(r);
    }
    
    void init(Rep& r, const Rep& a) const {
        r = Storage_t(a.n_row(), a.n_col());
        assign(r,a);
    }

        //-- initialization of a new object, with limited memory allocation
        //-- r has been preallocated
        //-- only provided for convenience, should not be in there

        // Preconditions (some of them !):
        // -- r.n_col < a.n_col, r.n_row < a.n_row, forall i, r[ii].size < a[ii].size
        // -- Storage_t::Row_t has random access
    Rep& assign(Rep& r, const Rep& a) const {
        Indice ni=r.n_row(),nj=r.n_col();
        Type_t ktmp;
        for(Indice ii=0; ii<ni; ++ii) {
            Indice k = 0;
            for(Indice jj=0; jj<r[ii].size(); ++jj)
                if (! _domain.iszero( _domain.assign(ktmp, a[ii][jj].getvalue()) ) )
                    r[ii][k++].assign(a[ii][jj].getindex(), ktmp );
            r[ii].resize(k);
        }
        return r;
    }

    Rep& assign(Rep& r) const {
        Indice ni=r.n_row(),nj=r.n_col();
        Type_t ktmp;
        Row_t v;
        for(Indice ii=0; ii<ni; ++ii) {
            Indice k=0;
            for(Indice jj=0; jj<r[ii].size(); ++jj)
                if (! _domain.iszero( _domain.assign(ktmp, r[ii][jj].getvalue()) ) )
                    r[ii][k++].assign(r[ii][jj].getindex(), ktmp );
            r[ii].resize(k);
        }
        return r;
    }
            


        //-- BlackBox methods

    template<class OutMatrix, class InMatrix>
    OutMatrix& Apply(OutMatrix& res, const Rep& M, const InMatrix& vect ) {
        Element_t toto;
        Indice k,i;
        for(k=M.size(); k-- ; )
            for( res[k]=_domain.zero, i=M[k].size(); i-- ; ) {
                toto = M[k][i];
                _domain.axpyin(res[k],toto.getvalue(),vect[toto.getindex()]);
            }
        return res;
    }
    
    template<class OutMatrix, class InMatrix>
    OutMatrix& ApplyTrans(OutMatrix& res, const Rep& M, const InMatrix& vect ) {
        Element_t toto;
        Indice k,i;
        for(i=res.size(); i-- ; )
            res[i] = _domain.zero;
        for(k=M.size(); k-- ; )
            for(i=M[k].size(); i-- ; ) {
                toto = M[k][i];
                _domain.axpyin(res[ toto.getindex() ],toto.getvalue(), vect[k]);
            }
        return res;
    }


        //-- Special Methods

    template<class Left, class Right>
    void rank_precondition(const Left& l, Rep& M, const Right& r) {
        Type_t tmp;
        for(Indice ii=M.size()-1; ii>=0; --ii)
            for(Indice jj=M[ii].size()-1; jj>=0; --jj)
//                 M[ii][jj].affect( l[ii] * M[ii][jj].getvalue() * r[M[ii][jj].j()] );
                M[ii][jj].affect( _domain.mulin( _domain.mul(tmp, l[ii], M[ii][jj].getvalue()), r[M[ii][jj].getindex()] ) );
    }
            
    
};



#endif

