#ifndef __SPARSE_B_B_H__
#define __SPARSE_B_B_H__

// =========================================================
// Copyright (C)  LinBox Project 1999
//
// Common object Givaro Black Blox Matrices
// Linbox wrapper for Givaro sparse vectors
// file : lin_sparse_b_b.h
// =========================================================

#include <fstream.h> 
#include <sparse_vector.h>      // Givaro sparse vectors
#include "lin_b_b.h"
#include <list>               // STL lists
#include <vector>             // STL vector
using std::list;
using std::vector;
#ifndef _IBB_VECTOR
#define _IBB_VECTOR_ std::vector
#endif

/*************
 * Field trait
 *************/
class default_field {};
class muladd_field : public virtual default_field {};

template<class K> struct FT { typedef default_field field_info; };

/*************************************************************
 * BB Givaro sparse matrices wrapper
 * Specialisation of BlackBox for _IBB_VECTOR_<Sparse_Vector<K> >
 *************************************************************/
template<class K>
class BlackBox < _IBB_VECTOR_ < Sparse_Vector < K > > > 
    : public _IBB_VECTOR_ < Sparse_Vector < K > > {
public:
    typedef BB_with_transpose BB_info;
    typedef typename Sparse_Vector<K>::elements Elements;
    
    BlackBox() : _col_dim(0), Zero(zero){};

    BlackBox(Indice n) 
            : _IBB_VECTOR_ < Sparse_Vector < K > >(n),_col_dim(n),Zero(zero) {}    

    BlackBox(Indice ni, Indice nj) 
            : _IBB_VECTOR_ < Sparse_Vector < K > >(ni) , _col_dim(nj),Zero(zero) {}    

    BlackBox(char * P_File_Name) {
        fileset(P_File_Name);
    }

    Indice get_n_row(char * P_File_Name) {
        Indice ni,nj;
	char *UT, *File_Name;
        int is_gzipped = 0;
	size_t s = strlen(P_File_Name);
	if ((P_File_Name[--s] == 'z') && (P_File_Name[--s] == 'g') && (P_File_Name[--s] == '.')) {
            is_gzipped = 1;
            File_Name = tempnam("/tmp","bbx_");
            UT = new char[s+34+strlen(File_Name)];
            sprintf(UT,"gunzip -c %s > %s", P_File_Name, File_Name);
            system(UT);
            sprintf(UT,"\\rm %s", File_Name);
	} else
            File_Name = P_File_Name;
        
        FILE* FileDes = fopen(File_Name, "r");
        if (FileDes != NULL) {
            fscanf(FileDes,"%ld %ld M\n",&ni, &nj) ;
        }
//             std::ifstream FileDes(File_Name);
//             char tmp;
//             FileDes >> ni >> nj >> tmp ;
 
        fclose(FileDes);
//             FileDes.close();
        if (is_gzipped) {
            system(UT);
        }
        _col_dim = nj;
        return ni;
    }

    void fileset(char * P_File_Name) {
	char *UT, *File_Name;
        int is_gzipped = 0;
	size_t s = strlen(P_File_Name);
	if ((P_File_Name[--s] == 'z') && (P_File_Name[--s] == 'g') && (P_File_Name[--s] == '.')) {
            is_gzipped = 1;
            File_Name = tempnam("/tmp","bbx_");
            UT = new char[s+34+strlen(File_Name)];
            sprintf(UT,"gunzip -c %s > %s", P_File_Name, File_Name);
            system(UT);
            sprintf(UT,"\\rm %s", File_Name);
	} else
            File_Name = P_File_Name;
        
        FILE* FileDes = fopen(File_Name, "r");
        if (FileDes != NULL) {
            Indice ni,nj;
  
            fscanf(FileDes,"%ld %ld M\n",&ni, &nj) ;
//            *this = BlackBox(ni,nj);
            resize(ni); _col_dim = nj; 
                // vecteur temporaire de taille nj trop grande ...
            Sparse_Vector<K> v(nj);

            long i,j,val;
            fscanf(FileDes,"%ld %ld %ld\n",&i, &j, &val) ;
            Elements cour;
           
            for(Indice ii=0; ii<ni; ii++) {
                Indice k=0;
                v.resize(nj);
                while (i == (ii+1)) {
                    cour = Elements(j-1, K(val));
                    if (! iszero( cour.getvalue() )) 
                        v[k++] = cour;
                    fscanf(FileDes,"%ld %ld %ld\n",&i, &j, &val) ;
                }
                    // maintenant je connais la taille de mon vecteur ...
//                 Sparse_Vector<K> w(k);
//                 for (Indice l=0; l<k; l++)
//                     w[l] = v[l];
//                 (*this)[ ii ] = w;
                v.resize(k);
                (*this)[ii] = v;
            }

            fclose(FileDes);
            if (is_gzipped) {
                system(UT);
            }
        }
    }

    template<class M>
    void fileset(char * P_File_Name, M MOD) {
	char *UT, *File_Name;
        int is_gzipped = 0;
	size_t s = strlen(P_File_Name);
	if ((P_File_Name[--s] == 'z') && (P_File_Name[--s] == 'g') && (P_File_Name[--s] == '.')) {
            is_gzipped = 1;
            File_Name = tempnam("/tmp","bbx_");
            UT = new char[s+34+strlen(File_Name)];
            sprintf(UT,"gunzip -c %s > %s", P_File_Name, File_Name);
            system(UT);
            sprintf(UT,"\\rm %s", File_Name);
	} else
            File_Name = P_File_Name;
        
        FILE* FileDes = fopen(File_Name, "r");
        if (FileDes != NULL) {
            Indice ni,nj;
  
            fscanf(FileDes,"%ld %ld M\n",&ni, &nj) ;
//            *this = BlackBox(ni,nj);
            resize(ni); _col_dim = nj; 
                // vecteur temporaire de taille nj trop grande ...
            Sparse_Vector<K> v(nj);

            long i,j,val;
            fscanf(FileDes,"%ld %ld %ld\n",&i, &j, &val) ;
            Elements cour;
           
            for(Indice ii=0; ii<ni; ii++) {
                Indice k=0;
                v.resize(nj);
                while (i == (ii+1)) {
                    long tr = val;
                    if (tr<0) {
                        tr = -tr;
                        if (tr > MOD) tr = tr % MOD;
                        if (tr) val = MOD - tr;
                        else val = 0L;
                    } else
                        if (tr > MOD) val = tr % MOD;
                    cour = Elements(j-1, K(val));
                    if (cour.getvalue()) 
                        v[k++] = cour;
                    fscanf(FileDes,"%ld %ld %ld\n",&i, &j, &val) ;
                }
                v.resize(k);
                (*this)[ii] = v;
            }

            fclose(FileDes);
            if (is_gzipped) {
                system(UT);
            }
        }
    }

    template<class Field>
    void fileset(const Field& F, char * P_File_Name) {
	char *UT, *File_Name;
        int is_gzipped = 0;
	size_t s = strlen(P_File_Name);
	if ((P_File_Name[--s] == 'z') && (P_File_Name[--s] == 'g') && (P_File_Name[--s] == '.')) {
            is_gzipped = 1;
            File_Name = tempnam("/tmp","bbx_");
            UT = new char[s+34+strlen(File_Name)];
            sprintf(UT,"gunzip -c %s > %s", P_File_Name, File_Name);
            system(UT);
            sprintf(UT,"\\rm %s", File_Name);
	} else
            File_Name = P_File_Name;
        
        FILE* FileDes = fopen(File_Name, "r");
        if (FileDes != NULL) {
            Indice ni,nj;
  
            fscanf(FileDes,"%ld %ld M\n",&ni, &nj) ;
//            *this = BlackBox(ni,nj);
	    resize(ni); _col_dim = nj; 
                // vecteur temporaire de taille nj trop grande ...
            Sparse_Vector<K> v(nj);

            long i,j,val;
            fscanf(FileDes,"%ld %ld %ld\n",&i, &j, &val) ;
            Elements cour;
           
            for(Indice ii=0; ii<ni; ii++) {
                Indice k=0;
                v.resize(nj);
                while (i == (ii+1)) {
                    F.assign(val,val);
                    cour = Elements(j-1, val);
                    if (! F.iszero( cour.getvalue() )) 
                        v[k++] = cour;
                    fscanf(FileDes,"%ld %ld %ld\n",&i, &j, &val) ;
                }
                    // maintenant je connais la taille de mon vecteur ...
//                 Sparse_Vector<K> w(k);
//                 for (Indice l=0; l<k; l++)
//                     w[l] = v[l];
//                 (*this)[ ii ] = w;
                v.resize(k);
                (*this)[ii] = v;
            }

            fclose(FileDes);
            if (is_gzipped) {
                system(UT);
            }
        }
    }

    void fileset_transpose(char * P_File_Name) {
	char *UT, *File_Name;
        int is_gzipped = 0;
	size_t s = strlen(P_File_Name);
	if ((P_File_Name[--s] == 'z') && (P_File_Name[--s] == 'g') && (P_File_Name[--s] == '.')) {
            is_gzipped = 1;
            File_Name = tempnam("/tmp","bbx_");
            UT = new char[s+34+strlen(File_Name)];
            sprintf(UT,"gunzip -c %s > %s", P_File_Name, File_Name);
            system(UT);
            sprintf(UT,"\\rm %s", File_Name);
	} else
            File_Name = P_File_Name;
        
        FILE* FileDes = fopen(File_Name, "r");
        if (FileDes != NULL) {
        
            Indice ni,nj;
  
            fscanf(FileDes,"%ld %ld M\n",&ni, &nj) ;
      
            list<Elements> * TheT = new list<Elements>[nj];
            
            long i,j,val;
            fscanf(FileDes,"%ld %ld %ld\n",&i, &j, &val) ;
            Elements cour;
            
            for(Indice ii=0; ii<ni; ii++)
                while (i == (ii+1)) {
                    cour = Elements(i-1, K(val));
                    if (! iszero(cour.getvalue())) 
                        TheT[j-1].push_back(cour);
                    fscanf(FileDes,"%ld %ld %ld\n",&i, &j, &val) ;
                }     
            
//            *this = BlackBox(nj,ni);
	      resize(nj); _col_dim = ni;
           for(Indice k=0; k<nj;k++) {
                Sparse_Vector< K > w(TheT[k].size());
                list<Elements>::iterator l=TheT[k].begin();
                Indice i=0;
                for (; l!=TheT[k].end(); ++l,++i)
                    w[i] = (*l);
                (*this)[k] = w;
            }
        }

        fclose(FileDes);
        if (is_gzipped) {
            system(UT);
        }
    }

    template<class Field>
    void fileset_transpose(const Field& F, char * P_File_Name) {
	char *UT, *File_Name;
        int is_gzipped = 0;
	size_t s = strlen(P_File_Name);
	if ((P_File_Name[--s] == 'z') && (P_File_Name[--s] == 'g') && (P_File_Name[--s] == '.')) {
            is_gzipped = 1;
            File_Name = tempnam("/tmp","bbx_");
            UT = new char[s+34+strlen(File_Name)];
            sprintf(UT,"gunzip -c %s > %s", P_File_Name, File_Name);
            system(UT);
            sprintf(UT,"\\rm %s", File_Name);
	} else
            File_Name = P_File_Name;
        
        FILE* FileDes = fopen(File_Name, "r");
        if (FileDes != NULL) {
        
            Indice ni,nj;
  
            fscanf(FileDes,"%ld %ld M\n",&ni, &nj) ;
      
            list<Elements> * TheT = new list<Elements>[nj];
            
            long i,j,val;
            fscanf(FileDes,"%ld %ld %ld\n",&i, &j, &val) ;
            Elements cour;
            
            for(Indice ii=0; ii<ni; ii++)
                while (i == (ii+1)) {
                    F.assign(val,val);
                    cour = Elements(i-1, val);
                    if (! F.iszero(cour.getvalue())) 
                        TheT[j-1].push_back(cour);
                    fscanf(FileDes,"%ld %ld %ld\n",&i, &j, &val) ;
                }     
            
//            *this = BlackBox(nj,ni);
	   resize(nj); _col_dim = ni;
           for(Indice k=0; k<nj;k++) {
                Sparse_Vector< K > w(TheT[k].size());
                list<Elements>::iterator l=TheT[k].begin();
                Indice i=0;
                for (; l!=TheT[k].end(); ++l,++i)
                    w[i] = (*l);
                (*this)[k] = w;
            }
        }

        fclose(FileDes);
        if (is_gzipped) {
            system(UT);
        }
    }

    template<class Left, class Right>
    void rank_precondition(const Left& l, const Right& r) {
        for(Indice ii=(*this).size()-1; ii>=0; --ii)
            for(Indice jj=(*this)[ii].size()-1; jj>=0; --jj)
                (*this)[ii][jj].affect( l[ii] * (*this)[ii][jj].getvalue() * r[(*this)[ii][jj].j()] );
    }
            
    template<class Left, class Right, class M>
    void rank_precondition(const Left& l, const Right& r, M MOD, M PRIME) {
        for(Indice ii=(*this).size()-1; ii>=0; --ii)
            for(Indice jj=(*this)[ii].size()-1; jj>=0; --jj)
                (*this)[ii][jj].affect( ( ( (l[ii] * (*this)[ii][jj].getvalue())%MOD) * r[(*this)[ii][jj].j()] ) % MOD );
    }
            
    void save_mat(char * File_Name) {
        ofstream fic(File_Name);
        fic << size() << " " << _col_dim << " M\n";
        for(Indice ii=0; ii<(*this).size(); ++ii)
            for(Indice jj=0; jj<(*this)[ii].size(); ++jj)
                fic << ii+1 << " " << (*this)[ii][jj].j()+1 << " " << (*this)[ii][jj].getvalue() << "\n";
        fic << "0 0 0\n";
        fic.close();
   }
    
        
            
    template<class OtherField>
    BlackBox(const BlackBox < _IBB_VECTOR_ < Sparse_Vector < OtherField > > >& a) 
        {
//        *this = BlackBox(a.n_row(),a.n_col());
	resize(a.n_row());
        _col_dim = a.n_col();
       
        }
            
            


    template<class OtherField>
    void firsthomomorphic(const BlackBox < _IBB_VECTOR_ < Sparse_Vector < OtherField > > >& a) {
        Indice ni=a.n_row(),nj=a.n_col();
//        *this = BlackBox(ni,nj);
	resize(ni);
        _col_dim = nj;
       
            // vecteur temporaire de taille nj trop grande ...
        Sparse_Vector<K> v(nj);
        Elements cour;
        for(Indice ii=0; ii<ni; ii++) {
            Indice k = 0;
            for(Indice jj=0; jj<a[ii].size(); jj++) {
                cour = Elements(a[ii][jj].j(), K(a[ii][jj].getvalue()));
                if (! iszero(cour.getvalue())) 
                    v[k++] = cour;
            }
                // maintenant je connais la taille de mon vecteur ...
            Sparse_Vector<K> w(k);
            for (Indice l=0; l<k; l++)
                w[l] = v[l];
            (*this)[ ii ] = w;
        }
    }
            
    template<class Field, class OtherRep>
    void firsthomomorphic(const Field& F, const BlackBox < _IBB_VECTOR_ < Sparse_Vector < OtherRep > > >& a) {
//         Indice ni=a.n_row(),nj=a.n_col();
//         *this = BlackBox(ni,nj);
//          _col_dim = nj;
       
//             // vecteur temporaire de taille nj trop grande ...
//         Sparse_Vector<K> v(nj);
//         K val;
//         Elements cour;
//         for(Indice ii=0; ii<ni; ii++) {
//             Indice k = 0;
//             for(Indice jj=0; jj<a[ii].size(); jj++) {
//                 F.assign(val,a[ii][jj].getvalue());
//                 cour = Elements(a[ii][jj].j(), val);
//                 if (! F.iszero(cour.getvalue()))
//                     v[k++] = cour;
//             }
//                 // maintenant je connais la taille de mon vecteur ...
//             Sparse_Vector<K> w(k);
//             for (Indice l=0; l<k; l++)
//                 w[l] = v[l];
//             (*this)[ ii ] = w;
//         }
        Indice ni=a.n_row(),nj=a.n_col();
//	        *this = BlackBox(ni,nj);
	resize(ni);
        _col_dim = nj;
       
        K ktmp; 
        for(Indice ii=0; ii<ni; ii++)
            for(Indice jj=0; jj<a[ii].size(); jj++)
                if (! F.iszero( F.assign(ktmp, a[ii][jj].getvalue() ) )) 
                   (*this)[ii].push_back( Elements(a[ii][jj].j(), ktmp) );
    }
            
    template<class OtherField>
    void abshomomorphic(const BlackBox < _IBB_VECTOR_ < Sparse_Vector < OtherField > > >& a) {
        Indice ni=a.n_row(),nj=a.n_col();
//        *this = BlackBox(ni,nj);
	resize(ni); 
        _col_dim = nj;
       
            // vecteur temporaire de taille nj trop grande ...
        Sparse_Vector<K> v(nj);
        Elements cour;
        for(Indice ii=0; ii<ni; ii++) {
            Indice k = 0;
            for(Indice jj=0; jj<a[ii].size(); jj++) {
                cour = Elements(a[ii][jj].j(), K(GIVABS(a[ii][jj].getvalue())));
                if (! iszero(cour.getvalue())) 
                    v[k++] = cour;
            }
                // maintenant je connais la taille de mon vecteur ...
            Sparse_Vector<K> w(k);
            for (Indice l=0; l<k; l++)
                w[l] = v[l];
            (*this)[ ii ] = w;
        }
    }
            
    void abshomomorphic() {
        Indice ni=n_row(),nj=n_col();
        for(Indice ii=0; ii<ni; ii++) {
            Indice k = 0;
            for(Indice jj=0; jj<(*this)[ii].size(); jj++) {
                (*this)[ii][jj].affect( GIVABS((*this)[ii][jj].getvalue()));
            }
        }
    }
            
    template<class Field, class OtherRep>
    void absthomomorphic(const Field& F, const BlackBox < _IBB_VECTOR_ < Sparse_Vector < OtherRep > > >& a) {
        Indice ni=a.n_row(),nj=a.n_col();
//        *this = BlackBox(ni,nj);
	resize(ni);
        _col_dim = nj;
       
        K ktmp; 
        for(Indice ii=0; ii<ni; ii++) {
            (*this)[ii].resize(0);
            for(Indice jj=0; jj<a[ii].size(); jj++)
                if (! F.iszero( F.assign(ktmp, GIVABS(a[ii][jj].getvalue()) ) )) 
                    (*this)[ii].push_back( Elements(a[ii][jj].j(), ktmp) );
        }
    }
            
    template<class OtherField>
    void homomorphic(const BlackBox < _IBB_VECTOR_ < Sparse_Vector < OtherField > > >& a) {
        Indice ni=a.n_row(),nj=a.n_col();
        for(Indice ii=0; ii<ni; ii++)
            for(Indice jj=0; jj<a[ii].size(); jj++)
                (*this)[ii][jj].affect( a[ii][jj].getvalue() );
    }
            
    long n_row() const { return size(); }
    long n_col() const { return _col_dim; }// don't know ...

    typedef _IBB_VECTOR_<K> PreferredInMatrix;
    typedef _IBB_VECTOR_<K> PreferredOutMatrix;
    typedef K coefficientSpace;
    
    template<class OutMatrix, class InMatrix>
    OutMatrix& Apply(OutMatrix& res,  const InMatrix& vect ) {
        return Apply_specialist(FT<K>::field_info(), res, vect);
    }
    
    template<class Field, class OutMatrix, class InMatrix>
    OutMatrix& Apply(const Field& F, OutMatrix& res,  const InMatrix& vect ) {
        Elements toto;
        Indice k,i;
        for(k=size(); k-- ; )
            for( res[k]=F.zero, i=operator[](k).size(); i-- ; ) {
                toto = operator[](k)[i];
                F.axpyin(res[k],toto.getvalue(),vect[toto.j()]);
            }
        return res;
    }
    
    template<class OutMatrix, class InMatrix, class M>
    OutMatrix& Apply(OutMatrix& res,  const InMatrix& vect, M MOD, M PRIME ) {
        Elements toto;
        Indice k,i;
        for(k=size(); k-- ; )
            for( res[k]=Zero, i=operator[](k).size(); i-- ; ) {
                toto = operator[](k)[i];
                res[k] = (res[k] + toto.getvalue() * vect[toto.j()]) % MOD;
            }
        return res;
    }
    
    template<class OutMatrix, class InMatrix>
    OutMatrix& ApplyTrans(OutMatrix& res,  const InMatrix& vect ) {
        return ApplyTrans_specialist(FT<K>::field_info(), res, vect);
    }

    template<class Field, class OutMatrix, class InMatrix>
    OutMatrix& ApplyTrans(const Field& F, OutMatrix& res,  const InMatrix& vect ) {
        Elements toto;
        Indice k,i;
        for(i=res.size(); i-- ; )
            res[i] = F.zero;
        for(k=size(); k-- ; )
            for(i=operator[](k).size(); i-- ; ) {
                toto = operator[](k)[i];
                F.axpyin(res[ toto.j() ],toto.getvalue(), vect[k]);
            }
        return res;
    }
    
    template<class OutMatrix, class InMatrix, class M>
    OutMatrix& ApplyTrans(OutMatrix& res,  const InMatrix& vect, M MOD, M PRIME) {
        Elements toto;
        Indice k,i;
        for(i=res.size(); i-- ; )
            res[i] = Zero;
        for(k=size(); k-- ; )
            for(i=operator[](k).size(); i-- ; ) {
                toto = operator[](k)[i];
                res[ toto.j() ] = (res[ toto.j() ] + toto.getvalue() * vect[k]) % MOD;
            }
        return res;
    }

            
private:
    template<class OutMatrix, class InMatrix>
    OutMatrix& Apply_specialist(default_field xxx, OutMatrix& res,  const InMatrix& vect ) {
            // res is supposed to be preallocated with the good size
            // but no need to set it to zero.
        Elements toto;
        Indice k,i;
        for(k=size(); k-- ; )
            for( res[k]=Zero, i=operator[](k).size(); i-- ; ) {
                toto = operator[](k)[i];
                res[k] += toto.getvalue() * vect[toto.j()];
            }
        return res;
    }

    template<class OutMatrix, class InMatrix>
    OutMatrix& Apply_specialist(muladd_field xxx, OutMatrix& res,  const InMatrix& vect ) {
            // res is supposed to be preallocated with the good size
            // but no need to set it to zero.
        Elements toto;
//        K Zero(zero);
        Indice k,i;
        for(k=size(); k-- ; )
            for( res[k]=Zero, i=operator[](k).size(); i-- ; ) {
                toto = operator[](k)[i];
                res[k].automuladd( toto.getvalue() , vect[toto.j()] );
            }
        return res;
    }

    template<class OutMatrix, class InMatrix>
    OutMatrix& ApplyTrans_specialist(default_field xxx, OutMatrix& res,  const InMatrix& vect ) {
            // res is supposed to be preallocated with the good size
            // but no need to set it to zero.
        Elements toto;
//        K Zero(zero);
        Indice k,i;

        for(i=res.size(); i-- ; )
            res[i] = Zero;

        for(k=size(); k-- ; )
            for(i=operator[](k).size(); i-- ; ) {
                toto = operator[](k)[i];
                res[ toto.j() ] += toto.getvalue() * vect[k];
            }
        return res;
    }

    template<class OutMatrix, class InMatrix>
    OutMatrix& ApplyTrans_specialist(muladd_field xxx,  OutMatrix& res,  const InMatrix& vect ) {
            // res is supposed to be preallocated with the good size
            // but no need to set it to zero.
        Elements toto;
//        K Zero(zero);
        Indice k,i;

        for(i=res.size(); i-- ; )
            res[i] = Zero;

        for(k=size(); k-- ; )
            for(i=operator[](k).size(); i-- ; ) {
                toto = operator[](k)[i];
                res[ toto.j() ].automuladd( toto.getvalue() , vect[k] );
            }
        return res;
    }

    Indice _col_dim;
    K Zero;
};



#endif


