#ifndef _POLY1_H_
#define _POLY1_H_
// =========================================================================== //
// Univariate polynomial
// class Poly1<T>
// (c) copyright GIVARO 1994
// author: still Th. Gautier
// version: 1.0
// date: 1994
// bugs: unknown
// =========================================================================== //
// files needed: pacmemory.h givarotype.h indeter.h
#include "givtype.h"
#include "indeter.h"
#include "array1.h"

#ifdef __GNUC__
template <class T> class Poly1;
template <class T>
inline T eval(const Poly1<T>&P, const T& val);
template <class T>
inline ulong length(const Poly1<T> &P) ;
template <class T>
inline int iszero (const Poly1<T> &P) ;
template <class T>
inline int isone (const Poly1<T> &P) ;
template <class T>
inline int isconstant (const Poly1<T> &P) ;
template <class T>
inline const Poly1<T> operator*(const T& u, const Poly1<T>& P);
template <class T>
inline const Poly1<T> operator+(const T& u, const Poly1<T>& P);
template <class T>
inline const Poly1<T> operator-(const T& u, const Poly1<T>& P);
template <class T>
inline const Poly1<T>  pow(const Poly1<T>&P,  const long n) ;
template <class T>
inline const Poly1<T>  powmod(const Poly1<T>&P,  const long n, const Poly1<T>&U) ; 
template <class T>
inline const Poly1<T> Karatsuba(const Poly1<T>& P, const Poly1<T>& Q) ;
template <class T>
inline void bezout (const Poly1<T> &P, const Poly1<T> &Q, Poly1<T> &d, Poly1<T> &u, Poly1<T> &v) ;
#endif



//  -------------------------------------------- Class Poly1<T>
template <class T>
class Poly1 : public Array1<T> { 
public :

  // Default cstor : ancien cas n = zero    (JGD 01.04.97)
       Poly1 ();

       // Default cstor
       Poly1 (const Neutral n) ;

//       Poly1 (const Neutral n =zero);


/*
static const Poly1<T> Zero ;
static const Poly1<T> One ;
*/
	
       // Cste polynomial 
       Poly1 (const T &cste )  ;      
      
       // Cstor of recopy 
       Poly1 (const Poly1<T>& P) ;
   
       // Allocate a polynomial with deg+1 coefficients
       Poly1 (const Indeter& Y, const Degree deg = deginfty) ;

       // Ctsor of a new polynomial from a array pointed
       // by tabcoef. All elements are recopied.
       Poly1 (const Indeter& X, const Degree deg , const T* tabcoef) ;
       Poly1 (const Indeter& X, const Degree deg , const Array0<T>& tabcoef) ;

       // For polynomial = lcoeff X^deg 
       Poly1 (const Indeter& X, const Degree deg , const T& lcoeff) ;
       
       // Logical dstor
       ~Poly1 () ;

       // Assignment operator
       Poly1<T>& operator= (const Poly1<T> &P) ;

       // Logical recopy:
inline       Poly1<T>& logcpy (const Poly1<T>& P) ;

	// Physical recopy
inline       Poly1<T>& memcpy (const Poly1<T>& P) ;
	
       // Evaluation on one value.
       T eval(const T& val ) const ;

       // Returns the differentiate polynomial
       const Poly1<T> diff() const ;

       // Returns the degree of polynomial
       Degree degree() const  ;

       // Returns the valuation of polynomial
       Degree val() const  ;

       // Returns the leading coefficient
       const T leadcoef() const ;

       // Returns the trailing coefficient
       const T tcoef() const ;


       // Access operators : read and write !!!
inline const T&  operator[] (const int i) const  ;
inline T& operator[] (const int i) ;


       // Comparaison operator
       int operator== (const Poly1<T> &P) const  ;
       int operator!= (const Poly1<T> &P) const
	{ return !(operator== (P)) ; }

       // Returns the variate of the polynomial  
inline const Indeter& variable() const  ;
inline const Indeter& indeter() const  ;
inline const Indeter& indeter(const Indeter& Y)  ;
inline const Indeter& variable(const Indeter& Y) ;

       // Arithemetic operators 
       const Poly1<T>& operator*= (const Poly1<T> &P) ;
       const Poly1<T>& operator+= (const Poly1<T> &P) ;
       const Poly1<T>& operator-= (const Poly1<T> &P) ;
       const Poly1<T>& operator%= (const Poly1<T> &P) ;  // JGD 14.04.97
       const Poly1<T>& operator/= (const Poly1<T> &P) ;  // JGD 14.04.97

       const Poly1<T>  operator * (const Poly1<T> &P) const ;
       const Poly1<T>  operator + (const Poly1<T> &P) const ;
       const Poly1<T>  operator - (const Poly1<T> &P) const ;

       const Poly1<T>  operator - () const ;


       const Poly1<T>  operator * (const T&) const ;
       const Poly1<T>  operator + (const T&) const ;
       const Poly1<T>  operator - (const T&) const ;

#ifdef __GNUC__
//        // Evaluation on one value.
// friend T eval<>(const Poly1<T>&P, const T& val);
//        // Returns the size in byte of the polynomial 
// friend ulong length<>(const Poly1<T> &P) ;
//        // If null
// friend int iszero<> (const Poly1<T> &P) ;
// friend int isone<> (const Poly1<T> &P) ;
// friend int isconstant<> (const Poly1<T> &P) ;
// friend const Poly1<T> operator* <>(const T& u,
//                                    const Poly1<T> & P);
// friend const Poly1<T> operator+ <>(const T& u,
//                                    const Poly1<T>& P);
// friend const Poly1<T> operator- <>(const T& u,
//                                    const Poly1<T>& P);
// friend const Poly1<T>  pow <>(const Poly1<T>&P,  const long n) ;
// friend const Poly1<T>  powmod <>(const Poly1<T>&P,  const long n, const Poly1<T>&U) ; // JGD 14.04.97
// friend const Poly1<T> Karatsuba <>(const Poly1<T>& P, const Poly1<T>& Q) ;
// friend void bezout  <>(const Poly1<T> &P, const Poly1<T> &Q,
//              Poly1<T> &d, Poly1<T> &u, Poly1<T> &v) ;
#else
       // Evaluation on one value.
friend T eval(const Poly1<T>&P, const T& val);
       // Returns the size in byte of the polynomial 
friend ulong length(const Poly1<T> &P) ;
       // If null
friend int iszero (const Poly1<T> &P) ;
friend int isone (const Poly1<T> &P) ;
friend int isconstant (const Poly1<T> &P) ;
friend const Poly1<T> operator* (const T&, const Poly1<T>&);
friend const Poly1<T> operator+ (const T&, const Poly1<T>&);
friend const Poly1<T> operator- (const T&, const Poly1<T>&);
friend const Poly1<T>  pow (const Poly1<T>&P,  const long n) ;
friend const Poly1<T>  powmod (const Poly1<T>&P,  const long n, const Poly1<T>&U) ; // JGD 14.04.97
friend const Poly1<T> Karatsuba (const Poly1<T>& P, const Poly1<T>& Q) ;
friend void bezout  (const Poly1<T> &P, const Poly1<T> &Q,
             Poly1<T> &d, Poly1<T> &u, Poly1<T> &v) ;
#endif


public:  //  ------------------------------------ Private part 
  
  Degree deg ;
  Indeter X ;
  
    // pour les input
static void read_var (std::istream& in, Indeter& X, Degree& deg) ;


} ; //  ------------------------------- End Of The Class Poly1<T>



template<class T>
inline const T&  Poly1<T>::operator[] (const int i) const  
       { return Array1<T>::operator[](i) ; }

template<class T>
inline T& Poly1<T>::operator[] (const int i) 
       { return Array1<T>::operator[](i) ; }

template<class T>
inline const Indeter& Poly1<T>::variable() const 
     { return X ; }
template<class T>
inline const Indeter& Poly1<T>::indeter() const 
     { return X ; }
template<class T>
inline const Indeter& Poly1<T>::indeter(const Indeter& Y) 
     { X.logcpy(Y) ; return X ; }  // JGD 07.05.97
//     { X = Y ; return X ; }
template<class T>
inline const Indeter& Poly1<T>::variable(const Indeter& Y)
     { X.logcpy(Y) ; return X ; }  // JGD 07.05.97


  // External definition :
template <class T>
inline T eval(const Poly1<T>&P, const T& val)
	{ return P.eval(val) ; }

template <class T>
inline int iszero (const Poly1<T> &P) 
        { return P.degree() == deginfty ; }

template <class T>
inline int isone (const Poly1<T> &P) 
        { return (P.degree() == 0)&&((P[0]==T(one)) || (P[0]==-T(one))) ; }

template <class T>
inline int isconstant (const Poly1<T> &P) 
        { return (P.degree() == 0); }


	// Ouput-Input
template <class T>
std::ostream& operator<< (std::ostream &o, const Poly1<T> &P) ;

template <class T>
std::istream& operator>> (std::istream &o, Poly1<T> &P) ;

template <class T>
const Poly1<T> operator/ (const Poly1<T> &P, const T& u) ;

template <class T>
const Poly1<T> div (const Poly1<T> &P,
                    const Poly1<T> &Q,
                    Poly1<T> &quotient,
                    Poly1<T> &remainder) ;

template <class T>
const Poly1<T> operator/ (const Poly1<T> &P,const Poly1<T> &Q) ;

template <class T>
const Poly1<T> operator% (const Poly1<T> &P,const Poly1<T> &Q) ;

template <class T>
void bezout (const Poly1<T> &P, const Poly1<T> &Q,
             Poly1<T> &d, Poly1<T> &u, Poly1<T> &v) ;

template <class T>
inline const Poly1<T> gcd (const Poly1<T> &P, const Poly1<T> &Q ) ;

template<class T>
inline const Poly1<T> prime_part(const Poly1<T> &P, const Poly1<T> &Q) ;

template<class T>
inline void decomposition(const Poly1<T> &P, const Poly1<T> &Q, Poly1<T> & prime_Q, Poly1<T> & divisor_Q) ;

#ifdef _MACINTOSH_
#  include "poly1.mac"
#else
#  include "poly1.c"
#endif

template <class T1, class T2>
void change_base(const Poly1<T1>& P1, Poly1<T2>& P2) {
  P2 = Poly1<T2>(P1.variable(),P1.degree());
  for(Degree i=0; i<=P1.degree();i++)
    P2[i] = T2(P1[i]);
}


#include "poly1.inl"

#endif
