// ======================================================================== //
// Univariate polynomial
// class Poly1<T>
// (c) copyright GIVARO 1994
// author: Th. Gautier....
// date: 1994
// bugs: unknown
// ======================================================================== //


// ======================================================================== //
//
// Description des fonctions :
//	- operator+ 
//	- operator* 
//	- operator- 
//	- operator+= 
//	- operator*= 
//	- operator-= 
//
// ======================================================================== //


template <class T>
const Poly1<T> Poly1<T>::operator+ (const Poly1<T> &P) const 
{
  Degree degThis = degree() ;
  Degree degP = P.degree() ;

  if (degThis == deginfty) return P ;
  if (degP == deginfty)    return *this ;

  if (variable() != P.variable() ) 
    {
       if (degP == 0) 
	 {
            Poly1<T> Res ;
	    Res.memcpy(*this) ;
	    Res[0] = Res[0] + P[0] ;
	    return Res ;
	 }
       else  if (degThis == 0) 
        {
            Poly1<T> Res ;
	    Res.memcpy(P) ;
	    Res[0] = (*this)[0] + Res[0] ; 
	    return Res ;
         }
       cerr << "in Poly1::operator+, variables are not equals" << endl ;
       cerr << *this << endl << P << endl ;
       return *this ;
    }

  // Ok here we can compute the sum :
  Degree degmin, degmax ;
  Degree i ;
  int this_is_of_lower_degree ;

  if (degThis > degP ) {
    this_is_of_lower_degree = 0 ;
    degmin = degP ;
    degmax = degThis ;
  }
  else {
    this_is_of_lower_degree = 1 ;
    degmin = degThis ;
    degmax = degP ;
  }

  // Computation of the coefficients of the result
  Poly1<T> Res(P.variable(), degmax ) ;   // new object

  for (i=0; i<= degmin; i++) 
    Res[i] = (*this)[i] + P[i] ;
  for (i=degmin+1; i<= degmax; i++) 
    if (this_is_of_lower_degree)  
         Res[i] = P[i] ;
    else 
         Res[i] = (*this)[i] ;

  // The computation of the effective degree of the result
  // is done in the next cstor 
  return Res ;
}


template <class T>
const Poly1<T>& Poly1<T>::operator+= (const Poly1<T> &P)
{
   return logcpy(*this + P) ;
}

template <class T>
const Poly1<T> Poly1<T>::operator- (const Poly1<T> &P) const 
{
  return operator+(-P) ; // we can save some memory
}

template <class T>
const Poly1<T>& Poly1<T>::operator-= (const Poly1<T> &P)
{
  return logcpy(*this - P) ;
}

template <class T>   // JGD 14.04.97
const Poly1<T>& Poly1<T>::operator%= (const Poly1<T> &P)
{
   return logcpy(*this % P) ;
}

template <class T>   // JGD 14.04.97
const Poly1<T>& Poly1<T>::operator/= (const Poly1<T> &P)
{
   return logcpy(*this / P) ;
}



template <class T>
const Poly1<T> Poly1<T>::operator- () const 
{
  Degree deg_autre_nom = degree() ;
  if (deg_autre_nom == deginfty) return *this ;
  
  Poly1<T> Res (variable(), deg_autre_nom) ;   // new object
  Degree i ;
  for (i=0 ; i <= deg_autre_nom ; i++) Res[i] = - (*this)[i] ;
  return Res ;
}


template <class T>
const Poly1<T> Poly1<T>::operator* (const Poly1<T>& P) const 
{
   Degree degthis, degP;
   degthis = degree() ;
   degP = P.degree() ;
   if (degthis == deginfty) return *this ;  // Res = 0
   if (degP == deginfty) return  P ;        // idem

   // Multiplication with cstes
   if (degP == 0)    return operator*(P[0]) ;
   if (degthis == 0) return P.operator*( (*this)[0]) ; 
	
   if (variable() != P.variable()) 
   { 
      cerr << "in Poly1::operator* : variables are not equals" << endl ;
//      cerr << *this << endl << P << endl ;
      return *this ; // erreur ici 
   }
/*
  if ((degP > KARA_THRESHOLD) && (degthis > KARA_THRESHOLD)) 
      return Karatsuba(*this,P) ;
*/


   Poly1<T> Res(P.indeter(),degthis+degP);
   Degree i,j ;
   T Zero = zero ;
	
   for (i = 0 ; i <= degthis+degP ; i++) Res[i] = Zero ;
        
   for (i=0 ; i<= degthis ; i++) 
      for (j=0; j<= degP ; j++)
         Res[i+j] = Res[i+j] + (*this)[i] * P[j] ;
   return Res;
}

template <class T>
const Poly1<T>& Poly1<T>::operator*= (const Poly1<T>& P)
{
    return logcpy( *this * P);
}


template <class T>
const Poly1<T>  pow (const Poly1<T>&P, const long pwr) 
{
   Poly1<T> Res (one) ;
   Poly1<T> Puiss ;
   Puiss.memcpy( P );            // must recopy the object
   ulong n = (pwr < 0 ? -pwr : pwr); 

   if (n>0) {
      while (n != 0) {
         if (n & 0x1) {
            Res *= Puiss ;
         }
         n >>= 1 ;
         Puiss *= Puiss ;
      }
   }
   return Res ;
}



template <class T>  // JGD 14.04.97
const Poly1<T>  powmod (const Poly1<T>&P, const long pwr, const Poly1<T>&U) 
{
   Poly1<T> Res (one) ;
   Poly1<T> Puiss ;
   Puiss.memcpy( P );            // must recopy the object
   ulong n = (pwr < 0 ? -pwr : pwr); 

   Poly1<T> tmp ;

   Puiss %= U;
   if (n>0) {
      while (n != 0) {
         if (n & 0x1) {
	   Res = (Res * Puiss) % U;
	    //Res = Karatsuba(Puiss, Res) % U;
           
         }
         n >>= 1 ;
	 Puiss = (Puiss * Puiss) % U;
         // Puiss = Karatsuba(Puiss , Puiss) % U ;
         
      }
   }
   return Res ;
}

// ======================================================================== //
// Mixed polynomial-scalar operators
//

// Operations avec *

template <class T>
const Poly1<T>  Poly1<T>::operator* (const T& u) const 
{
   if (iszero(u)) 
      return Poly1<T>(zero);
   else if (iszero(*this)) 
      return *this ;
   else
   {
      Degree i, deg_autre_nom ;
      deg_autre_nom = degree() ;
      Poly1<T> Res (variable(), deg_autre_nom);  // new object
      for (i =0 ; i<= deg_autre_nom  ; i++) 
         Res[i] = (*this)[i] * u ;
      return Res ;
   }
}

template <class T>
const Poly1<T> operator* (const T& u, const Poly1<T>& P) 
{
   if (iszero(u)) 
      return Poly1<T>(zero);
   else if (iszero(P)) 
      return P ;
   else
   {
      Degree i, deg_autre_nom ;
      deg_autre_nom = P.degree() ;
      Poly1<T> Res (P.variable(), deg_autre_nom);  // new object
      for (i =0 ; i<= deg_autre_nom ; i++) 
         Res[i] = u * P[i] ;
      return Res ;
   }
}

// Operators with +
template <class T>
const Poly1<T>  Poly1<T>::operator+ (const T& u) const 
{
   if (iszero(u)) return *this ;
   if (iszero(*this)) return Poly1<T>(u) ;
   Poly1<T> Res ;
   Res.memcpy(*this) ;   // Must recopy the object
   Res[0] = Res[0] + u ;
   if (iszero(Res)) {
       return Poly1<T>(zero) ;
   }
   return Res ;
}

template <class T>
const Poly1<T>  operator+ (const T& u, const Poly1<T>& P) 
{
   if (iszero(u)) return P ;
   if (iszero(P)) return Poly1<T>(u) ;

   Poly1<T> Res ;
   Res.memcpy(P) ;       // Must recopy the object
   Res[0] = u + Res[0] ;
   if (iszero(Res)) {
       return Poly1<T>(zero) ;
   }
   return Res ;
}

// Operations with -

template <class T>
const Poly1<T>  Poly1<T>::operator- (const T& u) const 
{
   if (iszero(u)) return *this ;
   if (iszero(*this)) return Poly1<T>(u) ;

   Poly1<T> Res ;
   Res.memcpy( *this );   // Must recopy the object
   Res[0] = Res[0] - u ;
   if (iszero(Res)) {
        return Poly1<T>(zero) ;
   }
   return Res ;
}

template <class T>
const Poly1<T>  operator- (const T& u, const Poly1<T>& P)
{
   if (iszero(u)) return P ;
   if (iszero(P)) return Poly1<T>(u) ;

   Poly1<T> Res = -P ;  // we can do that because -P create a new object !!!
   Res[0] = u + Res[0] ;
   if (iszero(Res)) {
        return Poly1<T>(zero) ;
   }
   return Res ;
}
