// =================================================================== //
// Givaro / Athapascan-1
// Cyclotomic polynomilas
// Time-stamp: <07 Jun 99 18:11:35 Jean-Guillaume.Dumas@imag.fr> 
// =================================================================== //
#ifndef _GIVARO_POLY_CYCLO_
#define _GIVARO_POLY_CYCLO_

#include <poly1.h>
#include <intnumtheo.h>

// ---------------------------------------------------------------
// Composition by a power of X
// ---------------------------------------------------------------
template<class T, class U>
Poly1<T>& power_compose(Poly1<T>& W, const Poly1<T>& P, U b) {
    T Zero(zero);
    Degree dp = P.degree();
    Poly1<T> Res(P.indeter(),b*dp,P.leadcoef());
    Degree r=0;
    for(Degree i=0;i<dp;++i) {
        Res[r++] = P[i];
        for(Degree j=b;--j;)
            Res[r++] = Zero;
    }
    return W = Res;
}




// ---------------------------------------------------------------
// n th Cyclotomic polynomial 
// ---------------------------------------------------------------
 
template<class T, class U>
Poly1<T>& cyclotomic( Poly1<T>& P, U n) ;


template<class T, class U>
Poly1<T>& cyclotomic( Poly1<T>& P, U n) {
        // P must provide an indeterminate
    T un(one);
    Integer In(n);
    if (n == U(one)) {
        Poly1<T> Res(P.indeter(),Degree(1),un);
        Res[0] = -un;
        return P = Res;
    } else if (isprime(In)) {
        Poly1<T> Res(
            P.indeter(),
            Degree(n-1),
            un);
        for(Degree i=n-1;i--;)
            Res[i] = un;
        return P = Res;
    } else {
        T Zero(zero);
        U q,r;
        q = n / 2;
        r = n % 2;
        if (iszero(r)) {
            if (iszero(q%2)) {
                    // q even
                Poly1<T> inter(P.indeter());
                cyclotomic(inter,q);
                Degree di = inter.degree();
                Poly1<T> Res;
                return P = power_compose(Res,inter,2);
            } else {
                    // q odd
                Poly1<T> Res(P.indeter());
                cyclotomic(Res,q);
                Degree di = Res.degree();
                for(Degree i=1;i<=di;i+=2)
                    Res[i] = - Res[i];
                return P = Res;
            }
        } else { 
            Integer Ir = Integer_factor(In);
            while (! isprime(Ir))
                Ir = Integer_factor(Ir);
            r = Ir % n;
            q = n / r;
            Poly1<T> cq(P.indeter());
            cyclotomic(cq,q);
            Poly1<T> Res;
            if (iszero (q % r))
                return P = power_compose(Res,cq,r);
            else
                return P = power_compose(Res,cq,r) / cq;
        }
    }
}
    
#endif
