// =================================================================== //
// Givaro / Athapascan-1
// Irreducibily test
// Factorisations de  Polynomes dans Fp[X] :
//      Distinct Degree
//      Cantor-Zassenhaus
//      Berlekamp : in LinBox
// Time-stamp: <23 Jun 99 18:36:02 Jean-Guillaume.Dumas@imag.fr> 
// =================================================================== //
#ifndef _GIVARO_POLY_FACTOR_
#define _GIVARO_POLY_FACTOR_

#include <stdlib.h>
#include <list>
using std::list;
#include <integer.h>
#include <poly1.h>


// -----------------------------------
// Polynomial utilities
// -----------------------------------
#include <givrandom.h>

template<class T>
Poly1<T>& RANDOMPOLY1(Poly1<T>& VV, const Degree& n, const Indeter& Y) 
{
    T Zero = zero;
    T V; NONZEROGIVRANDOM(V);
    VV = Poly1<T>(Y, n, V);
    for(Indice i=0; i<n; i++)
        GIVRANDOM(VV[i]);
    return  VV;
};


template<class T>
const Poly1<T>  powmod (const Poly1<T> &P, const Integer pwr, const Poly1<T>&U) 
{
    Poly1<T> Res (one) ;
    Poly1<T> Puiss ;
    Puiss.memcpy( P );            // must recopy the object
    Integer n = (pwr < 0 ? -pwr : pwr);
    Integer q,r;
    Integer deux(2);
 
    Puiss %= U;
    if (n>0) {
        while (n != 0) {
            divide(n,deux,q,r);
            n = q;
            if (r != 0)
                Res = (Res * Puiss) % U;
            Puiss = (Puiss * Puiss) % U;
        }
    }
    return Res ;
}


// ---------------------------------------------------------------
// Splits a polynomial into prime factors of same degree
// ---------------------------------------------------------------

template<class T>
void SplitFactor(Integer MOD, const Poly1<T>& G, Degree d, list< Poly1<T> > & L) {
    Degree dG = G.degree();
    if (dG == d)
        L.push_back(G);
    else {
        int splitted = 0;
        Integer pp = (pow(MOD, long(d)) - 1)/2;
        while (! splitted) {
            Poly1<T> tmp; RANDOMPOLY1(tmp, d, G.indeter() );
            Poly1<T> G1( gcd(G, tmp ) );
            Degree dG1 = G1.degree();
            if ( dG1 != dG) {
                if (dG1 > 0 ) {
                    splitted = 1;
                    SplitFactor ( MOD, G1, d, L) ;
                }

                Poly1<T> tp = powmod(tmp, pp, G);
                Poly1<T> G2( gcd(G, tp-1 ) );
                Degree dG2 = G2.degree();
                if ( dG2 != dG) {
                    if ( dG2 > 0 ) {
                        splitted = 1 ;
                        SplitFactor ( MOD, G2, d, L) ;
                    }
                    Poly1<T> G3( gcd(G, tp+1 ) );
                    Degree dG3 = G3.degree();
                    if (( dG3 != dG) && (dG3 > 0 )) {
                        splitted = 1 ;
                        SplitFactor ( MOD, G3, d, L) ;
                    }
                }
            }
        }
    }
}


template<class T>
const Poly1<T> SplitFactor(Integer MOD, const Poly1<T>& G, Degree d) {
    Degree dG = G.degree();
    if (dG == d)
        return G ;
    else {
        Integer pp = (pow(MOD, long(d)) - 1)/2;
        while (1) {
            Poly1<T> tmp; RANDOMPOLY1(tmp, d, G.indeter() );
            Poly1<T> G1( gcd(G, tmp ) );
            Degree dG1 = G1.degree();
            if ( dG1 != dG) {
                if (dG1 > 0 ) {
                    return G1;
                }
                Poly1<T> tp = powmod(tmp, pp, G);
                Poly1<T> G2( gcd(G, tp-1 ) );
                Degree dG2 = G2.degree();
                if ( dG2 != dG) {
                    if ( dG2 > 0 ) {
                        return G2;
                    }
                    Poly1<T> G3( gcd(G, tp+1 ) );
                    Degree dG3 = G3.degree();
                    if (( dG3 != dG) && (dG3 > 0 )) {
                        return G3;
                    }
                }
            }
        }
    }
}


// ---------------------------------------------------------------
// Splits a polynomial into divisors of homogenous prime factors
// ---------------------------------------------------------------

template<class Modulo, class T>
void DistinctDegreeFactor(Modulo MOD, const Poly1<T>& f, list< Poly1<T> > & L) {
    //srand48(BaseTimer::seed());
    Integer IMOD(MOD);
    Poly1<T> P = f;
    Poly1<T> Unit(P.indeter(), 1, T(one));
    Poly1<T> W = Unit;
    for(Degree ld = 1; ld<= (P.degree() / 2); ld++) {
        W = powmod(W, MOD, P);
        Poly1<T> G1( gcd (W - Unit, P) );
        if ( G1.degree() != 0 ) {
            P = P / G1;
            SplitFactor ( IMOD, G1, ld, L);
        }
    }
    if (P.degree() != 0 )
        L.push_back(P);
}

// ---------------------------------------------------------------
// Cantor-Zassenhaus Polynomial factorization over Z/pZ
// ---------------------------------------------------------------

template<class Modulo, class T>
void CZfactor( Modulo MOD, const Poly1<T>& P, list<Poly1<T> >& L) {
    int nb;
    Poly1<T> * g = new Poly1<T>[P.degree()+1];
    sqrfree(P,nb,g);
    for(Indice i=1; i<nb; i++)
        DistinctDegreeFactor(MOD, g[i], L) ;
    ::delete [] g;
}



// ---------------------------------------------------------------
// Irreducibility tests
// ---------------------------------------------------------------

template<class Modulo, class T>
int is_irreducible(Modulo MOD, const Poly1<T>& P) {
        // Square free ?
    Poly1<T> W = gcd(P.diff(),P);
    if (W.degree() > 0)
        return 0;
        // Distinct degree free ?
    T un(one);
    Poly1<T> Unit(P.indeter(), Degree(1), T(one));
    W = Unit;
    Poly1<T> G1;
    for(Degree ld = 1; ld<= (P.degree() / 2); ++ld) {
        W = powmod(W, MOD, P);
        G1 =  gcd (W - Unit, P) ;
        if ( G1.degree() > 0 )
            return 0;
    }
    return 1;
}

// ---------------------------------------------------------------
// Gives one non-trivial factor of P if P is reducible
// returns P otherwise
// ---------------------------------------------------------------

template<class Modulo, class T>
const Poly1<T> factor(Modulo MOD, const Poly1<T>& P) {
        // Square free ?
    Poly1<T> W = gcd(P.diff(),P);
    if (W.degree() > 0)
        return W;
        // Distinct degree free ?
    //srand48(BaseTimer::seed());
    T un(one);
    Poly1<T> Unit(P.indeter(), 1, T(one));
    W = Unit;
    Poly1<T> G1;
    Degree dP = P.degree();
    for(Degree ld = 1; ld<= (dP / 2); ++ld) {
        W = powmod(W, MOD, P);
        G1 =  gcd (W - Unit, P) ;
        Degree dG1 = G1.degree();
        if ( dG1 > 0 )
            if (dG1 < dP)
                return G1;
            else
                return SplitFactor(Integer(MOD),G1, ld);
    }
    return P;
}


#endif
