// =========================================================================== //
// Univariate polynomial
// class Poly1<T>
// (c) copyright GIVARO 1994
// author: J.L. Roch
// date: 1994
// =========================================================================== //

// Description des fonctions :
// Division euclidienne
//     - div  
//     - operator /
//     - operator %
// friend Poly1<T> div (const Poly1<T> &P,
//                                const Poly1<T> &Q,
//                                Poly1<T> &quotient,
//                                Poly1<T> &remainder) ;
// computes quotient and remainder such that :
//    P == Q * quotient + remainder

// Poly1<T>  operator / (const Poly1<T> &Q) const ;
// Poly1<T>  operator % (const Poly1<T> &Q) const ;

// GCD
//     - gcd
//     - bezout
// friend void bezout (const Poly1<T> &P,
//                                   const Poly1<T> &Q,
//                                   Poly1<T> &d,
//                                   Poly1<T> &u,
//                                   Poly1<T> &v) ;
// computes d = unitary gcd(P,Q) and u,v such that :
//   u*P+v*Q = d
// u and v are the unique polynomisals such that :
//   deg(u) <= deg(Q)-deg(d) and deg(v) <= deg(P) - deg(d)

// friend Poly1<T> gcd (const Poly1<T> &P, const Poly1<T> &Q ) ;
// computes d = unitary gcd(P,Q) 


// ===========================================================================

template <class T>
const Poly1<T> operator/ (const Poly1<T> &P, const T& u) 
{
  if (iszero(u)) {
    cerr << "*** division par zero " << endl ;
    return P ;
  }
  Poly1<T> res (P.variable(), P.degree()) ;
  for (int i=0; i<= P.degree(); i++)
    res[i] = (P)[i] / u ;
  return(res) ;
}

#include <iostream>
using std::cerr;
using std::endl;

template <class T>
const Poly1<T> div(const Poly1<T>& P, const Poly1<T>& Q, Poly1<T>& quotient, Poly1<T>& remainder) {

 if (iszero(P))
  { 
      remainder = P ; 
      return quotient = Poly1<T>(zero) ;
  }
  else if (iszero(Q)) {
    cerr << "*** error : Division by 0 in div " << endl ;
//    cerr << "P : " << P << endl ;
//    cerr << "Q : " << Q << endl ;
  }
  else if (Q.degree() == 0)
    {
          remainder = Poly1<T>(zero) ;
#ifdef __GNUC__
          return quotient = P / (const T&) Q[0] ;
#else
          return quotient = P / Q[0] ;
#endif
    }
  else if (P.degree() == 0)
    {
          remainder = P ;
          return quotient = Poly1<T>(zero) ;
    }

  if (P.variable() != Q.variable())
  { 
    cerr << "Erreur Division : operands are of different variables." << endl ;
    cerr << "P/Q . P : " << P.degree() << endl << "Q :" << endl  ;
    cerr << Q.degree() << endl ;
    quotient = P;
    remainder = Q;
  }

  //  Poly1<T> D=Q;
  /*
  if (D.leadcoef() != (T(one)) ) { 
    D = D / D.leadcoef() ;
  }
  */

  T cpq = Q.leadcoef();
  Poly1<T> D(Q.indeter(),Q.degree(),cpq);
  for(Degree iq=0;iq<Q.degree();iq++)
    D[iq] = Q[iq] / cpq;
  

  Degree degQ = Q.degree() ;
  Degree degP = P.degree() ;
  
  if (degP < degQ) { 
    remainder = P ;
    quotient = Poly1<T>(zero) ;
    return quotient ;
  }

  Poly1<T> quot(P.variable(), degP - degQ ) ;
   
  T tmp ;
  T* rem = new T [ degP + 1 ] ;

  Degree i= degP;
  for (; i>=0; i-- ) 
    rem[i] = P[i] ;

  for ( i= degP - degQ; i>=0; i-- ) {
    // computation of the monomial of quotient with degree i
    tmp = rem[i+degQ] ;
    quot[i] = tmp ;
    for (Degree j = degQ; j>=0 ; j-- ) {
      rem[j+i] -= tmp * D[j] ;
    }
  }

  for ( i = degQ-1 ; (rem[i] == T(zero)) && (i>0) ; i-- ) ;
  Poly1<T> remfinal ( P.variable(), i ) ;
  for (Degree j=i ; j >=0 ; j-- ) remfinal[j] = rem[j] ;
  delete [] rem ;

  remainder = remfinal ;
  quotient = quot / cpq ;
  return quotient ;
}




template <class T>
const Poly1<T> ancien_div (const Poly1<T> &P,
                    const Poly1<T> &Q,
                    Poly1<T> &quotient,
                    Poly1<T> &remainder) {
  if (iszero(P))
  { 
      remainder = P ; 
      return quotient = Poly1<T>(zero) ;
  }
  else if (iszero(Q)) {
    cerr << "*** error : Division by 0 in div " << endl ;
    cerr << "P : " << P << endl ;
    cerr << "Q : " << Q << endl ;
  }
  else if (Q.degree() == 0)
    {
	  remainder = Poly1<T>(zero) ;
#ifdef __GNUC__
          return quotient = P / (const T&) Q[0] ;
#else
          return quotient = P / Q[0] ;
#endif
    }
  else if (P.degree() == 0)
    {
          remainder = P ;
          return quotient = Poly1<T>(zero) ;
    }


  if (P.variable() != Q.variable())
  { 
    cerr << "Erreur Division : operands are of different variables." << endl ;
    cerr << "P/Q . P : " << P.degree() << endl << "Q :" << endl  ;
    cerr << Q.degree() << endl ;
    quotient = P;
    remainder = Q;
  }

  Degree degQ = Q.degree() ;

  Poly1<T> rem ;
  rem = P ;
  Degree degR = rem.degree() ;

  Poly1<T> quot(P.variable(), 0, T(zero));
  quotient = quot ;
       
  while ( (degR=rem.degree()) >= degQ) {
    Poly1<T> new_quot(P.variable(), degR-degQ, rem[degR] / Q[degQ] ) ;
    rem = rem - new_quot * Q ;
    quot = quot + new_quot ;
  }
     
  quotient = quot ;
  remainder = rem ;
  Poly1<T> tmp( Q*quotient ) ;
  tmp = P - tmp ;
  tmp = tmp - remainder ;
  if (!iszero(P-Q*quotient-remainder)) {
    cout << "Erreur dans la verif de div. " << endl 
         << "   P  :=" << P << endl 
         << "   Q  :=" << Q << endl
         << "   P/Q:=" << quotient  << endl 
         << "   P%Q:=" << remainder << endl ;
  }
  return quotient ;
}



template <class T>
const Poly1<T> operator/ (const Poly1<T> &P,const Poly1<T> &Q)
{
  Poly1<T> quotient ;
  Poly1<T> reste ;
  
  return  div(P, Q, quotient, reste) ;
}


template <class T>
const Poly1<T> operator% (const Poly1<T> &P,const Poly1<T> &Q)
{
  Poly1<T> quotient ;
  Poly1<T> reste ;
  div(P, Q, quotient, reste) ;
  return(reste) ;
}


// A coder par PRS : cf Geddes page 283
template <class T>
void bezout (const Poly1<T> &P,
                    const Poly1<T> &Q,
                    Poly1<T> &d,
                    Poly1<T> &u,
                    Poly1<T> &v) {
  // computes d = unitary gcd(P,Q) and u,v such that :
  //   u*P+v*Q = d
  // u and v are the unique polynomisals such that :
  //   deg(u) <= deg(Q)-deg(d) and deg(v) <= deg(P) - deg(d)
 
  // Auxilliary matrix : [oldu oldv] which is left multiplies by [0  1]
  //                     [newu newv]                             [1 -q]
  // at each step. 
  // At the end, u = oldu, v = old v

  static T csteT0(zero) ;
  static T csteT1(one) ;
  Poly1<T> A ;
  Poly1<T> B ;
  Poly1<T> quot ;
  Poly1<T> rem ;
  int permuter ;

  if (P.degree() < Q.degree()) { A = Q; B = P ; permuter = 1; }
  else { A = P; B = Q; permuter = 0; }

  if (iszero(B)) {
    T inv_lcoeff_A ;
    if (iszero(A)) { inv_lcoeff_A = csteT1 ; }
    else { inv_lcoeff_A = csteT1 /A[A.degree()] ; }

    d = A * inv_lcoeff_A ;
    Poly1<T> Zero(A.variable(), (Degree)0, csteT0);
    Poly1<T> cste(A.variable(), (Degree)0, inv_lcoeff_A) ;
    if (permuter) { u = Zero ; v = cste ; }
    else { u = cste ; v = Zero ; }
  }
  else {
    
    // On rend B unitaire 
    T inv_lcoeff_B = csteT1 / B[B.degree()] ; 
    B = B * inv_lcoeff_B ;

    Poly1<T> oldu(P.variable(), Degree(0), csteT1) ;
    Poly1<T> oldv(P.variable(), (Degree)0, csteT0 ) ;
    Poly1<T> newu(P.variable(), (Degree)0, csteT0 ) ;
    Poly1<T> newv(P.variable(), (Degree)0, inv_lcoeff_B);

    int cont ;
    do {
      div(A, B, quot, rem) ;
      A = B ;
      if (cont = !iszero(rem)) { 
        inv_lcoeff_B = csteT1/rem[rem.degree()] ; 
        B = rem*inv_lcoeff_B ;

        Poly1<T> tmpu = (oldu - quot*newu) * inv_lcoeff_B ;
        Poly1<T> tmpv = (oldv - quot*newv) * inv_lcoeff_B ;
        oldu = newu ;
        oldv = newv ;
        newu = tmpu ;
        newv = tmpv ;
      }
    } while (cont) ; 

    
    d = A ;

    if (permuter) { u = newv ; v = newu ; }
    else { u = newu; v = newv ; }
  }

  if (!iszero(u*P+v*Q-d)) {
    cout << "Erreur dans la verif de bezout. " << endl
         << "   P  =" << P << endl
         << "   Q  =" << Q << endl
         << "   d  =" << d  << endl
         << "   u  =" << u << endl 
         << "   v  =" << v << endl ;
  }
}


template <class T>
const Poly1<T> gcd (const Poly1<T> &P, const Poly1<T> &Q ) 
{
  Poly1<T> A ;
  Poly1<T> B ;

  if (P.degree() < Q.degree()) { A = Q; B = P ; }
  else { A = P; B = Q; }

  if (iszero(B)) {
    if (iszero(A)) { return A ; }
    else { return A/A[A.degree()] ; }
  }

  B = B/B[B.degree()] ; // B is made unitary
  Poly1<T> rem = A % B ;
  while (!iszero(rem = A % B)) {
    A = B ;
    B = rem / rem.leadcoef() ;
  }
  return B ;
}


template<class T>
inline const Poly1<T> prime_part(const Poly1<T>& P, const Poly1<T>& Q)
{ 
  Poly1<T> E,D ;
  E = P ;
  D = Q ;
  while (D.degree() >0)
  {
    E = E / D ;
    D = gcd(E,D) ;
  }  
  return E ;
}

// Split the argument into two factors prime_Q and divisor_Q such that :
//  divisor_Q = gcd(P, Q)
//  There exists a constant k such that P divides (prime_Q * divisor_Q)^k
template<class T>
inline void decomposition(const Poly1<T> &P, const Poly1<T> &Q, Poly1<T> & prime_Q, Poly1<T> & divisor_Q)
{
    divisor_Q = gcd(P, Q) ;
    prime_Q= P / gcd( Poly1<T>::pow(divisor_Q,((long)(degree() - divisor_Q.degree()))),
                        P ) ;
}


