// ======================================================================== //
// Univariate polynomial
// class Poly1<T>
// (c) copyright GIVARO 1994
// author: Th. Gautier....
// date: 1994
// ======================================================================== //

#include <ctype.h>

// ==================================================================== //
// Description des fonctions :
//	- operator >>:	entree
//	- operator <<:	sortie

#define OUTPUT_SIZE_BOUND 2

template<class T>
std::ostream&  operator<< (std::ostream& o, const Poly1<T>& P) 
{
   T tmp ;
   Degree i ;

   if (iszero(P) ) {
      return o << "0" ;
   }
   for (i = P.degree() ; i>=0 ;  i--) 
   {
      tmp = P[i] ;
      if (iszero(tmp)) continue ;
      if (i==0) {
          if (sizeof(tmp) > OUTPUT_SIZE_BOUND)
            o << "+(" ;
          else
            o << "+"; // JGD 09.02.98
          o << tmp ;
          if (sizeof(tmp) > OUTPUT_SIZE_BOUND)
            o << ")" ;
      } else  {
	 if (sizeof(tmp) > OUTPUT_SIZE_BOUND)
	   o << "+(" ;
	 else
	   o << "+"; // JGD 09.02.98
         o << tmp ;
	 if (sizeof(tmp) > OUTPUT_SIZE_BOUND)
	   o << ")" ;
	 o << '*' << P.variable() ;
         if (i>1)  o << '^' << i ;
         o << ' ' ;
      } 
   }
   return o ;
}

// Read a identifier that begins with a character.
// The separator are ' ', '\t' , '+' '-' '^' '*'
// Pre.Cond:  in.get(ch) -> ch =) is true.
template <class T>
void Poly1<T>::read_var (std::istream& in, Indeter& X, Degree& deg)
{
   char ch ;
   int len = 1 ;
   char *tmp = new char[len] ;
   tmp[0] = '\0' ;
   in.get(ch) ;
   // Read the string
   while ((in) && !(ch == '\n') && !(ch == ' ') && !(ch == '\t') && !(ch == '+')  
               && !(ch == '-') && !(ch == '^') && !(ch == '*') )
  {
      len++ ;
      char * t = new char[len] ;
      for (int i= 0 ; i <len-2 ; i++) t[i] = tmp[i] ;
      tmp[len-2] = ch ;
      tmp[len-1] = '\0' ;
      in.get(ch) ;
  }
  if (ch == '^')
  {
    in >> deg ;
    X = Indeter(tmp) ;
    delete [] tmp ;
    return ;
  }
  else if (ch == '*')
  {
    in.get(ch) ;
    if (ch != '*') 
     {
       in.putback('*') ;
       in.putback('*') ;
       deg = 1 ;
     }
    else in >> deg ;
    X = Indeter(tmp) ;
    delete [] tmp ;
  }
  else {
    deg = 1 ;
    X = Indeter(tmp) ;
    delete [] tmp ;
    in.putback(ch) ;
    return ;
  }
  return ;
}

template<class T>
std::istream&  operator>> (std::istream& in, Poly1<T>& P) 
{
   if (!in) return in ;

   char ch = ' ';
   Degree deg = deginfty ;
   Indeter X="" ;
   Indeter Xvoid = X ;
   T Zero = zero ;
   T One = one ;
   int sign = 1 ;
   T cste ; 
   int is_first_mon = 1 ;
   
   P = Poly1<T> (zero) ;
   
   in.get(ch);        // JGD 03.04.97
   while((in) && (ch == '\n')) in.get(ch);
   int first = 1;

   while ((in))  // tant qu'il y a des monomes
   {
     if ((!in) || (ch == '\n')) break;        // JGD 03.04.97
     if (!first)  in.get(ch) ;
     else first = 0;

     while ((in) && ((ch == ' ') || (ch == '\t'))) in.get(ch) ;
     if ((!in) || (ch =='\n'))
       { break; }

     // read a sign ? 
     if ((ch == '-') || (ch == '+'))
     {
        if (ch == '-') sign = -1 ;
        else sign = 1 ;
        in.get(ch) ;
        while ((in) && ((ch == ' ') || (ch == '\t'))) in.get(ch) ;
     } 
     else if (!is_first_mon) {
       // ch != '-' and ch != '+' : error
       cerr << "*** error in operator>> on Poly1 " << endl ;
       cerr << "    - / + missing between monomials " << endl ;
       cerr << "    found : " << ch << endl ;
       return in ;
     }
     is_first_mon = 0  ;

     
     // read a cste ?
     if (!isalpha(ch)) 
     {
        in.putback(ch) ;
        in >> cste ;
        if (in.eof()) 
        {
           if (sign == 1) P += cste ;
           else P -= cste ;
           return in ;
        }
        in.get(ch) ;
     }
     else cste = One ;

     deg = 0 ;
 
     while ((in) && ((ch == ' ') || (ch == '\t'))) in.get(ch) ;

     // read a variable ^ deg ?
     if ((ch == '*') || (isalpha(ch))) 
     {
        if (ch == '*') 
        {
           in.get(ch) ;
           while ((in) && ((ch == ' ') || (ch == '\t'))) in.get(ch) ;
        }
        if (!(in) || (ch == '\n')) 
        {
            cerr << "*** error in operator>> on poly1 :" << endl;
            cerr << "    EOF reached before to read a variable." << endl;
            return in ;
        }
        // read a variable
        in.putback(ch) ;
        Indeter Y ;
        Poly1<T>::read_var(in, Y, deg)  ;
        if (deg < 0) 
        {
            cerr << "*** deg < 0 in operator>> on poly1 !!! " << endl;
            return in ;
        }
        if (X == Xvoid) X = Y ;
        else if (X != Y) 
          {
             cerr << "*** error in operator>> on poly1 : found different" << endl ;
             cerr << "    variables: " << X << " and: " << Y << endl ;
             return in ;
          }
     }
     else if (ch != '\n') in.putback(ch) ;

     // We add the monom cste X ^ deg ;
     if (sign ==1) P += Poly1<T> (X, deg, cste) ;
     else P -= Poly1<T> (X, deg, cste) ;

   } // while(in) 
   
   return in ;
}

