// ======================================================================== //
// Univariate polynomial
// class Poly1<T>
// (c) copyright GIVARO 1994
// author: Th. Gautier....
// date: 1994
// ======================================================================== //


// ==================================================================== //
//
// Description des fonctions :
//	- memcpy :	copie physique  JGD 25.04.97 : dans l'.inl
//	- logcpy ;	copie logique   JGD 25.04.97 : dans l'.inl
//	- operator= :	affectation
//	- eval :	evaluation en une valeur ds T
//	- operator[] : 	acces au coef de degre i
//	- operator== :	comparaison


template <class T>
Poly1<T>& Poly1<T>::operator= (const Poly1<T> &P)
{
   Array1<T>::operator=(P) ;
   X.logcpy(P.X) ;
   deg = P.deg ;
   return *this ;
}



template<class T>
int Poly1<T>::operator== (const Poly1<T>& P) const 
{
  if (this == &P) return 1 ;
  Degree degP = P.degree() ;
  Degree degThis = degree() ;
  if (degThis != degP) return 0 ;
  if (degThis == deginfty)   return 1 ;  // this == 0 and P == 0 
  // Here this and P are != 0, and have same degree
  // Look the constant case
  if (degP == 0) return  (*this)[0] == P[0] ;
  if (indeter() != P.indeter()) return 0 ; 
  // We cannot call the comparizon operator on the array
  // because may be there are some extra coefficient
  for (int i= degP ; i>=0 ; i--)
     if ( (*this)[i] != P[i] ) return 0 ;
  return 1 ;
}


// Horner's scheme for evaluation
template <class T>
T Poly1<T>::eval (const T& val) const 
{
   Degree deg_autre_nom = degree() ;
   if (deg_autre_nom == deginfty) return T(zero);   // Gilles 01.09.94
   else { 
     T res = leadcoef() ;
     for (int i = deg_autre_nom ; i>0 ; i--) 
     {
         res *= val ;
         res += (*this)[i-1] ;
     }
     return res ;
   }
}


// Retourne le degre du polynom 
template <class T>
Degree  Poly1<T>::degree() const 
{
   int i = deg ; 
   if (i == deginfty) return deginfty ;
   else if (!iszero((*this)[i]) ) return i ;
   // We must re-compute the degree :
   for (int j=i-1; j>=0 ; j--)
      if (!iszero((*this)[j]) ) return j ;
   return deginfty ;
}

template <class T>
Degree  Poly1<T>::val() const
{
   int i = deg ;   
   if (i == deginfty) return deginfty ;
   else if (!iszero((*this)[0])) return 0 ;
   // We must compute the valuation :
   for (int j=1; j<=i ; j++)
      if (!iszero((*this)[j])) return j ;
   return deginfty ;
}


// Retourne le coefficient de tete
template <class T>
const T  Poly1<T>::leadcoef() const 
{ return (*this)[degree()] ; }


// Retourne le coefficient de base
template <class T>
const T  Poly1<T>::tcoef() const 
{ return (*this)[val()] ; }


// Retourne le polynome derive 
template <class T>
const Poly1<T> Poly1<T>::diff() const 
{
   Degree deg_autre_nom = degree() ;
   if ((deg_autre_nom == deginfty) || (deg_autre_nom == 0))
       return Poly1<T>(zero) ;

   Poly1<T> Res(indeter(), deg_autre_nom-1) ; 
   T One = one ;
   T ii = One ;
   for (int i = 0 ; i < deg_autre_nom; i++, ii +=  One )
         Res[i] = (*this)[i+1] * ii ; 
   return Res ;
}


// Retourne le polynome derive 
template <class T>
ulong length(const Poly1<T> &P)
{
   Degree deg_autre_nom = P.degree() ;
   ulong sum = 0UL ;
   if (deg_autre_nom == deginfty) return 0UL ;
   for (int i = 0 ; i <= deg_autre_nom ; i++) 
        sum += length(P[i]) ;
   return sum ;
}

