#ifndef _STURM_QUERY_
#define _STURM_QUERY_

// ============================================================ //
// Univariate polynomial, file poly1sq.C
// class Poly1<T>
// (c) copyright GIVARO 1997
// author: J.G. Dumas
// date: 22.04.1997
// ============================================================ //
//
// Strum-Query 
//
// ============================================================ //



template<class T>
inline int sturm_query(const Poly1<T>& P, const Poly1<T>& Q) {

  Poly1<T> B;
  Poly1<T> tmp;
  Poly1<T> A(P.indeter(),P.degree(),P.leadcoef());

  for(int aa=0;aa<=P.degree();aa++)
	A[aa] = P[aa];

  B = A.diff() * Q;
  Degree p=A.degree(),q=B.degree();
  T nul(zero);

  int* sm = new int [p+3];
  int* sp = new int [p+3];
  T temp;

  temp = A.leadcoef();
  if (temp < nul) 
    sp[0] = -1;
  else 
    sp[0] = 1;
  if (A.degree() % 2)
    sm[0] = sp[0] * -1;
  else 
    sm[0] = sp[0];

  temp = B.leadcoef();
  if (temp < nul)
    sp[1] = -1;
  else
    sp[1] = 1;
  if (B.degree() % 2)
    sm[1] = sp[1] * -1;
  else
    sm[1] = sp[1];

  int nb(2);

  // signs in + and - infinity of the Sturm-Sylvester's monic sequence
  while (!iszero(tmp = A % B)) {
    A = B ;
    if (nul > (temp = tmp.leadcoef())) {
        sp[nb] = 1;
        B = tmp / temp;
    }
    else {
        sp[nb] = -1;
        B = tmp / -temp;
    }
    if (tmp.degree() % 2)
        sm[nb] = sp[nb] * -1;
    else
        sm[nb] = sp[nb];
    nb++;
  }

  int wm(0),wp(0),prec,etat,cz;

  // # of sign change in - infinity
  cz = 0 ;
  prec=sm[0];
  while (!prec) prec=sm[++cz];
  int i = cz;
  for(;i<nb;i++) {
    cz = 0;
    etat=sm[i];
    while ((!etat) && ((i+cz)<nb)) etat=sm[i + ++cz];
    if (!etat) break;
    if (etat != prec)
      wm++;
    else
      if (cz == 2)
        wm += 2;
    prec = etat;
    i += cz;
  }

  // # of sign change in + infinity
  cz = 0 ;
  prec=sp[0];
  while (!prec) prec=sp[++cz];
  for(i=cz;i<nb;i++) {
    cz = 0;
    etat=sp[i];
    while ((!etat) && ((i+cz)<nb)) etat=sp[i + ++cz];
    if (!etat) break;
    if (etat != prec)
      wp++;
    else
      if (cz == 2)
        wp += 2;
    prec = etat;
    i += cz;
  }

  return wm-wp;
};




#endif

