// =========================================================================== //
// Univariate polynomial
// class Poly1<T>
// (c) copyright GIVARO 1994
// author: Th. Gautier....
// date: 1994
// =========================================================================== //


// Sqrfree decomposition 
// Nfact : is the number of factor in the sqrfree decomposition
// Fact : is a array of dimension degre + 1 of P which contains the factors
// ie P = Fact[0]^0 * Fact[1]^1 * ... * Fact[P.degre()]^(P.degre())
// with Fact[0] the leading coefficient.
// The array Fact must be allocated before calling the function.
// The size of Fact must be degP+1. 
// For more readeable version of the algorithm, see Geddes, p342.

template <class T>
void sqrfree(const Poly1<T>& P, int& Nfact, Poly1<T>* Fact)
{
  Degree count = 1 ;
  Indeter X = P.indeter() ;
  Poly1<T> A,B,C,W,Z,Y ;
  Poly1<T> One = one ;
  Fact[0] = P.leadcoef() ;
  
  A = P / Fact[0] ;
  B = A.diff() ;
  C = gcd(A,B) ;
  
  if (C == One) W = A ;
  else {
    W = A / C ;
    Y = B / C ;
    Z = Y - W.diff() ;
    while (!iszero(Z)) {
       Fact[count] = gcd(W,Z) ;
        W = W / Fact[count] ;
        Y = Z / Fact[count] ;
        Z = Y - W.diff() ;
        count++ ;
    } 
  }
  Fact[count] = W ;
  Nfact = 1+ count ;
  return ;
} 
