#ifndef _PACTYPE_H_
#define _PACTYPE_H_
// =========================================================================== //
// Basic type definitions
// (c) copyright GIVARO 1994
// author: Th. Gautier
// date: 1996
// =========================================================================== //
#include <stdlib.h> // for size_t
#include "givbase.h"

// Neutral type:
enum Neutral { zero = 0L , one = 1L, none = -1  } ;

// Usefull definition de ulong, uint, uchar :
typedef unsigned short ushort ;
typedef unsigned int uint ;
typedef unsigned long ulong ;
typedef unsigned long long ulonglong ;
typedef unsigned char uchar ;

typedef void* LocalAdr ; 

// Indice type for the matrix and vector classes.
typedef long Indice ;
typedef unsigned long UIndice ;


// Length function for basic type.
// Returns a size in byte.
template <class T>
inline ulong length( const T& ) { return sizeof(T) ; }
inline ulong length( const short ) { return sizeof(short) ; }
inline ulong length( const int ) { return sizeof(int) ; }
inline ulong length( const long ) { return sizeof(long) ; }
inline ulong length( const char ) { return sizeof(char) ; }
inline ulong length( const float ) { return sizeof(float) ; }
inline ulong length( const double ) { return sizeof(double) ; }
inline ulong length( const long double ) { return sizeof(long double);}

/*
template <class T>
inline int iszero (const T& a) { return a ==0 ; }
*/
inline int iszero( const float a) { return a == 0.0 ; }
inline int iszero( const double a) { return a == 0.0 ; }
inline int iszero( const long double a) { return a == 0.0 ; }
inline int iszero( const short a) { return a == 0 ; }
inline int iszero( const int a) { return a == 0 ; }
inline int iszero( const long a) { return a == 0 ; }
inline int iszero( const ulong a) { return a == 0 ; }
inline int iszero( const long long a) { return a == 0 ; }
inline int iszero( const ulonglong a) { return a == 0 ; }
inline int iszero( const char a) { return a == '0' ; }

#ifndef GIVMIN
#define GIVMIN(a,b) ((a)<(b)?(a):(b))
#endif

#ifndef GIVMAX
#define GIVMAX(a,b) ((a)<(b) ? b:a)
#endif

#ifndef GIVABS
#define GIVABS(l) ((l)<0 ? -(l): (l))
#endif



template<class XXX> struct Signed_Trait {
    typedef XXX self_type;
    typedef XXX signed_type;
//     typedef XXX unsigned_type;
};

template<> struct Signed_Trait<float> {
    typedef float self_type;
    typedef float signed_type;
    typedef ulong unsigned_type;
};

template<> struct Signed_Trait<double> {
    typedef double self_type;
    typedef double signed_type;
    typedef ulong unsigned_type;
};


template<> struct Signed_Trait<short> {
    typedef short self_type;
    typedef short signed_type;
    typedef ushort unsigned_type;
};

template<> struct Signed_Trait<uchar> {
    typedef uchar self_type;
    typedef short signed_type;
    typedef uchar unsigned_type;
};

template<> struct Signed_Trait<char> {
    typedef char self_type;
    typedef char signed_type;
    typedef uchar unsigned_type;
};

template<> struct Signed_Trait<long> {
    typedef long self_type;
    typedef long signed_type;
    typedef ulong unsigned_type;
};

template<> struct Signed_Trait<ulong> {
    typedef ulong self_type;
    typedef long signed_type;
    typedef ulong unsigned_type;
};

template<> struct Signed_Trait<long long> {
    typedef long long self_type;
    typedef long long signed_type;
    typedef ulonglong unsigned_type;
};


template<> struct Signed_Trait<ulonglong> {
    typedef ulonglong self_type;
    typedef long long signed_type;
    typedef ulonglong unsigned_type;
};



#endif
