#ifndef _GIVARO_STL_ARRAY_ALGS_H_
#define _GIVARO_STL_ARRAY_ALGS_H_

// ==========================================================================
// file: stlarrays.h 
// Time-stamp: <23 Jun 99 13:55:19 Jean-Guillaume.Dumas@imag.fr>
// (c) Givaro Team
// ==========================================================================

// -------------------------------------------------------------
// Requirements : iszero needed for V::value_type
// -------------------------------------------------------------
// -------------------------------------------------------------
// Poly in Arrays
// -------------------------------------------------------------
// -------------------------------------------------------------
// Multiplication
// -------------------------------------------------------------
template<class V, class DEGREE> void vector_mod_mul(V& R, const V& A, const V& B, const DEGREE db) {
        // Precondition : size A > 0
        // Precondition : size R >= size A + size B - 1
    typename V::const_iterator ai=A.begin(),bi=B.begin();
    typename V::iterator ri=R.begin(), rig=R.begin();
    DEGREE idb=0;
    if (iszero(*ai))
        for(;idb<=db;++bi,++ri,++idb)
            *ri = *ai;
    else
        for(;idb<=db;++bi,++ri,++idb)
            if (iszero(*bi))
                *ri = (*bi);
            else
                *ri = (*ai) * (*bi);
    typename V::value_type Zero(zero);
    for(;ri!=R.end();++ri)
        *ri = Zero;
    for(++ai,++rig;ai!=A.end();++ai,++rig)
        if (! iszero(*ai))
            for(ri=rig,bi=B.begin(),idb=0;idb<=db;++bi,++ri,++idb)
                if (! iszero(*bi))
                    *ri += (*ai) * (*bi);
}               

// -------------------------------------------------------------
// In place Monic Remainder
// -------------------------------------------------------------
template<class V, class DEGREE> void vector_mod_monic_rem(V& A, DEGREE& da, const V& B, const DEGREE db) {
    typename V::value_type Zero(zero);
        // Precondition : B is monic
    typedef typename V::value_type TT;
    TT l;
    typename V::iterator ai,aai;
    typename V::const_iterator bi;
    while (da >= db) {
        ai = A.begin();
        l = *(A.begin());
        aai = A.begin();
        bi=B.begin();
        for(++bi,++ai;bi!=B.end();++bi,++ai,--da) {
            if (!iszero(*bi))
                *aai = *ai - ((l) * (*bi));
            else
                *aai = *ai;
            if (! iszero(*aai)) break;
        }
        if (bi!=B.end()) 
            for(++bi,++ai,++aai;bi!=B.end();++bi,++ai,++aai)
                if (!iszero(*bi))
                    *aai = *ai - ((l) * (*bi));
                else
                    *aai = *ai;
        for(;ai!=A.end();++ai,++aai)
            *aai = *ai;
        *aai = Zero;
        --da;
    }
}               

// -------------------------------------------------------------
// Horner's scheme for evaluation
// -------------------------------------------------------------
template<class V, class DEGREE, class UTT> UTT vector_mod_eval(const V& P, const DEGREE db, const UTT a) {
// Precondition : convertion form V::value_type to UTT is made via "residu()"
    typename V::const_iterator pi=P.begin();
    DEGREE dip = 0;
    UTT res = (*pi).residu() ;
    for (++pi;dip<db;++pi,++dip) {
        res *= a;
        res += (*pi).residu();
    }
    return res;
}

#endif
