#ifndef _TIMER_H_
#define _TIMER_H_
// ======================================================================= //
// Class Timer :
// (c) copyright GIVARO 1994
// author : T. Gautier
// date: 1994
// version :
// ======================================================================= //

 
class ostream ;
class RealTimer ; class SysTimer ; class UserTimer ;

class BaseTimer {
public:
  enum { 
    MSPSEC = 1000000  // microsecond per second
  } ;

  // Clear timer :
  void clear() { _t = 0 ; }

  // Set timer :
  void set(double s) { _t = s ; }

  // total amount of second spent 
  double time() const { return _t ; }

  // Return a value to initialize random generator
  static long seed() ;

  // Output the value of the timer :
  friend ostream& operator<<(ostream& o, const BaseTimer& T)  ;
       
  // Some arithmetic operators to compute cumulative time :
  BaseTimer& operator = (const BaseTimer & T)  ;
  const BaseTimer operator - (const BaseTimer & T)  const ;
  const BaseTimer operator - ()  ;
  const BaseTimer operator +  (const BaseTimer & T)  const ;
  const BaseTimer operator += (const BaseTimer & T) { return *this = *this + T; } ;
  const BaseTimer operator -= (const BaseTimer & T) { return *this = *this - T; } ;
  const BaseTimer operator / (int i) ;
  const BaseTimer& operator /= (int i) { return *this = *this / i; }


  // Pack and unpack :
  //friend a1cOStream& operator<< ( a1cOStream& o, const BaseTimer& T) ;
  //friend a1cIStream& operator>> ( a1cIStream& o, BaseTimer& T) ;
  
protected:
  double _t ;  // time  
};

class RealTimer : public BaseTimer {
public:
  // Start timer
  void start() ;

  // Stop timer 
  void stop() ;
} ;


class UserTimer : public BaseTimer {
public:
  // Start timer
  void start() ;

  // Stop timer 
  void stop() ;
} ;


class SysTimer : public BaseTimer {
public:
  // Start timer
  void start() ;

  // Stop timer 
  void stop() ;
} ;


class Timer : protected RealTimer, protected UserTimer, protected SysTimer {
  public :

  // Clear timer :
  void clear() ; 

  // Start timer
  void start() ;

  // Stop timer 
  void stop() ;

  // total amount of second spent in user mode
  double usertime() const { return UserTimer::_t ; }

  // set user mode
  void setuser(double s) { UserTimer::set(s); }

  // total amount of second spent in system mode
  double systime () const { return SysTimer::_t ; }

  // set system mode
  void setsys(double s) { SysTimer::set(s); }

  // real total amount of second spent.  
  double realtime () const { return RealTimer::_t ; }

  // set real mode
  void setreal(double s) { RealTimer::set(s); }

  // retourne une petite graine
  static long seed()  { return RealTimer::seed() ; }

  // Output the value of the timer :
  friend ostream& operator<<(ostream& o, const Timer& T)  ;
       
  // Some arithmetic operators to compute cumulative time :
  Timer& operator = (const Timer & T)  ;
  const Timer operator - (const Timer & T)  const ;
  const Timer operator - ()  ;
  const Timer operator + (const Timer & T)  const ;
  const Timer operator += (const Timer & T) { return *this = *this + T; }
  const Timer operator -= (const Timer & T) { return *this = *this - T; }
  const Timer operator/(int i);
  const Timer& operator/=(int i) { return  *this = *this / i; } 

  // Pack and unpack :
  //friend a1cOStream& operator<< ( a1cOStream& o, const Timer& T) ;
  //friend a1cIStream& operator>> ( a1cIStream& o, Timer& T) ;
  
} ;

#endif 
