// ==========================================================================
// file: givdomgfq.C
// Time-stamp: <27 Jul 00 12:24:43 Jean-Guillaume.Dumas@imag.fr>
// ==========================================================================
#include "givdomgfq.h"

#include <givrandom.h>
#include <givpower.h>
#include <stlarrays.h>
#include <zpz.h>
#include <intnumtheo.h>
#include <intfactor.C>
#include <intproot.C>
#include <poly1proot.C>

#include <vector>
using std::vector;

template<class TT> GFqDom<TT>::GFqDom(const GFqDom<TT>::UTT P, const GFqDom<TT>::UTT e) {
        // Precondition P prime
    _caracteristic = P ;
    _exponent = e ;
    zero = 0;
    _q = mod_power(P,e);
    _qm1 = _q - 1;
    one = _qm1;
//    _qm1o2 = _q >> 1;
    _qm1o2 = (P==2)?(one):(_q>>1) ; // 1 == -1 in GF(2^k)
//     _log2pol = new UTT[ _q ];
//     _pol2log = new UTT[ _q ];
//     _plus1 = new TT[ _q ];
    _log2pol.resize(_q);
    _pol2log.resize(_q);
    _plus1.resize(_q);
    UTT i;
    _log2pol[0] = 0;
    if (e == 1) {
        UTT seed = (UTT) ( prim_root( Integer(P) ) % P );
        UTT accu = 1;
        for(i=1; i<P; i++) {
            accu = (accu * seed) % P;
            _log2pol[i] = accu;
        }      
    } else {
            // Fisrt compute an irreductible polynomial F over Z/pZ of degree e
            // Then a primitive root G (i.e. a generator of GF(q))
        Indeter X("X");
        Modulo old_prime; Zpz::getZpz(old_prime);
        Zpz::setZpz(P);
        Poly1< Zpz > F(X),G(X);
        random_prim_root(F,G,e,P);
          // Then compute p-adic tables
          // log2pol[ i ] = G^i(p)     
        Degree df=F.degree(),dg=G.degree(),dh=df+dg,dhi;
        vector< Zpz > lF, lG;Degree i;
        for(i=df;i>=0;--i)
            lF.push_back(F[i]);
        for(i=dg;i>=0;--i)
            lG.push_back(G[i]);
      
        _log2pol[1] = vector_mod_eval(lG, dg, P);
        
        vector<Zpz>::iterator hi,qi;
        vector<Zpz> H(dh);
        vector<Zpz> Q(dh); 
        for(i=dg;i>=0;--i)
            H[i] = lG[i];
        dh = dg;
        for (i = 2; i < _q; ++i) {
            vector_mod_mul(Q,lG,H,dh);
            dh += dg;
            vector_mod_monic_rem(Q,dh,lF,df);
            for(hi=H.begin(),qi=Q.begin(),dhi=0;dhi<=dh;++hi,++qi,++dhi)
                *hi = *qi;
            _log2pol[i] = vector_mod_eval ( H, dh, P);
        }
 

        Zpz::setZpz(old_prime);
    }
        // pol2log[ j ] = i such that log2pol[i] = j
    for (i = 0; i < _q; ++i) 
        _pol2log[ _log2pol[i] ] = i;

      // plus1[i] = k such that G^i + 1 = G^k
    _plus1[0] = _qm1;
    UTT a,b,r;
    for (i = 1; i < _q; ++i) {
        a = _log2pol[i];
        r = a % P;
        if (r == (P - 1))
            b = a - r;
        else
            b = a + 1;
//        GFqDom<TT>::_plus1[i] = GFqDom<TT>::_pol2log[b]; 
        _plus1[i] = _pol2log[b] - _qm1; 
    }
   _plus1[_qm1o2] = 0; 
}


template<class TT> GFqDom<TT>::GFqDom() : zero(0),_log2pol(0),_pol2log(0),_plus1(0) {};

template<class TT> GFqDom<TT>::GFqDom( const GFqDom<TT>::UTT P ) 
{
   *this = GFqDom( P, UTT(one) );
}


template<class TT> GFqDom<TT>::~GFqDom() 
{
//    delete [] _log2pol;
//    delete [] _pol2log;
//    delete [] _plus1;
}

template<class TT> void GFqDom<TT>::Init() {} 

template<class TT> void GFqDom<TT>::End() {}


