// ==========================================================================
// file: givdomgfq.inl
// Description:
//   Arithmetic on GF(q)
// Bugs:
// Time-stamp: <13 Aug 99 11:49:11 Jean-Guillaume.Dumas@imag.fr> 
// ==========================================================================

// Warning : valid iff b != c
#define _GIVARO_GFQ_ADD(c, a, b, mun, plun) { if ((b)==0) (c)=(a); else if ((a)==0) (c)=(b); else { \
                                   (c) = (a)-(b); \
                                   (c) = ((c)>0)?(c):(c)+ (mun); \
                                   (c) = (plun)[(UTT)(c)]; \
                                   if (c) { \
                                        (c) = (c)+(b); \
                                        (c) = ((c)>0)?(c):(c)+(mun); \
                                   } } }

#define _GIVARO_GFQ_NEG(res, a, mo, mun) { if ( (a)==0 ) (res)=0; else\
                              { (res) = (a) - (mo) ; (res) = ((res)>0)?(res):(res)+(mun); } }

// Warning : valid iff a != c
// if not use AUTOSUB ...
#define _GIVARO_GFQ_SUB(c, a, b, mo, mun, plun) { if ((a)==0) {_GIVARO_GFQ_NEG(c,b,mo,mun);} else if ((b)==0) (c)=(a); else { \
                                   (c) = (b)-(a)-(mo); \
                                   (c) = ((c)>0)?(c):(c)+(mun); \
                                   (c) = ((c)>0)?(c):(c)+ (mun); \
                                   (c) = (plun)[(UTT)(c)]; \
                                   if (c) { \
                                        (c) = (c)+(a); \
                                        (c) = ((c)>0)?(c):(c)+(mun); \
                                   } } }

#define _GIVARO_GFQ_AUTOSUB(c, b, mo, mun, plun) { if ((c)==0) {_GIVARO_GFQ_NEG(c,b,mo,mun);} else if ((b)!=0) { \
                                   (c) = (c)-(b)-(mo); \
                                   (c) = ((c)>0)?(c):(c)+(mun); \
                                   (c) = ((c)>0)?(c):(c)+ (mun); \
                                   (c) = (plun)[(UTT)(c)]; \
                                   if (c) { \
                                        (c) = (c)+(b); \
                                        (c) = ((c)>0)?(c)-(mo):(c)+(mo); \
                                        (c) = ((c)>0)?(c):(c)+(mun); \
                                   } } }



#define _GIVARO_GFQ_MUL(res, a, b, mun) { if ( ((a)==0) || ((b)==0) ) (res) = 0; else \
                                  { (res) = (a)+(b)-(mun); \
                                    (res) = ((res)>0)?(res):(res)+ (mun); } }

// Warning :  if a==1, a /= a --> 0 !!!
#define _GIVARO_GFQ_INV(res, a, mun)    { (res) = (mun)-(a); (res)=(res)?(res):(a); }

#define _GIVARO_GFQ_DIV(res, a, b, mun) {  if ( (a)==0 ) (res)=0; else\
                                   { (res) = (a) - (b) ; (res) = ((res)>0)?(res):(res)+(mun) ; } }


// Warning : valid iff b != c
#define _GIVARO_GFQ_MULADD(c,a1,a2,b,mun,plun) { \
if (((a1)==0) || ((a2)==0)) (c)=(b); \
else if ((b)==0) { \
        (c)=(a1)+(a2) - (mun); (c) = ((c)>0)?(c):(c) + (mun); \
 } else { \
       (c) = (a1)+(a2)-(b)-(mun); \
       (c) = ((c)<0)?(c)+(mun):(c); \
       (c) = ((c)>0)?(c):(c)+(mun); \
       (c) = (plun)[(UTT)(c)]; \
       if (c) { \
                (c) = (c)+(b); \
                (c) = ((c)>0)?(c):(c)+(mun); } \
     }\
}

 // ------------------------- Accesses functions

template<class TT> inline GFqDom<TT>::Residu_t GFqDom<TT>::residu( ) const
{ return _q; }

template<class TT> inline GFqDom<TT>::Residu_t GFqDom<TT>::caracteristic( ) const
{ return _caracteristic; }

template<class TT> inline GFqDom<TT>::Residu_t GFqDom<TT>::exponent( ) const
{ return _exponent; }

template<class TT> inline GFqDom<TT>::Residu_t GFqDom<TT>::size( ) const
{ return _q; }

 // ------------------------- Miscellaneous functions

template<class TT> inline short GFqDom<TT>::iszero(const Rep a) const
  { return a == zero ; }

template<class TT> inline short GFqDom<TT>::isone(const Rep a) const
  { return a == one ; }

template<class TT> inline size_t GFqDom<TT>::length(const Rep ) const
  { return sizeof(TT) ;}

  // Input - Output 
template<class TT> inline std::istream& GFqDom<TT>::read (std::istream& i) {
    char ch;
    s >> ch;
    if (ch != '(')
        cerr << "GFqDom::read: syntax error: no '('" << endl;
    UTT p;
    s >> p;
    s >> ch;
    if (ch == ')') 
        *this = GFqDom<TT>(p,UTT(1));
    else {
        if (ch != '^')
            cerr << "GFqDom::read: syntax error: no '^'" << endl;
        UTT k;
        s >> k;
        if (ch != ')')
            cerr << "GFqDom::read: syntax error: no ')'" << endl;

        *this = GFqDom<TT>(p,k);
    }
}   

template<class TT> inline std::ostream& GFqDom<TT>::write (std::ostream& o) const {
    return o << '(' <<  _caracteristic << '^' << _exponent << ')';
}


template<class TT> inline std::istream& GFqDom<TT>::read (std::istream& i, Rep& a) const {
    i >> a;
    assign(a,a);
    return i;
}

template<class TT> inline std::ostream& GFqDom<TT>::write (std::ostream& o, const Rep a) const {
  return o << _log2pol[ (UTT)a] ;
}

template<class TT> inline GFqDom<TT>::Rep GFqDom<TT>::access(const Rep a) const {
  return _log2pol[ (UTT)a ] ;
}

template<class TT> inline GFqDom<TT>::UTT GFqDom<TT>::write(const Rep a) const {
  return _log2pol[ (UTT)a ] ;
}

  // ----------- Usefull method :
template<class TT> inline GFqDom<TT>::Rep& GFqDom<TT>::mul (Rep& r, const Rep a, const Rep b) const 
  { _GIVARO_GFQ_MUL(r,a,b, _qm1) ; return r; }

template<class TT> inline GFqDom<TT>::Rep& GFqDom<TT>::mulin (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a) const 
  { _GIVARO_GFQ_MUL(r,r,a, _qm1) ; return r; }

template<class TT> inline GFqDom<TT>::Rep& GFqDom<TT>::div (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a, const GFqDom<TT>::Rep b) const 
  { _GIVARO_GFQ_DIV(r, a, b, _qm1) ; return r; }

template<class TT> inline GFqDom<TT>::Rep& GFqDom<TT>::divin (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a) const 
  { _GIVARO_GFQ_DIV(r, r, a, _qm1) ; return r; }

template<class TT> inline GFqDom<TT>::Rep& GFqDom<TT>::add (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a, const GFqDom<TT>::Rep b) const 
  { _GIVARO_GFQ_ADD(r, a, b, _qm1, _plus1) ; return r; }

template<class TT> inline GFqDom<TT>::Rep& GFqDom<TT>::addin (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a) const 
  { _GIVARO_GFQ_ADD(r, r, a, _qm1, _plus1) ; return r; }

template<class TT> inline GFqDom<TT>::Rep& GFqDom<TT>::sub (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a, const GFqDom<TT>::Rep b) const 
  { _GIVARO_GFQ_SUB(r, a, b, _qm1o2, _qm1, _plus1) ; return r; }

template<class TT> inline GFqDom<TT>::Rep& GFqDom<TT>::subin (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a) const 
  { _GIVARO_GFQ_AUTOSUB(r, a, _qm1o2, _qm1, _plus1) ; return r; }

template<class TT> inline GFqDom<TT>::Rep& GFqDom<TT>::neg (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a) const 
  { _GIVARO_GFQ_NEG(r, a, _qm1o2, _qm1) ; return r; }

template<class TT> inline GFqDom<TT>::Rep& GFqDom<TT>::inv (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a) const 
  { _GIVARO_GFQ_INV(r, a, _qm1) ; return r; }

template<class TT> inline GFqDom<TT>::Rep& GFqDom<TT>::axpy (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a, const GFqDom<TT>::Rep b, const GFqDom<TT>::Rep c) const 
  { _GIVARO_GFQ_MULADD(r,a,b,c, _qm1, _plus1) ; return r; }

template<class TT> inline GFqDom<TT>::Rep& GFqDom<TT>::axpyin (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a, const GFqDom<TT>::Rep b) const { 
    Rep tmp = r; 
    _GIVARO_GFQ_MULADD(r,a,b,tmp, _qm1, _plus1) ; return r; 
}

template<class TT> inline GFqDom<TT>::Rep& GFqDom<TT>::axmy (GFqDom<TT>::Rep& r, const GFqDom<TT>::Rep a, const GFqDom<TT>::Rep b, const GFqDom<TT>::Rep c)  const { 
    _GIVARO_GFQ_MUL(r,a,b, _qm1) ; 
    _GIVARO_GFQ_AUTOSUB(r,c, _qm1o2, _qm1, _plus1) ; return r; 
}


 // -- inline array operations between ZpzDom<Log16>::Rep
template<class TT> inline void GFqDom<TT>::mul (const size_t sz, Array r, constArray a, constArray b) const {
    for (register size_t i=sz-1; i!=0; --i) {
        _GIVARO_GFQ_MUL(r[i],a[i],b[i], _qm1) ; 
    }
}

template<class TT> inline void GFqDom<TT>::mul (const size_t sz, Array r, constArray a, Rep b) const {
    for (register size_t i=sz-1; i!=0; --i) {
        _GIVARO_GFQ_MUL(r[i],a[i],b, _qm1) ; 
    }
}

template<class TT> inline void GFqDom<TT>::div (const size_t sz, Array r, constArray a, constArray b) const {
    for (register size_t i=sz-1; i!=0; --i) {
        _GIVARO_GFQ_DIV(r[i],a[i],b[i], _qm1) ; 
    }
}

template<class TT> inline void GFqDom<TT>::div (const size_t sz, Array r, constArray a, Rep b) const {
    for (register size_t i=sz-1; i!=0; --i) {
        _GIVARO_GFQ_DIV(r[i],a[i],b, _qm1) ; 
    }
}

template<class TT> inline void GFqDom<TT>::add (const size_t sz, Array r, constArray a, constArray b) const {
    for (register size_t i=sz-1; i!=0; --i) {
        _GIVARO_GFQ_ADD(r[i], a[i], b[i], _qm1, _plus1) ;
    }
}

template<class TT> inline void GFqDom<TT>::add (const size_t sz, Array r, constArray a, Rep b) const {
    for (register size_t i=sz-1; i!=0; --i) {
        _GIVARO_GFQ_ADD(r[i], a[i], b, _qm1, _plus1) ;
    }
}

template<class TT> inline void GFqDom<TT>::sub (const size_t sz, Array r, constArray a, constArray b) const {
    for (register size_t i=sz-1; i!=0; --i) {
        _GIVARO_GFQ_SUB(r[i], a[i], b[i], _qm1o2, _qm1, _plus1) ;
    }
}

template<class TT> inline void GFqDom<TT>::sub (const size_t sz, Array r, constArray a, Rep b) const {
    for (register size_t i=sz-1; i!=0; --i) {
        _GIVARO_GFQ_SUB(r[i], a[i], b, _qm1o2, _qm1, _plus1) ;
    }
}

template<class TT> inline void GFqDom<TT>::neg (const size_t sz, Array r, constArray a) const {
    for (register size_t i=sz-1; i!=0; --i) {
        _GIVARO_GFQ_NEG(r[i], a[i], _qm1o2, _qm1) ;
    }
}

template<class TT> inline void GFqDom<TT>::inv (const size_t sz, Array r, constArray a) const {
    for (register size_t i=sz-1; i!=0; --i) {
        _GIVARO_GFQ_INV(r[i], a[i], _qm1) ;
    }
}

template<class TT> inline void GFqDom<TT>::axpy (const size_t sz, Array r, Rep a, constArray x, constArray y) const {
    for (register size_t i=sz-1; i!=0; --i) {
        _GIVARO_GFQ_MULADD(r[i], a, x[i], y[i], _qm1, _plus1) ;
    }
}

template<class TT> inline void GFqDom<TT>::axpyin (const size_t sz, Array r, Rep a, constArray x) const {
    Rep tmp;
    for (register size_t i=sz-1; i!=0; --i) {
        tmp = r[i];
        _GIVARO_GFQ_MULADD(r[i], a, x[i], r[i], _qm1, _plus1) ;
    }
}

template<class TT> inline void GFqDom<TT>::axpy (const size_t sz, Array r, Rep a, constArray x, Rep y) const {
    for (register size_t i=sz-1; i!=0; --i) {
        _GIVARO_GFQ_MULADD(r[i], a, x[i], y, _qm1, _plus1) ;
    }
}

template<class TT> inline void GFqDom<TT>::axmy (const size_t sz, Array r, Rep a, constArray x, constArray y) const {
    for (register size_t i=sz-1; i!=0; --i) {
        _GIVARO_GFQ_MUL(r[i], a, x[i], _qm1) ;
        _GIVARO_GFQ_AUTOSUB(r[i], y[i], _qm1o2, _qm1, _plus1) ;
    }
}

template<class TT> inline void GFqDom<TT>::axmy (const size_t sz, Array r, Rep a, constArray x, Rep y) const {
    for (register size_t i=sz-1; i!=0; --i) {
        _GIVARO_GFQ_MUL(r[i], a, x[i], _qm1) ;
        _GIVARO_GFQ_AUTOSUB(r[i], y, _qm1o2, _qm1, _plus1) ;
    }
}

// ---------
// -- misc operations
// ---------
template<class TT> inline void GFqDom<TT>::assign( const size_t sz, Array r, constArray a ) const {
    TT tr;
    for (register size_t i=sz-1; i!=0; --i) {
        tr = a[i] ;
        if (tr <0) {
                // -a = b [p]
                // a = p-b [p]
            tr = -tr;
            if (tr >=_caracteristic ) tr = tr % _caracteristic ;
            if (tr)
                r[i] = _pol2log[ _caracteristic - tr ];
            else
                r[i] = 0;
        } else {
            if (tr >=_caracteristic ) tr = tr % _caracteristic ;
            r[i] = _pol2log[ tr ];
        }
    }
}


template<class TT> inline GFqDom<TT>::Rep& GFqDom<TT>::assign( Rep& r, const Rep residu ) const {
    Rep tr = residu ;
    if (tr <0) {
            // -a = b [p]
            // a = p-b [p]
        tr = -tr;
        if (tr >=_caracteristic ) tr = (UTT)tr % _caracteristic ;
        if (tr)
            return r = _pol2log[ _caracteristic - (UTT)tr ];
        else
            return r = 0;
    } else {
        if (tr >=_caracteristic ) tr = (UTT)tr % _caracteristic ;
        return r = _pol2log[ (UTT)tr ];
    }
}

template<class TT> inline GFqDom<TT>::Rep& GFqDom<TT>::assign( Rep& r, const Integer residu ) const {
    Integer Itr = residu ;
    Rep tr;
    if (Itr <0) {
            // -a = b [p]
            // a = p-b [p]
        Itr = -Itr;
        tr = Itr % _caracteristic ;
        if (tr)
            return r = _pol2log[ _caracteristic - (UTT)tr ];
        else
            return r = 0;
    } else {
        tr = Itr % _caracteristic ;
        return r = _pol2log[ (UTT)tr ];
    }
}

template<class TT> inline void GFqDom<TT>::init( Rep& r) const { r = zero; }

template<class TT> inline void GFqDom<TT>::dotprod 
( Rep& r, const size_t sz, constArray a, constArray b ) const {
    if (sz) {
        _GIVARO_GFQ_MUL(r,a[0],b[0],_qm1);
        Rep tmp;
// cerr << "d0:" << access(r) << endl;
        for( register int i= sz-1; i>0; --i) {
            _GIVARO_GFQ_MUL(tmp,a[i],b[i],_qm1);
            _GIVARO_GFQ_ADD(r,r,tmp,_qm1,_plus1);
// cerr << "d" << i << ":" << access(r) << endl;
        }
    } else
        r = zero;
}

    
   // ----- random generators
template<class TT> inline GFqDom<TT>::Rep& GFqDom<TT>::NONZEROGIVRANDOM(Rep& a) {
    do 
        a = Rep( (UTT)(lrand48()) % _q);
    while (iszero(a));
    a = (a<0?a+_q:a);
    return a;
}

template<class TT> inline GFqDom<TT>::Rep& GFqDom<TT>::GIVRANDOM(Rep& a) {
    a = Rep( (UTT)(lrand48()) % _q);
    a = (a<0?a+_q:a);
    return a;
}
    

    
