#include "zpz.h"

Modulo Zpz::p = 0;

Zpz::Zpz( const Modulo residu )
{
  r = residu ;
  if (r <0) r += Zpz::p ;
  if ((r <0) || (r >=Zpz::p )) r %= Zpz::p ;
  if (r <0) r += Zpz::p ;
}

Zpz::Zpz( const Modulo residu, const Modulo P )
{
  setZpz( P ) ;
  r = residu ;
  if ((r <0) || (r >=Zpz::p )) r %= Zpz::p ;
  if (r <0) r += Zpz::p ;
}


  // Returns d, and u and v such that u a + v b = d ;
Modulo Zpz::bezout(Modulo a, Modulo b, Modulo& u, Modulo& v)
{
   register Modulo u1,u2,u3 ;
   register Modulo v1,v2,v3 ;
   u1 = 1 ; u2 = 0 ; u3 = a ;
   v1 = 0 ; v2 = 1 ; v3 = b ;
   while (v3 != 0)
     {
        register Modulo q , t1, t2 ,t3 ;
        q = u3 / v3 ;
        t1 = u1 - q * v1 ; t2 = u2 - q * v2 ; t3 = u3 - q * v3 ;
        u1 = v1 ; u2 = v2 ; u3 = v3 ; v1 = t1 ; v2 = t2 ; v3 = t3 ;
     }
   u = u1 ; v = u2 ;
   return u3 ;
} ;

#include <iostream>
using std::cerr;
using std::endl;

void Zpz::inv (Modulo& r,  const Modulo a, const Modulo pp)
{
  Modulo u,v, d ;
  d = Zpz::bezout(a,pp,u,v) ;

  if (d == -1) { d = -d ;  u = -u ; v = -v ; }
  if (d != 1) {
      cerr << "*** division by zero in: Modulo div " << endl;
      cerr << a << '*'<< u << '+' << v << '*' << pp << '=' << u*a+v*pp << endl ;
      cerr << "and gcd = " << d << endl ;
      return ;
    }
  if (u <0) u += pp ;
  r = u ;
}

void Zpz::inv (Zpz& r,  const Zpz& a, const Modulo pp)
{
   Zpz::inv( r.r, a.r, pp ) ; 
}

void Zpz::inv (Zpz& r,  const Zpz& a)
{
   Zpz::inv( r.r, a.r, Zpz::p ) ; 
}

 // Implementation functions : between array and element
void Zpz::mul (int size, Zpz* r,  const Zpz* a,  const Zpz* b)
{
  for (int i=0 ; i<size ; i++)
    ZPZ_MUL(r[i], a[i], b[i], Zpz::p) ;
}

void Zpz::mul (int size, Zpz* r,  const Zpz* a,  const Zpz b)
{
  for (int i=0 ; i<size ; i++)
    ZPZ_MUL(r[i], a[i], b, Zpz::p) ;
}

void Zpz::div (int size, Zpz* r,  const Zpz* a,  const Zpz* b)
{
  for (int i=0 ; i<size ; i++)
    ZPZ_DIV(r[i], a[i], b[i], Zpz::p) ;
}

void Zpz::div (int size, Zpz* r,  const Zpz* a,  const Zpz b)
{
  for (int i=0 ; i<size ; i++)
    ZPZ_DIV(r[i], a[i], b, Zpz::p) ;
}

void Zpz::add (int size, Zpz* r,  const Zpz* a,  const Zpz* b)
{
  for (int i=0 ; i<size ; i++)
    ZPZ_ADD(r[i], a[i], b[i], Zpz::p) ;
}

void Zpz::add (int size, Zpz* r,  const Zpz* a,  const Zpz b)
{
  for (int i=0 ; i<size ; i++)
    ZPZ_ADD(r[i], a[i], b, Zpz::p) ;
}

void Zpz::sub (int size, Zpz* r,  const Zpz* a,  const Zpz* b)
{
  for (int i=0 ; i<size ; i++)
    ZPZ_SUB(r[i], a[i], b[i], Zpz::p) ;
}
void Zpz::sub (int size, Zpz* r,  const Zpz* a,  const Zpz b)
{
  for (int i=0 ; i<size ; i++)
    ZPZ_SUB(r[i], a[i], b, Zpz::p) ;
}

void Zpz::neg (int size, Zpz*  r,  const Zpz* a)
{
  for (int i=0 ; i<size ; i++)
    ZPZ_NEG(r[i], a[i], Zpz::p) ;
}


// -------------------------------------------------------------
void Zpz::setZpz( const Modulo P ) 
{
  Zpz::p = P ;
}
