/****************************************************************************
**
*W  matrix.C          Homology source                  Dumas & Saunders
**
*H  @(#)$Id: matrix.C,v 0.0 1999/08
**
*Y  Copyright (C)  1999,  Us guys
*T Time-stamp: <04 Aug 01 11:11:07 Jean-Guillaume.Dumas@imag.fr> 
**
*/

#include <givtype.h>
#include <stdio.h> 
#include "matrix.h"
#include "commentator.h"
Commentator GlobalCommentator;

// -------------------------------------------------
// GMP compatibility 
// -------------------------------------------------
#include "integer.h"
#include "compgmp.h"
void AssignInteger(Integer& Res, const mpz_t cc) {
    mpz_set( (mpz_ptr)&(Res.gmp_rep), cc);
}
void AssignMPZ(mpz_t& cc, const Integer m) {
    mpz_init_set_str( cc, m.Int2str(), 10 );
}


#include "intnumtheo.h"
// -------------------------------------------------
// Is prime power
// -------------------------------------------------
long isprimepower(long& l, long a) {
    long ll = l = a;
    long r = 1;
    while (r && (! isprime(ll))) {
        r *= isnpower(l,ll);
        ll = l;
    }
    if (r)
        return r;
    return 0;
}


// -------------------------------------------------
// Field
// -------------------------------------------------
#include <zpz.h>
#define NEWFIELD(a,b) { Zpz::setZpz(a); }

// -------------------------------------------------
// Sparse structure and Methods
// -------------------------------------------------

#include "lin_sparse_b_b.h" 
#include "lin_smith.C"
#include "lin_cassini.h"
#include "lin_dom_smith.C"


typedef BlackBox< _IBB_VECTOR_ < Sparse_Vector < long > > > SPBBlong;
typedef BlackBox< _IBB_VECTOR_ < Sparse_Vector < Zpz > > > SPBBZpz;
template<> struct FT<Zpz> { typedef muladd_field field_info; };

typedef SPBBlong::coefficientSpace T;
typedef SPBBlong::Elements E;

#define MINERRORLEVEL 0.1

template<class BB>
void MatrixSmithFormValenceCPP (Commentator & Comm,
                                BB& TheMat,
                                long ni, long nj,
                                char * File_Name,
                                int Method, int Mode, const Integer& R,
                                Invariants<ulong>& Inv, 
				char * numerator, char * denominator) {
//cerr << "Received error: " << numerator << "/" << denominator << endl;
    double error = Integer(numerator).Int2double()/Integer(denominator).Int2double();
//cerr << "Computed error: " << error << endl;
        // Here I make a copy of the Commentator
    GlobalCommentator = Comm;

    startActivity(GlobalCommentator, "SNFV", "Smith Normal Form of Integer Matrix using Valence algorithm", LVL_NORMAL, INTERNAL_DESCRIPTION ); 
    activityReport(GlobalCommentator, "Loading the matrix ...", LVL_BLABLA, INTERNAL_DESCRIPTION);

    Inv = Invariants<ulong>(ni, nj);
    double wrong = 0;
    srand48(BaseTimer::seed());
    if (Mode == SMITHFORM) {
     
            // Weidemann Smith NOT YET IMPLEMENTED but INTEGRANK and MODULRANK are ...
        if (Method == hmValenceBB) {
//             activityReport(GlobalCommentator, "Sorry, Complete Smith Form via Weidemann is NOT YET AVAILABLE.\nBut Integer and prime ranks and P-Torsion with P square-free are.",LVL_NORMAL, INTERNAL_ERROR);
            unsigned long p1;
            Integer valence;
            UIndice rank;
            SPBBZpz A;
            Cassini<double>::Type_t OCB;
            {
                startActivity(GlobalCommentator, "Gersh", "Cassini Ovals method, to create a bound on the coefficients of A^t . A", LVL_NORMAL, INTERNAL_DESCRIPTION ); 
                { Cassini<double>(File_Name).operator() (OCB); }
                char * UT = new char[70]; sprintf(UT,"Cassini eigen value bound : %f", OCB); 
                stopActivity(GlobalCommentator, UT, LVL_IMP, PARTIAL_RESULT ); 
            }
            { seq_Integer_valence( File_Name, valence, rank, A, OCB, wrong, error); }
            { wiedemann_rank_N_torsion(Inv, File_Name, valence); }
        } else {
            Cassini<double>::Type_t OCB;
            {
                startActivity(GlobalCommentator, "Gersh", "Cassini Ovals method, to create a bound on the coefficients of A^t . A", LVL_NORMAL, INTERNAL_DESCRIPTION ); 
                { Cassini<double>(File_Name).operator() (OCB); }
                char * UT = new char[70]; sprintf(UT,"Cassini eigen value bound : %f", OCB); 
                stopActivity(GlobalCommentator, UT, LVL_IMP, PARTIAL_RESULT ); 
            }
            SPBBZpz A(TheMat);
            seq_Integer_smith(Inv, TheMat, A, OCB, wrong, error);
            char * UT = new char[70]; sprintf(UT,"Smith form correct with probability at least 1 - %e", wrong);
            activityReport(GlobalCommentator, UT, LVL_NORMAL, PARTIAL_RESULT);
	}
    } else if (Mode == INTEGRANK) {
        Cassini<double>::Type_t OCB;
        {
            startActivity(GlobalCommentator, "Gersh", "Cassini Ovals method, to create a bound on the coefficients of A^t . A", LVL_NORMAL, INTERNAL_DESCRIPTION ); 
            { Cassini<double>(File_Name).operator() (OCB); }
            char * UT = new char[70]; sprintf(UT,"Cassini eigen value bound : %f", OCB); 
            stopActivity(GlobalCommentator, UT, LVL_IMP, PARTIAL_RESULT ); 
        }
        unsigned long p1;
    	Integer valence;
        UIndice rank;
        if (Method == hmValenceBB) {
            SPBBZpz A;
            seq_Integer_valence( File_Name, valence, rank, A, OCB, wrong, error/2.0);
	} else {
            SPBBZpz A(TheMat);
            seq_Integer_valence(valence, rank, TheMat, A, OCB, wrong, error/2.0);
	}
 	double int_wrong = wrong;
        UIndice r;
 	double rank_wrong = 0;
        if (Method == hmValenceBB) {
            do 
                p1 = (prevprime(1+MAXGFQSIZE - (ulong)lrand48()%(ulong)MAXGFQSIZE/20) ).Int2long();
            while ((! isprime(p1)) || (! isone(gcd(valence,p1))));
            wiedemann_rank(p1, File_Name, r, rank_wrong);
            if (rank_wrong < 1) {
                    // Taking error into account ...
                int nbr = 0;
                if (error == 0) error = MINERRORLEVEL;
                while (error <= (int_wrong = rank_wrong+wrong)){ 
                    double r_w = 0;
                    do
                        p1 = (prevprime(1+MAXGFQSIZE - (ulong)lrand48()%(ulong)MAXGFQSIZE) ).Int2long();
                    while ((! isprime(p1)) || (! isone(gcd(valence,p1))));
                    UIndice rr;
                    wiedemann_rank(p1, File_Name, rr, r_w);
                    if (rr > r) {        // Previous ranks were bad
                        nbr = 0;
                        rank_wrong = r_w;
                        r = rr;
                    } else if (rr == r) { // another rank
                        rank_wrong = (r_w)*(rank_wrong);
                    } else               // bad rank (happens with proba wrong)
                        --nbr;
                }
            } else {
                    // else is impossible : heuristic did not gave probability.
                    // We therefore just do it twice.
                int nbr = 0;
                while ( ++nbr < 2 ){ 
                    double r_w = 0;
                    do
                        p1 = (prevprime(1+MAXGFQSIZE - (ulong)lrand48()%(ulong)MAXGFQSIZE) ).Int2long();
                    while ((! isprime(p1)) || (! isone(gcd(valence,p1))));
                    UIndice rr;
                    wiedemann_rank(p1, File_Name, rr, r_w);
                    if (rr > r) {        // Previous ranks were bad
                        nbr = 0;
                        rank_wrong = r_w;
                        r = rr;
                    } else if (rr == r) { // another rank
                        rank_wrong = (r_w)*(rank_wrong);
                    } else               // bad rank (happens with proba wrong)
                        --nbr;
                }
                int_wrong = 1;
            }
        } else {
            do 
                p1 = (prevprime(MAXPRIMESIZE - (ulong)lrand48()%(ulong)MINLOWERSIZE) ).Int2long();
            while ((! isprime(p1)) || (! isone(gcd(valence,p1))));
            gauss_rank(p1, TheMat, r, Density());
	}

        char * UT = new char[130];
        if (int_wrong == 1)
            sprintf(UT,"Rank %ld mod %ld is the Integer rank with unknown probability ", r, p1);
        else
            sprintf(UT,"Rank %ld mod %ld is the Integer rank with probability at least 1 - %e", r, p1, int_wrong );
        activityReport(GlobalCommentator, UT, LVL_NORMAL, PARTIAL_RESULT);
        Inv.PushInvariant(1,r);
    } else if (Mode == TORSIONRANK) {
        if (Method == hmValenceBB) {
            wiedemann_rank_N_torsion(Inv, File_Name, R);
        } else {
            SPBBZpz A(TheMat);
            gauss_rank_N_torsion(Inv, TheMat, A, R);
        }
    } else if (Mode == PRIMERANK) {
        UIndice r;
        if (Method == hmValenceBB) {
            double rank_wrong;
            wiedemann_rank(R.Int2long(), File_Name, r, rank_wrong);
        } else {
            gauss_rank(R.Int2long(), TheMat, r, Density());
        }
        Inv.PushInvariant(R.Int2long(),r);
    } else if (Mode == PRIMEPOWERRANK) {
        if (Method == hmValenceBB) {
            activityReport(GlobalCommentator, "Sorry, prime power rank via Wiedemann is NOT YET AVAILABLE.",LVL_NORMAL, INTERNAL_ERROR);
        } else {
            long expo, prime;
            expo = (long)isprimepower(prime, R.Int2long());
            UIndice r;
            gauss_rank(R.Int2long(), prime, TheMat, r, Density());
            Inv.PushInvariant(R.Int2long(),r);
        }
    }


    stopActivity(GlobalCommentator, "End of SNFV", LVL_NORMAL, INTERNAL_DESCRIPTION);
}


extern "C" void MatrixSmithFormValence (Commentator * Comm,
                                        const TSparseMatrixArray * Matrix,
                                        char * File_Name,
                                        TCount MaxRank, TCount Dim,
                                        int Method, int Mode, const mpz_t * coeff,
                                        TSmithForm * SmithForm,
					char * numerator, char * denominator) {


    Invariants<ulong> Inv;
    Integer R; AssignInteger(R, *coeff);
    SPBBlong TheMat;
    long ni, nj;
    
    if (Matrix) {
            // Changing matrix structure
        ni = Matrix->NumRows; nj = Matrix->NumColumns;
        TheMat = SPBBlong(ni,nj);
        for (long i = ni; i--;) { 
            TheMat[i] = Sparse_Vector< T >( Matrix->Rows[i].Count );
            for (long j = TheMat[i].size(); j--;)
                TheMat[i][j] = E( Matrix->Rows[i].Components[j].Index, Matrix->Rows[i].Components[j].Val );
        }
    } else {
        ni = TheMat.get_n_row(File_Name);
        nj = TheMat.n_col();
        if (Method != hmValenceBB) {
            if (ni > nj)
                TheMat.fileset_transpose(File_Name);
            else
                TheMat.fileset(File_Name);
        }
    }

    MatrixSmithFormValenceCPP(*Comm, TheMat, ni, nj, File_Name, Method, Mode, R, Inv, numerator, denominator);
    
    SmithForm->Rows = Inv.n_row();
    SmithForm->Columns = Inv.n_col();
    SmithForm->NumberOfInvariantFactors = Inv.size();
    TInvariantFactor * factors = new TInvariantFactor[Inv.size()];
    for (long i=0; i<Inv.size(); ++i) {
        AssignMPZ(factors[i].Coefficient, Integer( Inv.getvalue(i) ));
        factors[i].Count = Inv.getexponent(i);
    }

    SmithForm->InvariantFactors = factors;
};

